#!/usr/bin/perl
#!/usr/bin/perl -d:ptkdb

###############################################################################
#
# geotag.pl -  tag JPEG images with GPS data from database.
#
# Copyright (C) 2011  Rainer Walther  (rainerwalther-mail@web.de)
#
# Creative Commons Lizenz mit den Zusaetzen (by, nc, sa)
#
# Es ist Ihnen gestattet: 
#     * das Werk vervielfltigen, verbreiten und ffentlich zugnglich machen
#     * Abwandlungen bzw. Bearbeitungen des Inhaltes anfertigen
# 
# Zu den folgenden Bedingungen:
#     * Namensnennung.
#       Sie mssen den Namen des Autors/Rechteinhabers in der von ihm festgelegten Weise nennen.
#     * Keine kommerzielle Nutzung.
#       Dieses Werk darf nicht fr kommerzielle Zwecke verwendet werden.
#     * Weitergabe unter gleichen Bedingungen.
#       Wenn Sie den lizenzierten Inhalt bearbeiten oder in anderer Weise umgestalten,
#       verndern oder als Grundlage fr einen anderen Inhalt verwenden,
#       drfen Sie den neu entstandenen Inhalt nur unter Verwendung von Lizenzbedingungen
#       weitergeben, die mit denen dieses Lizenzvertrages identisch oder vergleichbar sind.
# 
# Im Falle einer Verbreitung mssen Sie anderen die Lizenzbedingungen, unter welche dieses
# Werk fllt, mitteilen. Am Einfachsten ist es, einen Link auf diese Seite einzubinden.
# 
# Jede der vorgenannten Bedingungen kann aufgehoben werden, sofern Sie die Einwilligung
# des Rechteinhabers dazu erhalten.
# 
# Diese Lizenz lsst die Urheberpersnlichkeitsrechte unberhrt.
# 
# Weitere Details zur Lizenzbestimmung gibt es hier:
#   Kurzform: http://creativecommons.org/licenses/by-nc-sa/3.0/de/
#   Komplett: http://creativecommons.org/licenses/by-nc-sa/3.0/de/legalcode
#
###############################################################################
# 2011-02-10 0.0.1 rw created
#
###############################################################################

$Version = "0.0.1 - 2011-02-10";

# change working directory to program path
my $Cwd = substr ($0, 0, rindex ($0, "geotag.pl"));
chdir $Cwd;

# set path for local Perl libs
push @INC, $Cwd . "perl/lib";

use XML::Simple;                  # http://search.cpan.org/dist/XML-Simple-2.18/lib/XML/Simple.pm
use DBI;                          # http://search.cpan.org/~timb/DBI-1.615/DBI.pm
use Image::ExifTool qw(:Public);  # http://search.cpan.org/~exiftool/Image-ExifTool-8.25/lib/Image/ExifTool.pod
use Time::Local;

# Read Mission Cockpit configuration
$XmlConfigFile = "mkcockpit.xml";
$Cfg = XMLin($XmlConfigFile);

# Commandline parameter
%CmdLine = @ARGV;

# Copy Cfg-CmdLine parameter to Cfg-Hash
foreach my $Key ( keys %CmdLine )
    {
    if ( $Key =~ /^-(\S*):(\S*)/i )
        {
        $Section = $1;
        $Keyword = $2;
        $Cfg->{$Section}->{$Keyword} = $CmdLine{$Key};
        }
    }

#
# connect to database
#

# DSN examples
# MSSQL Server: DBI:ODBC:driver={SQL Server};Server=NOTEBOOK\SQLEXPRESS;database=MissionCockpit
# SQLite:       DBI:SQLite:dbname=plugin/DbLog.sqlite
# MySQL:        DBI:mysql:database=MissionCockpit;host=localhost

# Configuration
my $DSN1            = $Cfg->{'dblog'}->{'DSN1'}    || 'DBI:ODBC:driver={SQL Server};Server=NOTEBOOK\\SQLEXPRESS;database=MissionCockpit';
my $DbUser1         = $Cfg->{'dblog'}->{'DbUser1'} || "";
my $DbPwd1          = $Cfg->{'dblog'}->{'DbPwd1'}  || "";
my $DSN2            = $Cfg->{'dblog'}->{'DSN2'}    || 'DBI:SQLite:dbname=plugin/DbLog.sqlite';
my $DbUser2         = $Cfg->{'dblog'}->{'DbUser2'} || "";
my $DbPwd2          = $Cfg->{'dblog'}->{'DbPwd2'}  || "";
my $DbLogTable      = $Cfg->{'dblog'}->{'DbLogTable'} || "FlightLog";

# connect to primary DSN
my $DSN = $DSN1;
$dbh = DBI->connect("$DSN","$DbUser1","$DbPwd1",{ PrintError => 1, RaiseError => 0, AutoCommit => 1 });

if ( ! $dbh )
    {
    # connect to secondary DSN
    $DSN = $DSN2;
    $dbh = DBI->connect("$DSN","$DbUser2","$DbPwd2",{ PrintError => 1, RaiseError => 0, AutoCommit => 1 });
    }

if ( ! $dbh )
    {
    print "DbLog: Can't connect to DSN1 or DSN2 ... exit\n";
    exit 1;
    }

print "\ngeotag.pl: Connected to DSN: \'$DSN\'\n";

#
# read Image files
#

$Path = ".";
if ( $#ARGV == 0 )
    {
    $Path = $ARGV[0];
    }
elsif ( $CmdLine{'-path'} ne "" )
    {
    $Path = $CmdLine{'-path'};
    }

if ( -f $Path)
    {
    # is a single file
    &GeoTag ($Path, $CmdLine{'-geosync'});
    }
elsif ( -d $Path )
    {
    # is a directory
    opendir IMG, "$Path";
    my @Images = readdir IMG;
    closedir IMG;

    foreach $Image (@Images)
        {
        chomp $Image;
        if ( $Image =~ /\.jpg$/i )
            {
            &GeoTag ("${Path}/${Image}", $CmdLine{'-geosync'});
            }
        }
    }
else
    {
    print "Invalid path: $Path\n";
    }

# the end
exit 0;


# geotag one image
sub GeoTag()
    {
    my ($Image, $GeoSync) = @_;

    # Read EXIF creation date
    my $ExifTool = new Image::ExifTool;
    my $ImgInfo  = $ExifTool->ImageInfo($Image);
    my $CreateDate  = $$ImgInfo{'CreateDate'};

    if ( $CreateDate ne "" )
        {
        # take EXIF time

        $CreateDate =~ /(\d+):(\d+):(\d+)\s*(\d*):(\d*):(\d*)/;
        my $year = $1;
        my $mon  = $2 - 1;
        my $mday = $3;
        my $hour = $4;
        my $min  = $5;
        my $sec  = $6;
        $ImageTime = timelocal($sec,$min,$hour,$mday,$mon,$year);
        }
    else
        {
        # take file creation date

        my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime,$ctime,$blksize,$blocks) = stat($Image);
        $ImageTime = $mtime;
        }

    if ( $GeoSync ne "" )
        {
        # image time correction
        $ImageTime += $GeoSync;
        }

    # try to find dataset in db for exact timestamp
    $SearchTime = $ImageTime;
    my $Sql = "SELECT CurPosLat, CurPosLon, CurPosAlt, GroundSpeed, Heading, CompassHeading from $DbLogTable WHERE Timestamp = \'$SearchTime\'";
    ($CurPosLat, $CurPosLon, $CurPosAlt, $GroundSpeed, $Heading, $CompassHeading) = $dbh->selectrow_array($Sql);

    $SearchTime = $ImageTime;
    while ( ($CurPosLat eq ""  or  $CurPosLon eq "")  and  $ImageTime - $SearchTime < 2 )
        {
        # timestamp not found, try some seconds before

        $SearchTime --;

        my $Sql = "SELECT CurPosLat, CurPosLon, CurPosAlt, GroundSpeed, Heading, CompassHeading from $DbLogTable WHERE Timestamp = \'$SearchTime\'";
        ($CurPosLat, $CurPosLon, $CurPosAlt, $GroundSpeed, $Heading, $CompassHeading) = $dbh->selectrow_array($Sql);
        }

    $SearchTime = $ImageTime;
    while ( ($CurPosLat eq ""  or  $CurPosLon eq "")  and  $SearchTime - $ImageTime < 2)
        {
        # not found, try some seconds after

        $SearchTime ++;

        my $Sql = "SELECT CurPosLat, CurPosLon, CurPosAlt, GroundSpeed, Heading, CompassHeading from $DbLogTable WHERE Timestamp = \'$SearchTime\'";
        ($CurPosLat, $CurPosLon, $CurPosAlt, $GroundSpeed, $Heading, $CompassHeading) = $dbh->selectrow_array($Sql);
        }

    if ( $CurPosLat ne ""  and  $CurPosLon ne "" )
        {
        # found db entry 

        printf "Image: %s Lat: %.6f Lon: %.6f Alt: %.2f\n", $Image, $CurPosLat, $CurPosLon, $CurPosAlt; 

        $CurPosLatRef = "N";
        if ( $CurPosLat < 0 )
            {
            $CurPosLatRef = "S";
            $CurPosLat = -$CurPosLat;
            }

        $CurPosLonRef = "E";
        if ( $CurPosLon < 0 )
            {
            $CurPosLonRef = "W";
            $CurPosLon = -$CurPosLon;
            }

        my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = gmtime($SearchTime);
        my $GPSTimeStamp = sprintf "%02d:%02d:%02d", $hour, $min, $sec;
        my $GPSDateStamp  = sprintf "%04d:%02d:%02d", $year+1900, $mon+1, $mday;

        $GroundSpeed *= 0.036;   # km/s

        # write EXIF to file
        # Tag details: http://www.awaresystems.be/imaging/tiff/tifftags/privateifd/gps.html
        ($success, $errStr) = $ExifTool->SetNewValue('GPSLatitude',        $CurPosLat);
        ($success, $errStr) = $ExifTool->SetNewValue('GPSLatitudeRef',     $CurPosLatRef);
        ($success, $errStr) = $ExifTool->SetNewValue('GPSLongitude',       $CurPosLon);
        ($success, $errStr) = $ExifTool->SetNewValue('GPSLongitudeRef',    $CurPosLonRef);
        ($success, $errStr) = $ExifTool->SetNewValue('GPSAltitude',        $CurPosAlt);
        ($success, $errStr) = $ExifTool->SetNewValue('GPSAltitudeRef',     0);
        ($success, $errStr) = $ExifTool->SetNewValue('GPSTimeStamp',       $GPSTimeStamp);
        ($success, $errStr) = $ExifTool->SetNewValue('GPSDateStamp',       $GPSDateStamp);
        ($success, $errStr) = $ExifTool->SetNewValue('GPSSpeed',           $GroundSpeed);
        ($success, $errStr) = $ExifTool->SetNewValue('GPSSpeedRef',        "K");
        ($success, $errStr) = $ExifTool->SetNewValue('GPSTrack',           $Heading);
        ($success, $errStr) = $ExifTool->SetNewValue('GPSTrackRef',        "T");
        ($success, $errStr) = $ExifTool->SetNewValue('GPSImgDirection',    $CompassHeading);
        ($success, $errStr) = $ExifTool->SetNewValue('GPSImgDirectionRef', "M");

        $ExifTool->WriteInfo($Image);
        }
    }

__END__
