#!/usr/bin/perl
#!/usr/bin/perl -d:ptkdb

###############################################################################
#
# libgui.pl -  MK Mission Cockpit - Subroutined for GUI Frontend
#
# Copyright (C) 2011  Rainer Walther  (rainerwalther-mail@web.de)
#
# Creative Commons Lizenz mit den Zusaetzen (by, nc, sa)
#
# Es ist Ihnen gestattet: 
#     * das Werk vervielfltigen, verbreiten und ffentlich zugnglich machen
#     * Abwandlungen bzw. Bearbeitungen des Inhaltes anfertigen
# 
# Zu den folgenden Bedingungen:
#     * Namensnennung.
#       Sie mssen den Namen des Autors/Rechteinhabers in der von ihm festgelegten Weise nennen.
#     * Keine kommerzielle Nutzung.
#       Dieses Werk darf nicht fr kommerzielle Zwecke verwendet werden.
#     * Weitergabe unter gleichen Bedingungen.
#       Wenn Sie den lizenzierten Inhalt bearbeiten oder in anderer Weise umgestalten,
#       verndern oder als Grundlage fr einen anderen Inhalt verwenden,
#       drfen Sie den neu entstandenen Inhalt nur unter Verwendung von Lizenzbedingungen
#       weitergeben, die mit denen dieses Lizenzvertrages identisch oder vergleichbar sind.
# 
# Im Falle einer Verbreitung mssen Sie anderen die Lizenzbedingungen, unter welche dieses
# Werk fllt, mitteilen. Am Einfachsten ist es, einen Link auf diese Seite einzubinden.
# 
# Jede der vorgenannten Bedingungen kann aufgehoben werden, sofern Sie die Einwilligung
# des Rechteinhabers dazu erhalten.
# 
# Diese Lizenz lsst die Urheberpersnlichkeitsrechte unberhrt.
# 
# Weitere Details zur Lizenzbestimmung gibt es hier:
#   Kurzform: http://creativecommons.org/licenses/by-nc-sa/3.0/de/
#   Komplett: http://creativecommons.org/licenses/by-nc-sa/3.0/de/legalcode
#
###############################################################################
# 2011-03-11 0.7.0 rw moved from libmkcockpit.pl
#                     Refresh time changed from 100ms to 200ms
# 2011-04-02 0.7.1 rw Bugfix Event export
# 2011-12-25 0.9.0 rw NC 0.28
# 2012-04-02 0.9.1 rw Check Package TableMatrix
#
###############################################################################

$Version{'libgui.pl'} = "0.9.1 - 2012-04-02";

eval "use Tk::TableMatrix" ||              # http://search.cpan.org/dist/Tk-TableMatrix/pod/TableMatrix.pod
      print ">>> Package \"Tk::TableMatrix\" not installed!\n" .
            ">>> Waypoint List Editor not available!\n\n";

eval "use Tk::TableMatrix::Spreadsheet"  ||   # http://search.cpan.org/dist/Tk-TableMatrix/TableMatrix/Spreadsheet.pm
      print ">>> Package \"Tk::TableMatrix::Spreadsheet\" not installed!\n" .
            ">>> Waypoint List Editor not available!\n\n";

#
# Configuration and data-visualisation
#

# Display or Modify Hash
sub DisplayHash()
    {
    my ($hrefData, $Titel, $Mode) = @_;

    # $Mode: Display, Edit, Waypoint, PointOI, Refresh, Heartbeat, SerialChannel, ExternControl

    my $MaxRow = 22;   # number or Rows in multi column view
    my %Id;
    my $Label;
    my $Value;

    # Neues Fenster aufmachen
    my $popup = $main->Toplevel();
    $popup->title($Titel);
	
    # Frame mit den Buttons
    my $popup_button = $popup->Frame() -> pack('-side' => 'bottom',
                                               '-expand' => 'y',
                                               '-anchor' => 's',
                                               '-padx' => 5,
                                               '-pady' => 5,
                                               );

    $popup_button->Button('-text'    => 'Schlieen',
                          '-command' => sub
        {
        if ( $Mode =~ /edit/i  and  $Mode =~ /waypoint/i )
            {
            $WaypointsModified = 1;            
            &WpRedrawLines();
            &WpRedrawIcons();
            }

        $popup->destroy()
        })->pack;

    # Frame mit den Labels und Daten
    my $popup_data = $popup->Frame() -> pack('-side' => 'left',
                                             '-expand' => 'y',
                                             '-anchor' => 'w',
                                             '-padx' => 10,
                                             '-pady' => 10,
                                             );

    # Labels und Daten anzeigen
    my $Row = 0;
    my $Col = 0;

    foreach $Label ( sort keys %{$hrefData})
        {
        $LabelView = $Label;
        if ( $Translate{$LabelView} ne "" )
            {
            $LabelView = $Translate{$LabelView};
            }

        # Label
        $popup_data->Label ('-text' => $LabelView,
                            '-width' => 25,
                            '-anchor' => 'w',
                            ) -> grid( -row => $Row,
                                       -column => $Col,
                                       -padx => 10,
                                     );
        # Daten
        if ( $Mode =~ /display/i )
            {
            # Display
            if ( ref ${$hrefData}{$Label} )
                {
                $Text = "- can't display references -";
                }
            else
                {
                $Text = ${$hrefData}{$Label};
                }

            $Id{$Label} = $popup_data->Label ('-text' => $Text,
                                              '-width' => 20,
                                              '-anchor' => 'e',
                                              '-relief' => 'sunken',
                                             ) -> grid( -row => $Row,
                                                        -column => $Col + 1,
                                                        -padx => 10,
                                                      );
            }
        if ( $Mode =~ /edit/i )
            {
            # Edit
            $Id{$Label} = $popup_data->Entry ('-textvariable' => \${$hrefData}{$Label},
                                              '-exportselection' => '1',
                                              '-width' => 20,
                                              '-relief' => 'sunken',
                                             ) -> grid( -row => $Row,
                                                        -column => $Col + 1,
                                                        -padx => 10,
                                                      );
            if ( $Mode =~ /waypoint/i )
                {
                # einige Waypoint-Felder nicht aenderbar einstellen
                if ( "MapX MapY _MapX_Rel _MapY_Rel Pos_Lat Pos_Lon Tag" =~ /$Label/i )
                    {
                    $Id{$Label}->configure('-state' => 'disabled', );
                    }
                }

            if ( $Mode =~ /PointOI/i )
                {
                # einige POI-Felder nicht aenderbar einstellen
                if ( "Poi_x Poi_y Poi_Lat Poi_Lon" =~ /$Label/i )
                    {
                    $Id{$Label}->configure('-state' => 'disabled', );
                    }
                }

            }

        # multi Column wrap
        $Row++;
        if ( $Row > $MaxRow )
            {
            $Row = 0;
            $Col += 2;
            }
        }	


    if ( $Mode =~ /refresh/i )
        {
        # Timer: 0.1s
        $popup_data->repeat (200, sub 
            {
            # Datenfelder alle 100ms aktualisieren

            my $BgColor = 'white';
            if ( $Mode =~ /heartbeat/i )
                {
                $BgColor = 'red';
                if ( &MkOsdIsValid() )
    	            {
                    # gltige daten vom MK
                    $BgColor = 'white';
                    }
                }

            if ( $Mode =~ /serialchannel/i )
                {
                $BgColor = 'red';
                if ( $Cfg->{'serialchannel'}->{'SerialChannelSend'} =~ /y/i )
    	            {
                    # senden aktiv
                    $BgColor = 'white';
                    }
                }

            if ( $Mode =~ /externcontrol/i )
                {
                $BgColor = 'red';
                if ( $Cfg->{'externcontrol'}->{'ExternControlSend'} =~ /y/i )
    	            {
                    # senden aktiv
                    $BgColor = 'white';
                    }
                }
                
            foreach $Label ( sort keys %{$hrefData} )
                {
                # Eingebbare Waypoint/POI-Felder nicht aktualisieren
                if ( ! ( ($Mode =~ /waypoint/i  and
                         "Event_Flag EventChannelValue Heading ToleranceRadius HoldTime Pos_Alt Speed AltRate CamAngle" =~ /$Label/i)
                         or
                         ($Mode =~ /PointOI/i  and
                          "Poi_Alt Poi_EventFlag" =~ /$Label/i)
                        ))
                    {                
                    $Id{$Label}->configure('-text' => ${$hrefData}{$Label},
                                           '-background' => "$BgColor",
                                          );
                    }
                }
            });
        }

    return 0;
    }


# Konfigurations-Hash (aus XML-Datei) im Popup-Fenster editieren
sub Configure()
    {
    my ($CfgFile, $hrefCfg, $Mode) = @_;

    # get a copy of Cfg-Hash for editing
    my $CfgEdit = &CopyHash($hrefCfg);

    # Neues Fenster aufmachen
    my $popup = $main->Toplevel();
    $popup->title("Einstellungen - $CfgFile");

    # Display data in a notebook widget
    my $book = $popup->NoteBook(-dynamicgeometry => 1,
                               )->grid(-row        => 0,
                                       -column     => 0,
                                       -columnspan => 4,
                                       -sticky     => 'w',
                                       -padx       => 5,
                                       -pady       => 5,
                                      );
    # Show data
    &ConfigureShow($book, $CfgEdit);

    # Button: OK
    $popup->Button('-text'    => 'OK',
                   '-width' => '10',
                   '-command' => sub
        {
        # Save and activate config
        &ConfigureSave( $CfgFile, $hrefCfg, $CfgEdit);
        &CanvasRedraw();

        $popup->destroy();

        } )->grid(-row    => 1,
                  -column => 0,
                  -sticky => 'w',
                  -padx   => 15,
                  -pady   => 5,
                 );

    # Button: Apply
    $popup->Button('-text'    => $Translate{'Apply'},
                   '-width' => '10',
                   '-command' => sub
        {
        # Save and activate config
        &ConfigureSave( $CfgFile, $hrefCfg, $CfgEdit);
        &CanvasRedraw();

        } )->grid(-row    => 1,
                  -column => 1,
                  -sticky => 'w',
                  -padx   => 15,
                  -pady   => 5,
                 );


    # Button: Abort
    $popup->Button('-text'    => $Translate{'Abort'},
                   '-width' => '10',
                   '-command' => sub { $popup->destroy() },
                  )->grid(-row    => 1,
                          -column => 2,
                          -sticky => 'w',
                          -padx   => 15,
                          -pady   => 5,
                         );

    #
    # special handling for "Config" configuration
    #
    if ( $Mode =~ /CONFIG/i )
        {
        $popup->Label ('-text' => $Translate{'RestartRequired'},
                       '-anchor' => 'w',
                       '-foreground' => 'red',
                      )->grid(-row    => 1,
                              -column => 3,
                              -sticky => 'w',
                              -padx   => 5,
                              -pady   => 5,
                             );
        }


    #
    # special handling for "Event" configuration
    #
    if ( $Mode =~ /EVENT/i )
        {

        # notebook must have at least one tab
        if (scalar $book->pages() == 0 )
            {
            # create new record in hash
            my $NewEvent = sprintf ("Event%d", scalar $book->pages() + 1);
            &EventInit($NewEvent, $CfgEdit);

            # Display new event
            &ConfigureShow($book, $CfgEdit);
            $book->raise($NewEvent);
            }


        # Menu bar (New, Delete, Rename)

        my $menu_bar = $popup->Menu;
        $popup->optionAdd("*tearOff", "false");
        $popup->configure ('-menu' => $menu_bar);

        my $menu_event = $menu_bar->cascade('-label' => $Translate{'Event'});

        #
        # New Event
        #
        $menu_event->command('-label' => $Translate{'EventNew'},
                             '-command' => sub
            {
            # Event Name in neuem Fenster abfragen
            my $popup_new = $popup->Toplevel();
            $popup_new->title("Event - $Translate{'EventNew'}");

            $popup_new->Label (-text => $Translate{'EventNewName'},
                               -width => 20,
                               -anchor => 'w',
                               )->grid (-row    => 0,
                                        -column => 0,
                                        -sticky => 'w',
                                        -padx   => 5,
                                        -pady   => 5,
                                       );

            my $NewEvent = sprintf ("Event%d", scalar $book->pages() + 1);
            $popup_new->Entry ( -textvariable => \$NewEvent,
                                -exportselection => '1',
                                -width => 40,
                                -relief => 'sunken',
                               )->grid (-row    => 0,
                                        -column => 1,
                                        -sticky => 'w',
                                        -padx   => 5,
                                        -pady   => 5,
                                        );
            # Button: OK
            $popup_new->Button('-text'    => "OK",
                               '-width' => '10',
                               '-command' => sub
                {
                # create new record in hash
                $NewEvent = &EventnameAdjust($NewEvent);
                &EventInit($NewEvent, $CfgEdit);

                # Display new event
                &ConfigureShow($book, $CfgEdit);
                $book->raise($NewEvent);

                $popup_new->destroy();
                } )->grid (-row    => 1,
                           -column => 0,
                           -sticky => 'w',
                           -padx   => 20,
                           -pady   => 5,
                          );

            # Button: Abort
            $popup_new->Button('-text'    => $Translate{'Abort'},
                               '-width' => '10',
                               '-command' => sub
                {
                $popup_new->destroy()
                } )->grid (-row    => 1,
                           -column => 1,
                           -sticky => 'e',
                           -padx   => 20,
                           -pady   => 5,
                          );
            });

        #
        # Rename Event
        #
        $menu_event->command('-label' => $Translate{'EventRename'},
                             '-command' => sub
            {
            # Event Name in neuem Fenster abfragen
            my $popup_rename = $popup->Toplevel();
            $popup_rename->title("Event - $Translate{'EventRename'}");

            $popup_rename->Label (-text => $Translate{'EventName'},
                                  -width => 20,
                                  -anchor => 'w',
                                  )->grid (-row    => 0,
                                           -column => 0,
                                           -sticky => 'e',
                                           -padx   => 5,
                                           -pady   => 5,
                                          );

            my $CurrentEvent = $book->raised;
            $popup_rename->Entry ( -textvariable => \$CurrentEvent,
                                   -exportselection => '1',
                                   -width => 40,
                                   -relief => 'sunken',
                                   -state => 'disabled',
                                  )->grid (-row    => 0,
                                           -column => 1,
                                           -sticky => 'w',
                                           -padx   => 5,
                                           -pady   => 5,
                                           );

            $popup_rename->Label (-text => $Translate{'EventNewName'},
                                  -width => 20,
                                  -anchor => 'w',
                                  )->grid (-row    => 1,
                                           -column => 0,
                                           -sticky => 'e',
                                           -padx   => 5,
                                           -pady   => 5,
                                          );

            my $NewEvent = sprintf ("Event%d", scalar $book->pages() + 1);
            $popup_rename->Entry ( -textvariable => \$NewEvent,
                                   -exportselection => '1',
                                   -width => 40,
                                   -relief => 'sunken',
                                  )->grid (-row    => 1,
                                           -column => 1,
                                           -sticky => 'w',
                                           -padx   => 5,
                                           -pady   => 5,
                                           );

            # Button: OK
            $popup_rename->Button('-text'    => "OK",
                                  '-width' => '10',
                                  '-command' => sub
                {
                $NewEvent = &EventnameAdjust($NewEvent);

                # create new record in hash
                $CfgEdit->{$NewEvent} = $CfgEdit->{$CurrentEvent};
                delete $CfgEdit->{$CurrentEvent}

                # Display events again
                &ConfigureShow($book, $CfgEdit);

                $book->raise($NewEvent);

                $popup_rename->destroy();
                } )->grid (-row    => 2,
                           -column => 0,
                           -sticky => 'e',
                           -padx   => 20,
                           -pady   => 5,
                          );

            # Button: Abort
            $popup_rename->Button('-text'    => $Translate{'Abort'},
                                  '-width' => '10',
                                  '-command' => sub
                {
                $popup_rename->destroy()
                } )->grid (-row    => 2,
                           -column => 1,
                           -sticky => 'w',
                           -padx   => 20,
                           -pady   => 5,
                          );

            });

        #
        # Copy Event
        #
        $menu_event->command('-label' => $Translate{'EventCopy'},
                             '-command' => sub
            {
            # Event Name in neuem Fenster abfragen
            my $popup_copy = $popup->Toplevel();
            $popup_copy->title("Event - $Translate{'EventCopy'}");

            my $CurrentEvent = $book->raised;
            my $CopyEvent = sprintf ("Event%d", scalar $book->pages() + 1);

            $popup_copy->Label (-text => $Translate{'EventName'},
                                -width => 20,
                                -anchor => 'w',
                                )->grid (-row    => 0,
                                         -column => 0,
                                         -sticky => 'e',
                                         -padx   => 5,
                                         -pady   => 5,
                                        );

            $popup_copy->Entry ( -textvariable => \$CurrentEvent,
                                 -exportselection => '1',
                                 -width => 40,
                                 -relief => 'sunken',
                                 -state => 'disabled',
                                )->grid (-row    => 0,
                                         -column => 1,
                                         -sticky => 'w',
                                         -padx   => 5,
                                         -pady   => 5,
                                         );

            $popup_copy->Label (-text => $Translate{'EventNewName'},
                                -width => 20,
                                -anchor => 'w',
                                )->grid (-row    => 1,
                                         -column => 0,
                                         -sticky => 'w',
                                         -padx   => 5,
                                         -pady   => 5,
                                        );


            $popup_copy->Entry ( -textvariable => \$CopyEvent,
                                 -exportselection => '1',
                                 -width => 40,
                                 -relief => 'sunken',
                                )->grid (-row    => 1,
                                         -column => 1,
                                         -sticky => 'w',
                                         -padx   => 5,
                                         -pady   => 5,
                                         );
            # Button: OK
            $popup_copy->Button('-text'    => "OK",
                                '-width' => '10',
                                '-command' => sub
                {
                $CopyEvent = &EventnameAdjust($CopyEvent);

                # copy hash
                $CfgEdit->{$CopyEvent} = {%{$hrefCfg->{$CurrentEvent}}};

                # Display new event
                &ConfigureShow($book, $CfgEdit);
                $book->raise($CopyEvent);

                $popup_copy->destroy();
                } )->grid (-row    => 2,
                           -column => 0,
                           -sticky => 'w',
                           -padx   => 20,
                           -pady   => 5,
                          );

            # Button: Abort
            $popup_copy->Button('-text'    => $Translate{'Abort'},
                               '-width' => '10',
                               '-command' => sub
                {
                $popup_copy->destroy()
                } )->grid (-row    => 2,
                           -column => 1,
                           -sticky => 'e',
                           -padx   => 20,
                           -pady   => 5,
                          );
            });

        #
        # Delete event
        #
        $menu_event->command('-label' => $Translate{'EventDelete'},
                             '-command' => sub
            {
            my $CurrentBook = $book->raised;

            # delet event in Cfg-Hash
            delete $CfgEdit->{$CurrentBook};

            # Display events again
            &ConfigureShow($book, $CfgEdit);
            });

        #
        # Export current Event
        #
        $menu_event->command('-label' => $Translate{'EventExport'},
                             '-command' => sub
            {
            my $XmlFile = $popup->getSaveFile('-defaultextension' => ".xml",
                                              '-filetypes'        =>
                                               [['Event',     '.xml' ],
                                                ['All Files', '*', ],
                                                ],
                                              '-initialdir' => "event",
                                              '-title' => $Translate{'EventExport'},
                                             );
            if ( $XmlFile ne "" )
                {
                my $ExportCfg;
                my $CurrentEvent = $book->raised;

                # copy and quote event
                foreach $key (keys %{$CfgEdit->{$CurrentEvent}})
                    {
                    my $Line = $CfgEdit->{$CurrentEvent}->{$key};
                    $ExportCfg->{$key} = &QuoteXML($Line);
                    }

            # Event in XML-Datei speichern
            &XMLout ($ExportCfg,                 # save quoted hash
                     'OutputFile' => $XmlFile,
                     'AttrIndent' => '1',
                     'RootName'   => 'mkcockpit-Event',
                     'NoEscape'   => '1',
                    );
                }

            });

        #
        # Import XML to current Event
        #
        $menu_event->command('-label' => $Translate{'EventImport'},
                             '-command' => sub
            {
            my $XmlFile = $popup->getOpenFile(-defaultextension => ".xml",
                                              -filetypes        =>
                                               [['Event',     '.xml' ],
                                                ['All Files', '*', ],
                                                ],
                                              -initialdir => "event",
                                              -title => $Translate{'EventImport'},
                                             );
            if ( -f $XmlFile )
                {
                my $CurrentEvent = $book->raised;
                my $ImportCfg = XMLin($XmlFile);

                # copy event
                foreach $key (keys %{$ImportCfg})
                    {
                    $CfgEdit->{$CurrentEvent}->{$key} = $ImportCfg->{$key};
                    }

                # deactivate Event
                $CfgEdit->{$CurrentEvent}->{'Active'} = "NO";
                }
            });
        }
    }


# Copy a Cfg-Hash including real copy of hash-references
sub CopyHash()
    {
    my ($hrefCfg) = @_;

    my $CfgCopy = {%{$hrefCfg}};
    foreach $key (keys %{$hrefCfg})
        {
        if ( ref $hrefCfg->{$key} )
            {
            $CfgCopy->{$key} = {%{$hrefCfg->{$key}}};
            }
        }
    return $CfgCopy;
    }


# Initialize a new event
sub EventInit()
    {
    my ($EventName, $Cfg) = @_;

    $Cfg->{$EventName}->{'Active'}      = "no";
    $Cfg->{$EventName}->{'Action'}      = "";
    $Cfg->{$EventName}->{'ActionElse'}  = "";
    $Cfg->{$EventName}->{'Condition'}   = "";
    $Cfg->{$EventName}->{'Delay'}       = "";
    $Cfg->{$EventName}->{'Repeat'}      = "";
    $Cfg->{$EventName}->{'RepeatElse'}  = "";
    $Cfg->{$EventName}->{'Description'} = "";
    $Cfg->{$EventName}->{'Trigger'}     = "TRUE";
    }


# Event-Name XML konform anpassen
sub EventnameAdjust()
    {
    my ($Name) = @_;

    $Name =~ s/\W/_/g;
    if ( substr ($Name, 0, 1) =~ /\d/ )
        {
        substr ($Name, 0, 1) = "_";
        }

    return $Name;
    }



# display tabs with configuration data
sub ConfigureShow()
    {
    my ($book, $CfgEdit) = @_;

    # delete all existing tabs in notebook
    foreach my $Tab ($book->pages)
        {
        $book->delete($Tab);
        }

    # jede Sektion in eigenem Tab anzeigen
    foreach my $key (sort keys %{$CfgEdit})
        {    
        if ( ! ref $CfgEdit->{$key} )
            {
            # skip: is not a section
            next;
            }

        if ( keys %{$CfgEdit->{$key}} == 0 )
            {
            # skip: section has no parameters
            next;
            }

        my $TabLabel = "$key";
        if ( $Translate{$key} ne "" )
                {
                $TabLabel = $Translate{$key};
                }

        my $Tab = $book->add( "$key",
                              -label      => "$TabLabel",
                              -wraplength => "75",
                              -createcmd => [\&CbConfigureShow, $CfgEdit, $key],
                             );

        # save tab reference for use in Callback "&CbConfigureShow()"
        $ConfigureShowTabs{$key} = $Tab;  
        }
    }



# Callback; create fields for a Tab
sub CbConfigureShow()
    {
    my ($CfgEdit, $key) = @_;

    # Frame for label and data

    my $Tab = $ConfigureShowTabs{$key};
    my $popup_cfg = $Tab->Frame() -> pack('-anchor' => 'w',
                                          '-padx' => 5,
                                          '-pady' => 5,
                                          );
 
    # Eingabefelder/Optionmenu/Fileselection mit Daten anzeigen
    my $Col = 0;
    my $Row = 0;
    my $MaxRow = 20;   # number or Rows in multi column view
    foreach $Entry ( sort keys %{$CfgEdit->{$key}})
        {				
        # Label
        my $Label = $Entry;
        if ( $Translate{$Label} ne "" )
            {
            $Label = $Translate{$Label};
            }
		
        $popup_cfg->Label (-text => $Label,
                           -width => 30,
                           -anchor => 'w',
                           )->grid (-row    => $Row,
                                    -column => $Col,
                                   );
        #
        # Combo Box with optiones defined in libcfgopt.pl
        #
        if ( defined $CfgOpt{$Entry}[0] )
            {
            my $cbo = $popup_cfg->BrowseEntry( -label => "",
                                               -variable => \$CfgEdit->{$key}->{$Entry},
                                               -width => 37,
                                               -relief => 'sunken',
                                             )->grid (-row    => $Row,
                                                      -column => $Col + 1,
                                                      -columnspan => 2,
                                                      -sticky => 'w',
                                                     );
            # add options
            $cbo->insert("end", @{ $CfgOpt{$Entry} });
            }

        #
        # File selection, if defined in libcfgopt.pl
        #
        elsif ( defined $CfgFile{$Entry} )
            {
            # a) Text entry
            my $TextEntry = $popup_cfg->Entry ( -textvariable => \$CfgEdit->{$key}->{$Entry},
                                                -exportselection => '1',
                                                -width => 37,
                                                -relief => 'sunken',
                                              )->grid (-row    => $Row,
                                                       -column => $Col + 1,
                                                       -columnspan => 1,
                                                       -sticky => 'e',
                                                      );
            # b) button with file selection dialog
            my $Mode = $CfgFile{$Entry};
            $popup_cfg->Button('-text'  => '>',
                               '-width' => 1,
                               '-command' => sub
                {
                my $File = $popup_cfg->getOpenFile( '-title' => $Label );
                if ($File ne "" )
                    {
                    if ( $Mode =~ /Filename/i )
                        {
                        $File = substr ($File, rindex ($File, '/') +1 );
                        }
                    elsif ( $Mode =~ /Path/i )
                        {
                        # nothing to do
                        }
                    # show selection in text entry
                    $TextEntry->delete (0, length $TextEntry->get );
                    $TextEntry->insert (0, $File);
                    }
                } )->grid (-row    => $Row,
                           -column => $Col + 2,
                           -sticky => 'e',
                          );
            }

        #
        # Multiline Text widget, if defined in libcfgopt.pl
        #
        elsif ( defined $CfgText{$Entry} )
            {
            # a) Text entry

            my $State = 'normal';
            my $NumLines = grep /\n/, $CfgEdit->{$key}->{$Entry};
            if ( $NumLines > 0 )
                {
                $State = 'disabled';
                }
            my $TextEntry = $popup_cfg->Entry ( -textvariable => \$CfgEdit->{$key}->{$Entry},
                                                -exportselection => '1',
                                                -state => $State,
                                                -width => 37,
                                                -relief => 'sunken',
                                              )->grid (-row    => $Row,
                                                       -column => $Col + 1,
                                                       -columnspan => 1,
                                                       -sticky => 'e',
                                                      );

            # b) button with multiline Text-Edit dialog
            my ($Width, $Height) = split /;/, $CfgText{$Entry};
            my $Title = "Edit: $key -> $Entry";
            my $refVariable = \$CfgEdit->{$key}->{$Entry};
            $popup_cfg->Button('-text'  => '>',
                               '-width' => 1,
                               '-command' => sub
                {
                # popup mit Text-Widget
                my $popup_text = $popup_cfg->Toplevel();
                $popup_text->title($Title);
                my $text = $popup_text->Scrolled( "Text",
                                                   -height => $Height,
                                                   -width  => $Width,
                                                   -wrap   => 'none',
                                                   -scrollbars => 'se',
                                                 )->pack();

                $text->insert('end', $$refVariable);

                # OK Button
                $popup_text->Button('-text'    => 'OK',
                                    '-width'   => 10,
                                    '-command' => sub
                    {
                    # Text uebernehmen
                    $$refVariable = $text->get("1.0", "end");
                    chomp $$refVariable;
                    $popup_text->destroy()
                    } )->pack ('-side' => 'left',
                               '-expand' => 'y',
                               '-anchor' => 's',
                               '-padx' => 5,
                               '-pady' => 5,
                              );

                # Anwenden Button
                $popup_text->Button('-text'    => $Translate{'Apply'},
                                    '-width'   => 10,
                                    '-command' => sub
                    {
                    # Text uebernehmen
                    $$refVariable = $text->get("1.0", "end");
                    chomp $$refVariable;

                    } )->pack ('-side' => 'left',
                               '-expand' => 'y',
                               '-anchor' => 's',
                               '-padx' => 5,
                               '-pady' => 5,
                              );

                # Abort Button
                $popup_text->Button('-text'    => $Translate{'Abort'},
                                    '-width'   => '10',
                                    '-command' => sub
                    {
                    $popup_text->destroy();

                    } )->pack ('-side' => 'left',
                               '-expand' => 'y',
                               '-anchor' => 's',
                               '-padx' => 5,
                               '-pady' => 5,
                               );
                } )->grid (-row    => $Row,
                           -column => 2,
                           -sticky => 'e',
                          );
            }

        #
        # Text entry
        #
        else
            {
            $popup_cfg->Entry ( -textvariable => \$CfgEdit->{$key}->{$Entry},
                                -exportselection => '1',
                                -width => 40,
                                -relief => 'sunken',
                               )->grid (-row    => $Row,
                                        -column => $Col + 1,
                                        -columnspan => 2,
                                        -sticky => 'e',
                                       );
            }

        # next Variable in next row (Multi-Column)
        $Row ++;
        if ( $Row > $MaxRow )
            {
            $Row = 0;
            $Col += 3;
            }
        }
    }


# Quote for output to XML-file
sub QuoteXML()
    {
    my ($Line) = @_;

    $Line =~ s/\&/\&amp;/g;
    $Line =~ s/'/\&apos;/g;
    $Line =~ s/</\&lt;/g;
    $Line =~ s/>/\&gt;/g;
    $Line =~ s/"/\&quot;/g;
    $Line =~ s/\n/\&#10;/g;

    return $Line;
    }


# Activate and save Config to file
sub ConfigureSave()
    {
    my ($CfgFile, $hrefCfg, $CfgEdit) = @_;

    # set new timestamp
    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
    my $TimeStamp = sprintf ("%04d%02d%02d-%02d%02d%02d", $year+1900, $mon+1, $mday, $hour, $min, $sec);
    $CfgEdit->{'CreationDate'} = $TimeStamp;

    # set MkCockpit Version
    $CfgEdit->{'Version'} =  $Version{'mkcockpit.pl'};

    # empty original hash
    foreach $key (keys %{$hrefCfg})
        {
        delete $hrefCfg->{$key};
        }

    # Build %CfgXml with XML-quoting
    my $CfgXml = {};

    # then copy %CfgEdit back to $hrefCfg.
    foreach $key (keys %{$CfgEdit})
        {
        if ( ref $CfgEdit->{$key} )
            {
            # Reference
            foreach $val (keys %{$CfgEdit->{$key}})
                {
                my $Line = $CfgEdit->{$key}->{$val};
                $hrefCfg->{$key}->{$val} = $Line;
                $CfgXml->{$key}->{$val} = &QuoteXML($Line);
                }
            }
        else
            {
            # Scalar
            my $Line = $CfgEdit->{$key};
            $hrefCfg->{$key} = $Line;
            $CfgXml->{$key} = &QuoteXML($Line);
            }
        }

    # Cfg in XML-Datei speichern
    &XMLout ($CfgXml,                 # save quoted hash
             'OutputFile' => $CfgFile,
             'AttrIndent' => '1',
             'RootName'   => 'mkcockpit-Config',
             'NoEscape'   => '1',
            );
    }



# Display status text in upper status line
sub DispayUpperStatus()
    {
    my ($Num, $Condition, $Text, $ActiveColor) = @_;

    my $Color = 'lightgray';
    my $Relief = 'flat';
    if ( $Condition )
            {
            $Color = $ActiveColor;
            $Relief = 'sunken';
            }
    $map_status_top[$Num]->configure (-text       => $Text,
                                      -background => $Color,
                                      -relief     => $Relief,
                                     );
    }


#
# Waypoint List Editor
#

# Field definition for Wp List Editor
@WpEditTitle = ( "WP Nr", 
                 "Heading",
                 "Altitude",
                 "AltRate",
                 "Speed",
                 "Wait",
                 "Tolerance",
                 "CAM-Nick",
                 "Ev.Ch",
                 "Ev.Flag",
               );

@WpEditFields = ( "_Title_",
                  "Heading", 
                  "Pos_Alt",
                  "AltRate",
                  "Speed",
                  "Holdtime",
                  "ToleranceRadius",
                  "CamAngle",
                  "EventChannelValue",
                  "Event_Flag",
                );

# WP list editor
sub WpEdit()
    {
    # Neues Fenster aufmachen
    my $popup = $main->Toplevel();
    $popup->title("Waypoint Editor");

    my $Rows = 100 + 1;
    my $Cols = $#WpEditTitle + 1;

    $WpEditTable = $popup->Scrolled("Spreadsheet",
                                 -font            => '-*-Arial-Normal-R-Normal--*-120-*',
                                 -rows            => $Rows,
                                 -cols            => $Cols,
                                 -titlecols       => 1,
                                 -titlerows       => 1,
                                 -resizeborders   =>'col',
                                 -colstretchmode  =>'all',
                                 -colwidth        => 6,
                                 -cache           => 0,
                                 -drawmode        => 'slow',
                                 -scrollbars      => "osoe",
                                 -multiline       => '0',
                                 -selectmode      => 'extended',
                                 -anchor          => 'w',
                                 -maxheight       => '400',
                                 -maxwidth        => '250',
                                 -usecommand      => 1,
                                 -command         => \&CbWpEditCommand,
                                 -validate        => 1,
                                 -validatecommand => \&CbWpEditValidate,
                                ) -> pack(-expand => 1,
                                          -fill   => 'both');

    # Catch delete window event
    $popup->protocol( 'WM_DELETE_WINDOW' => sub
        {
        $popup->destroy();
        undef $WpEditTable;
        });

    return 0;
    }


# Refresh WP Edit Spreadsheet
sub WpEditRefresh()
    {
    if ( defined $WpEditTable )
        {
        # trick for refresh, see Tk::TableMatrix Example
        $WpEditTable->configure (-padx => $WpEditTable->cget(-padx));
        }
    }


# Get and set WP Values from @Waypoints for Spreadsheet
sub CbWpEditCommand()
    {
    my ($SetFlag, $Row, $Col, $Value) = @_;

    if ( $SetFlag == 1 )
        {
        #
        # write field
        #
        if ( $Row == 1 )
            {
            # Set Value for all WP
            my $WpCount = scalar @Waypoints;
            for ( my $Wp=1; $Wp <= $WpCount; $Wp ++)
                {
                # WP data
                my $Field = $WpEditFields[$Col];
                $Waypoints[$Wp-1]{$Field} = eval "$Value";    # $Wp may be used in eval as WP counter
                }

            &WpEditRefresh();
            }
        elsif ( $Row >= 2  and  $Row <= $#Waypoints + 2 and  $Col >= 1)
            {
            # WP data
            my $Field = $WpEditFields[$Col];
            my $WpIndex = $Row - 2;
            $Waypoints[$WpIndex]{$Field} = $Value;
            }
        }
    else
        {
        #
        # read field
        #
        if ( $Row == 0 )
            {
            # title
            return "$WpEditTitle[$Col]";
            }
        elsif ( $Row == 1 )
            {
            # "Edit-All" Row
            if ( $Col == 0 )
                {
                return ("Set All");
                }
            return "---";
            }
        elsif ( $Row <= $#Waypoints + 2 )
            {
            # WP data
            if ( $Col == 0 )
                {
                return ($Row - 1);
                }
            my $Field = $WpEditFields[$Col];
            my $WpIndex = $Row - 2;
            my $Val = $Waypoints[$WpIndex]{$Field};
            return $Val;
            }
        }

    return "";
    }


# Validate input
sub CbWpEditValidate()
    {
    my ($Row, $Col, $OldValue, $NewValue, $CellIndex) = @_;

    if ( $Row >= 2 )
        {
        # check only for WP-Rows

        $NewValue =~ s/[0-9.-]//g;
        if ( $NewValue ne "" )
            {
            return 0;  # NOK
            }
        }
    return 1;  # OK
    }

1;

__END__
