#!/usr/bin/perl
#!/usr/bin/perl -d:ptkdb

###############################################################################
#
# libmkcockpit.pl -  MK Mission Cockpit - Subroutines
#
# Copyright (C) 2009  Rainer Walther  (rainerwalther-mail@web.de)
#
# Creative Commons Lizenz mit den Zusaetzen (by, nc, sa)
#
# Es ist Ihnen gestattet: 
#     * das Werk vervielfltigen, verbreiten und ffentlich zugnglich machen
#     * Abwandlungen bzw. Bearbeitungen des Inhaltes anfertigen
# 
# Zu den folgenden Bedingungen:
#     * Namensnennung.
#       Sie mssen den Namen des Autors/Rechteinhabers in der von ihm festgelegten Weise nennen.
#     * Keine kommerzielle Nutzung.
#       Dieses Werk darf nicht fr kommerzielle Zwecke verwendet werden.
#     * Weitergabe unter gleichen Bedingungen.
#       Wenn Sie den lizenzierten Inhalt bearbeiten oder in anderer Weise umgestalten,
#       verndern oder als Grundlage fr einen anderen Inhalt verwenden,
#       drfen Sie den neu entstandenen Inhalt nur unter Verwendung von Lizenzbedingungen
#       weitergeben, die mit denen dieses Lizenzvertrages identisch oder vergleichbar sind.
# 
# Im Falle einer Verbreitung mssen Sie anderen die Lizenzbedingungen, unter welche dieses
# Werk fllt, mitteilen. Am Einfachsten ist es, einen Link auf diese Seite einzubinden.
# 
# Jede der vorgenannten Bedingungen kann aufgehoben werden, sofern Sie die Einwilligung
# des Rechteinhabers dazu erhalten.
# 
# Diese Lizenz lsst die Urheberpersnlichkeitsrechte unberhrt.
# 
# Weitere Details zur Lizenzbestimmung gibt es hier:
#   Kurzform: http://creativecommons.org/licenses/by-nc-sa/3.0/de/
#   Komplett: http://creativecommons.org/licenses/by-nc-sa/3.0/de/legalcode
#
###############################################################################
# 2009-08-09 0.2.5 rw subroutines moved from mkcockpit.pl
# 2009-09-05 0.2.6 rw POI heading control added
# 2009-10-10 0.2.7 rw Layout Config-dialog
#                     Fix Message-Balloon in KML-Mode
# 2009-10-25 0.3.0 rw NC 0.17
#                     Read/Write KopterTool WPL Waypoint list
#                     configuration Combo Box
# 2010-02-10 0.4.0 rw Show Grid on map
#                     Show crosshair in player pause mode
#                     joystick and 3D-Mouse support
#                     Cfg file selection dialog
#                     serial channel
#                     Event engine
#                     External control - Limit, expo
#                     Resize WP-Icon to 24x48 pixel
# 2010-02-14 0.4.1 rw ExpoLimit
# 2010-03-07 0.4.2 rw Input control parser added
#                     FctKey, RcStick, RcPoti input device
#                     MkIsLowBat     
# 2010-09-10 0.5.2 rw Redraw canvas after configuration change 
# 2010-10-16 0.5.4 rw Spline speed controlled mode
#                     FctKeyToggle() added
# 2010-11-01 0.6.0 rw Minor change SPD-Player
#                     NC 0.21
# 2011-03-06 0.7.0 rw WP EventChannelValue added
#                     NC bassed Altitude- and Gier-control
#                     WP List Editor added
#                     MKFlags2 check routines
#                     convert wP to Border
# 2011-04-02 0.7.1 rw WP EventChannelValue added
#                     Logging added
# 2011-09-22 0.8.0 rw disable gier when POI switched off
# 2011-12-24 0.9.0 rw NC 0.28
#                     MkIsOut1(), MkIsOut2()
#
###############################################################################

$Version{'libmkcockpit.pl'} = "0.9.0 - 2011-12-24";

# return default value, if value is empty
sub Default()
    {
    my ($Value, $DefaultValue1, $DefaultValue2) = @_;

    if ( $Value eq "" ) { $Value = $DefaultValue1; }
    if ( $Value eq "" ) { $Value = $DefaultValue2; }

    return $Value;
    }


# check, if %MkOsd is valid
sub MkOsdIsValid()
    {
    return ( $MkOsd{'_Timestamp'} >= time-2 );
    }

# check, if current GPS position is valid
sub CurPosIsValid()
    {
    return ( &MkOsdIsValid()  and  &MkGpsOk()  and  $MkOsd{'CurPos_Stat'} == 1 );
    } 

# check, if home GPS position is valid
sub HomePosIsValid()
    {
    return ( &MkOsdIsValid()  and  &MkGpsOk()  and  $MkOsd{'HomePos_Stat'} == 1 );
    }

# check, if target GPS position is valid
sub TargetIsValid()
    {
    return ( &MkOsdIsValid()  and  &MkGpsOk()  and  $MkOsd{'TargetPos_Stat'} == 1  );
    }

# check, if motor are on
sub MkIsMotorOn()
    {
    return ( &MkOsdIsValid()  and  $MkOsd{'MKFlags'} & 0x01  );
    }

# check, if MK is flying
sub MkIsFlying()
    {
    return ( &MkOsdIsValid()  and  $MkOsd{'MKFlags'} & 0x02  );
    }

# check, if MK is calibrating
sub MkIsCalibrating()
    {
    return ( &MkOsdIsValid()  and  $MkOsd{'MKFlags'} & 0x04  );
    }

# check, if Motor is starting
sub MkIsMotorStarting()
    {
    return ( &MkOsdIsValid()  and  $MkOsd{'MKFlags'} & 0x08  );
    }

# check, Emergency Landing
sub MkEmergencyLanding()
    {
    return ( &MkOsdIsValid()  and  $MkOsd{'MKFlags'} & 0x10  );
    }
# check, if Battery Low voltage
sub MkIsLowVoltage()
    {
    return ( &MkOsdIsValid()  and  $MkOsd{'MKFlags'} & 0x20  );
    }

# check, vario trim up
sub MkIsTrimUp()
    {
    return ( &MkOsdIsValid()  and  $MkOsd{'MKFlags'} & 0x40  );
    }

# check, vario trim down
sub MkIsTrimDown()
    {
    return ( &MkOsdIsValid()  and  $MkOsd{'MKFlags'} & 0x80  );
    }

# check, carefree
sub MkIsCarefree()
    {
    return ( &MkOsdIsValid()  and  $MkOsd{'MKFlags2'} & 0x01  );
    }

# check, altitude control
sub MkIsAltCtrl()
    {
    return ( &MkOsdIsValid()  and  $MkOsd{'MKFlags2'} & 0x02  );
    }

# check, OUT1
sub MkIsOut1()
    {
    return ( &MkOsdIsValid()  and  $MkOsd{'MKFlags2'} & 0x08  );
    }

# check, OUT2
sub MkIsOut2()
    {
    return ( &MkOsdIsValid()  and  $MkOsd{'MKFlags2'} & 0x10  );
    }


# check, if MK is FREE Mode
sub MkIsFreeMode()
    {
    return ( &MkOsdIsValid()  and  $MkOsd{'NCFlags'} & 0x01  );
    }

# check, if MK is in PH Mode
sub MkIsPhMode()
    {
    return ( &MkOsdIsValid()  and  $MkOsd{'NCFlags'} & 0x02  );
    }

# check, if MK is in WPT Mode
sub MkIsWptMode()
    {
    return ( &MkOsdIsValid()  and  $MkOsd{'NCFlags'} & 0x04  );
    }

# check, Range Limit
sub MkRangeLimit()
    {
    return ( &MkOsdIsValid()  and  $MkOsd{'NCFlags'} & 0x08  );
    }

# check, Serial Link
sub MkSerialLink()
    {
    return ( &MkOsdIsValid()  and  $MkOsd{'NCFlags'} & 0x10  );
    }

# check, Target reached
sub MkTargetReached()
    {
    return ( &MkOsdIsValid()  and  $MkOsd{'NCFlags'} & 0x20  );
    }

# check, Manual Control
sub MkManualControl()
    {
    return ( &MkOsdIsValid()  and  $MkOsd{'NCFlags'} & 0x40  );
    }

# check, GPS OK
sub MkGpsOk()
    {
    return ( &MkOsdIsValid()  and  $MkOsd{'NCFlags'} & 0x80  );
    }


# Get altitude (hoehensensor)
sub AltitudeAir ()
    {
    return ( $MkOsd{'Altimeter'} / $Cfg->{'map'}->{'AltFactor'} );
    }

# Get altitude (GPS)
sub AltitudeGPS ()
    {
    return ( $MkOsd{'CurPos_Alt'} - $MkOsd{'HomePos_Alt'} );
    }

# Get altitude (average hoehensensor , GPS)
sub Altitude ()
    {
    my $Alt =  ( 4 * &AltitudeAir + &AltitudeGPS ) / 5;
    return ($Alt);
    }


# check, if Fct-Key "Num" pressed, Num = 1..12
sub FctKey()
    {
    my ($Num) = @_;

    $Num--;
    return (($Stick{'FctKey'} >> $Num) & 1) == 1;
    }

# check, if Fct-Key "Num" pressed (Toggle), Num = 1..12
sub FctKeyToggle()
    {
    my ($Num) = @_;

    $Num--;
    return (($Stick{'FctKeyToggle'} >> $Num) & 1) == 1;
    }

# Get RcPoti value. Poti = 1..8
sub RcPoti()
    {
    my ($Poti) = @_;

    return $Stick{"RcPoti" . "$Poti"};
    }


# Get Rc Stick value. Stick = Nick, Roll, Gas, Gier
sub RcStick()
    {
    my ($Stick) = @_;

    return $Stick{"RcStick" . "$Stick"};
    }


# range 0 .. 255
sub CheckUnsignedChar()
    {
    my ($U8) = @_;

    if ( $U8 < 0)   { $U8 = 0; };
    if ( $U8 > 255) { $U8 = 255; };
    return $U8;
    }

# range -128 .. 127
sub CheckSignedChar()
    {
    my ($S8) = @_;

    if ( $S8 < -128) { $S8 = -128; };
    if ( $S8 > 127)  { $S8 = 127; };
    return $S8;
    }

# Set serial Channel value. Num: 0..11, Val: -128..0..127
sub SerialChannel()
    {
    my ($Num, $Val) = @_;

    my $Key = sprintf ("SerialChannel%02d", $Num + 1);

    lock (%MkSerialChannel);    # until end of block

    $MkSerialChannel{$Key} = &CheckSignedChar($Val);  

    # timestamp, when channel value was set
    $MkSerialChannel{'_Timestamp'} = time;  
    }


# Limit: 0% .. 100%
# Expo : -100% .. 0 .. 100%
sub ExpoLimit ()
    {
    my ($StickMin, $StickMax, $Stick, $Expo, $LimitMin, $LimitMax) = @_;

    if ( $Expo ne "" )
        {
        # neg. Expo: 1..0.2  (0% .. -100%)
        # pos. Expo: 1..5    (0% ..  100%)

        if ( $Expo >= 0 )
            {
            $Expo = 1 + $Expo / 100 * 4;
            }
        else
            {
            $Expo = 1 + $Expo / 100 * 0.8;
            }

        if( $Stick >= 0 )
            {
            $Stick = $StickMax * ( $Stick ** $Expo ) / ( $StickMax ** $Expo);
            }
        else
            {
            $Stick = $StickMin * ( (- $Stick) ** $Expo ) / ( (- $StickMin) ** $Expo);
            }
        }

    # Travel limiter
    if ( $Stick >= 0  and  $LimitMax ne "" )
        {
        $Stick = $Stick * $LimitMax / 100;
        }
    elsif ( $Stick < 0  and  $LimitMin ne "" )
        {
        $Stick = $Stick * $LimitMin / 100;
        }

    return ($Stick);
    }


# Multipoint Curve
sub Curve ()
    {
    my ($Min, $Max, $Resolution, $Stick, @Curve) = @_;

    my $Points = scalar @Curve;
    if ( $Points < 2 )
        {
        # need at least 2 points
        return $Stick;
        }

    my $Val;
    if ( $Stick < $Min )
        {
        $Val= $Curve[0];
        }
    elsif ( $Stick > $Max )
        {
        $Val = $Curve[$Points -1];
        }
    else
        {
        my $i = 0;
        my $dx = ($Max - $Min) / ($Points - 1);
        for (my $x = $Min; $x < $Max; $x += $dx)
            {
            if ( $Stick >= $x  and  $Stick <= $x + $dx)
                {
                my $y1 = $Curve[$i];
                my $y2 = $Curve[$i+1];
                my $dy = $y2 - $y1;

                $Val = $y1 + $dy / $dx * ($Stick - $x);
                last;
                }

            $i ++;
            }
        }

    # Prozent in steps umrechnen
    $Stick = $Val / 100 * $Resolution;

    return $Stick;
    }


# parse input controls for one output channel
# Syntax: [ChannelOption,par,par] + Control1[_Reverse],par,par,par + Control2 ...  + Control3 ...
sub ParseControls
    {
    my ($Channel, $ControlVal, $Expo, $Limit, $Timing) = @_;

    if ( $Expo  eq "" ) { $Expo  = 0; };
    if ( $Limit eq "" ) { $Limit = 100; };

    my $ChannelRes = 125;             # Channel Resolution pos+neg = 250 Steps
    my $ChannelInc = 0;               # Channel Incremental Mode: 0, 1, 2
    my $ChannelReverse = 1;           # Channel reverse factor, 1 , -1
    my $ChannelOffset = 0;            # Channel Offset in %
    my $ChannelTravelNeg = $Limit;    # Channel travel in %
    my $ChannelTravelPos = $Limit;    # Channel travel in %
    my $ChannelLimitNeg =  $Limit;    # Channel limit in %
    my $ChannelLimitPos =  $Limit;    # Channel limit in %
    my $ChannelExpo = $Expo;          # Channel expo in %
    my $ChannelSwitchVal = "OFF";     # Channel switch value in % or "OFF"
    my $ChannelSwitchMin = 100;       # Channel switch min output in %
    my $ChannelSwitchMax = 100;       # Channel switch max output in %
    my @ChannelCurve;                 # Channel curve

    my $IsAsymChannel = 0;
    if ( "ExternControlGas ExternControlHeight" =~ /$Channel/i )
        {
        # special handling for asymmetric channels
        $IsAsymChannel = 1;

        $ChannelLimitNeg = 0;
        $ChannelTravelNeg = 0;
        }
                
    # Channel output is sum of multiple input controls
    my $ChannelVal = 0;
    $ControlVal =~ s/ //g;
    my @Controls = split('\+', $ControlVal);    # controls separated by "+"
    foreach $ControlVal (@Controls)
        {
        # Control params separated by ","
        my @Par = split(',', $ControlVal);
        my $Control = $Par[0]; 
        my $ControlTravelNeg = $Par[1];   # in %
        my $ControlTravelPos = $Par[2];   # in %
        my $ControlExpo      = $Par[3];   # in %
        my $ControlOffset    = $Par[4];   # in %

        # take Pos as Neg, if only Neg is given
        if ( $ControlTravelNeg eq "" ) { $ControlTravelNeg = 100; }
        if ( $ControlTravelPos eq "" ) { $ControlTravelPos = $ControlTravelNeg; }

        if ( $ControlExpo eq "" )   { $ControlExpo = 0; }
        if ( $ControlOffset eq "" ) { $ControlOffset = 0; }

        my $Val = 0;               # control value
        my $ControlReverse = 1;    # control reverse factor: 1, -1

        my ($Control, $Option) = split('_', $Control, 2);      # Control options separeted by "_"
        if ( $Control ne "" )
            {

            #
            # Output Channel Options
            #

            if ( $Control =~ /Rev/i )
                {
                $ChannelReverse *= -1;
                next;
                }

            if ( $Control =~ /IncStop/i )
                {
                # stop at neutral point
                $ChannelInc = 2;
                next;
                }
            elsif ( $Control =~ /Inc/i )
                {
                # don't stop at neutral point
                $ChannelInc = 1;
                next;
                }

            if ( $Control =~ /Offset/i )
                {
                $ChannelOffset = $Par[1];
                if ( $ChannelOffset eq "" ) { $ChannelOffset = 0; };
                next;
                }

            if ( $Control =~ /Travel/i )
                {
                $ChannelTravelNeg = $Par[1];
                $ChannelTravelPos = $Par[2];

                # take symmetrical Pos as Neg, if only Neg is given
                if ( $ChannelTravelNeg eq "" ) { $ChannelTravelNeg = 100; }
                if ( $ChannelTravelPos eq "" ) { $ChannelTravelPos = $ChannelTravelNeg; }
                next;
                }

            if ( $Control =~ /Limit/i )
                {
                $ChannelLimitNeg = $Par[1];
                $ChannelLimitPos = $Par[2];

                if ( $IsAsymChannel  and  $ChannelLimitNeg ne ""  and $ChannelLimitPos eq "" )
                    {
                    # only Neg given. Take Neg as Pos.
                    $ChannelLimitPos = $ChannelLimitNeg;
                    $ChannelLimitNeg = 0;
                    }

                # take symmetrical Pos as Neg, if only Neg is given
                if ( $ChannelLimitNeg eq "" ) { $ChannelLimitNeg = 100; }
                if ( $ChannelLimitPos eq "" ) { $ChannelLimitPos = $ChannelLimitNeg; }
                next;
                }

            if ( $Control =~ /Expo/i )
                {
                $ChannelExpo = $Par[1];
                if ( $ChannelExpo eq "" ) { $ChannelExpo = 0; };
                next;
                }

            if ( $Control =~ /Switch/i )
                {
                $ChannelSwitchVal = $Par[1];
                $ChannelSwitchMin = $Par[2];
                $ChannelSwitchMax = $Par[3];
                if ( $ChannelSwitchVal eq "" ) { $ChannelSwitchVal = "OFF"; };
                if ( $ChannelSwitchMin eq "" ) { $ChannelSwitchMin = 100; };
                if ( $ChannelSwitchMax eq "" ) { $ChannelSwitchMax = 100; };
                next;
                }

            if ( $Control =~ /Curve/i )
                {
                @ChannelCurve = @Par;
                splice @ChannelCurve, 0, 1;
                next;
                }

            #
            # Input Control Options
            #

            if ( $Option =~ /Rev/i )
                {
                $ControlReverse *= -1;
                }

            #
            # Input controls
            #

            # Joystick Button
            if ( $Control =~ /^JoystickButton(\d+)/i )
                {
                my $Button = $1 - 1;
                $Val = &JoystickButton($Button) ? $ChannelRes : -$ChannelRes;
                }

            # Joystick POV Button
            elsif ( $Control =~ /^JoystickPov(\d+)/i )
                {
                my $Angle = $1;
                my $Pov = $Stick{'JoystickPov'} / 100;
                $Val = ($Pov == $Angle) ? $ChannelRes : -$ChannelRes;
                }

            # Mouse Button
            elsif ( $Control =~ /^MouseButton(\d+)/i )
                {
                my $Button = $1 - 1;
                $Val = &MouseButton($Button) ? $ChannelRes : -$ChannelRes;
                }

            # Function Key
            elsif ( $Control =~ /^FctKey(\d+)/i )
                {
                my $Key = $1 - 1;
                $Val = &FctKey($Key) ? $ChannelRes : -$ChannelRes;
                }

            # Function Key (Toggle Mode)
            elsif ( $Control =~ /^FctKeyToggle(\d+)/i )
                {
                my $Key = $1 - 1;
                $Val = &FctKeyToggle($Key) ? $ChannelRes : -$ChannelRes;
                }

            # Serial Channel
            elsif ( $Control =~ /^SerialChannel/i )
                {
                $Val = $MkSerialChannel{$Control};
                }

            # fixed value
            elsif ( $Control =~ /^(-*\d+)/i )
                {
                $Val = $1;
                if ( $IsAsymChannel )
                    {
                    $Val = $Val - $ChannelRes;
                    }
                }

            # Joystick
            elsif ( $Control =~ /^Joystick/i )
                {
                # Scale Stick 0..StickRange to -125..0..125
                $Val = $Stick{$Control} / $Stick{'StickRange'} * 2 * $ChannelRes - $ChannelRes;
                }

            # 3D-Maus
            elsif ( $Control =~ /^Mouse/i )
                {
                # Scale Stick 0..StickRange to -125..0..125
                $Val = $Stick{$Control} / $Stick{'StickRange'} * 2 * $ChannelRes - $ChannelRes;
                }

            # Rc Poti
            elsif ( $Control =~ /^RcPoti/i )
                {
                $Val = $Stick{$Control};
                }

            # Rc Stick
            elsif ( $Control =~ /^RcStick/i )
                {
                $Val = $Stick{$Control};
                }
            else
                {
                # unknown, fall through
                print "ParseControls: $Channel : Unknown Control \"$Control\"\n";
                next;
                }

            # Control Reverse
            $Val *= $ControlReverse;

            # Expo/Limit for each input control
            if ( $IsAsymChannel  and  $ChannelInc == 0 )
                {
                # asymmetric channel 0..250, if not in INC mode
                $Val += $ChannelRes;
                $Val = &ExpoLimit (0.001, 2 * $ChannelRes, $Val, $ControlExpo, $ControlTravelNeg, $ControlTravelPos);
                }
            else
                {
                # symmetric channel -125..0..125
                $Val = &ExpoLimit (- $ChannelRes, $ChannelRes, $Val, $ControlExpo, $ControlTravelNeg, $ControlTravelPos);
                }
 
            # Control Offset
            $Val = $Val + $ControlOffset / 100 * $ChannelRes;

            # sum controls
            $ChannelVal += $Val;
            }
        }

    # Incremental Channel Mode. Control must be neg ..0..pos
    if ( $ChannelInc > 0 )
        {
        # Channel travel time is 1s for 125 steps at control full speed

        # Channel neutral point crossing detection 
        my $Neutral = $ChannelVal;
        my $LastNeutral = $Controls{$Channel}{'Neutral'};
        $Controls{$Channel}{'Neutral'} = $Neutral;
              
        my $LastVal = $Controls{$Channel}{'Value'};
        $ChannelVal = $LastVal + $ChannelVal * $Timing / 1000;

        if ( $ChannelInc > 1 )
            {
            # stop at neutral position
            if ( $LastVal <= 0  and  $ChannelVal >= 0  and  $LastNeutral > 5  and $Neutral > 5 )
                {
                # coming from left
                $ChannelVal = 0;
                }
            elsif ( $LastVal >= 0  and  $ChannelVal <= 0  and  $LastNeutral < -5  and $Neutral < -5 )
                {
                # comimg from right
                $ChannelVal = 0;
                }
            }

        if ( $IsAsymChannel )
            {
            # asymmetric channel 0..250
            if ( $ChannelVal > 2 * $ChannelRes ) { $ChannelVal = 2 * $ChannelRes };
            if ( $ChannelVal < 0 )               { $ChannelVal = 0 };
            }
        else
            {
            # symmetric channel -125..0..125
            if ( $ChannelVal >  $ChannelRes ) { $ChannelVal =  $ChannelRes };
            if ( $ChannelVal < -$ChannelRes ) { $ChannelVal = -$ChannelRes };
            }

        $Controls{$Channel}{'Value'} = $ChannelVal;
        }


    # channel output processing
    if ( $IsAsymChannel )
        {
        # asymmetric channel 0..250

        # Channel Multipoint curve
        $ChannelVal = &Curve (0.001, 2 * $ChannelRes, 2 * $ChannelRes, $ChannelVal, @ChannelCurve);

        # Channel Expo, Travel
        $ChannelVal = &ExpoLimit (0.001, 2 * $ChannelRes, $ChannelVal, $ChannelExpo, $ChannelTravelNeg, $ChannelTravelPos);

        # Channel Switch
        if ( $ChannelSwitchVal ne "OFF" )
            {
            if ( $ChannelVal < 2 * $ChannelRes * $ChannelSwitchVal / 100 )
                {
                $ChannelVal = 2 * $ChannelRes * $ChannelSwitchMin / 100;
                }
            else
                {
                $ChannelVal = 2 * $ChannelRes * $ChannelSwitchMax / 100;
                }
            } 

        # Channel Reverse
        if ( $ChannelReverse == -1 )
            {
            $ChannelVal = 2 * $ChannelRes - $ChannelVal;
            }

        # Channel offset
        $ChannelVal = $ChannelVal + $ChannelOffset / 100 * 2 * $ChannelRes;

        # Channel Limiter
        my $Pos = 2 * $ChannelRes * $ChannelLimitPos / 100;
        if ( $ChannelVal > $Pos )
            {
            $ChannelVal = $Pos;
            }
        my $Neg = 2 * $ChannelRes * $ChannelLimitNeg / 100;
        if ( $ChannelVal < $Neg )
            {
            $ChannelVal = $Neg;
            }

        # round to integer
        if ( $ChannelVal >= 0 )
            {
            $ChannelVal = int ($ChannelVal + 0.5);
            }
        else
            {
            $ChannelVal = int ($ChannelVal - 0.5);
            }
        $ChannelVal = &CheckUnsignedChar($ChannelVal);
        }
    else
        {
        # symmetric channel -125..0..125

        # Channel Multipoint curve
        $ChannelVal = &Curve (-$ChannelRes, $ChannelRes, $ChannelRes, $ChannelVal, @ChannelCurve); 
       
        # Channel Expo, Travel
        $ChannelVal = &ExpoLimit (-$ChannelRes, $ChannelRes, $ChannelVal, $ChannelExpo, $ChannelTravelNeg, $ChannelTravelPos);

        # Channel Switch
        if ( $ChannelSwitchVal ne "OFF" )
            {
            if ( $ChannelVal < $ChannelRes * $ChannelSwitchVal / 100 )
                {
                $ChannelVal = $ChannelRes * $ChannelSwitchMin / 100;
                }
            else
                {
                $ChannelVal = $ChannelRes * $ChannelSwitchMax / 100;
                }
            }

        # Channel Reverse
        $ChannelVal *= $ChannelReverse;

        # Channel offset
        $ChannelVal = $ChannelVal + $ChannelOffset / 100 * $ChannelRes;

        # Channel Limiter
        my $Pos = $ChannelRes * $ChannelLimitPos / 100;
        if ( $ChannelVal > $Pos )
            {
            $ChannelVal = $Pos;
            }
        my $Neg = - $ChannelRes * $ChannelLimitNeg / 100;
        if ( $ChannelVal < $Neg )
            {
            $ChannelVal = $Neg;
            }

        # round to integer
        if ( $ChannelVal >= 0 )
            {
            $ChannelVal = int ($ChannelVal + 0.5);
            }
        else
            {
            $ChannelVal = int ($ChannelVal - 0.5);
            }
        $ChannelVal = &CheckSignedChar($ChannelVal);
        }

    return $ChannelVal;
    }


#
# Waypoint handling
#

# Add a Waypoint to @Waypoints List
sub WpAdd()
    {
    my %Param = @_;
    my $Wp_x              = $Param{'-x'};
    my $Wp_y              = $Param{'-y'};
    my $Lat               = $Param{'-lat'};
    my $Lon               = $Param{'-lon'};
    my $Alt               = $Param{'-alt'};
    my $AltRate           = $Param{'-altrate'};
    my $Heading           = $Param{'-heading'};
    my $ToleranceRadius   = $Param{'-toleranceradius'};
    my $Holdtime          = $Param{'-holdtime'};
    my $EventFlag         = $Param{'-eventflag'};
    my $EventChannelValue = $Param{'-eventchannelvalue'};
    my $Speed             = $Param{'-speed'};
    my $CamAngle          = $Param{'-camangle'};

    # Check validity
    $Alt               = &Default ($Alt,               $Cfg->{'waypoint'}->{'DefaultAltitude'});
    $AltRate           = &Default ($AltRate,           $Cfg->{'waypoint'}->{'DefaultAltRate'} );
    $Heading           = &Default ($Heading,           $Cfg->{'waypoint'}->{'DefaultHeading'} );
    $ToleranceRadius   = &Default ($ToleranceRadius,   $Cfg->{'waypoint'}->{'DefaultToleranceRadius'} );
    $Holdtime          = &Default ($Holdtime,          $Cfg->{'waypoint'}->{'DefaultHoldtime'} );
    $EventFlag         = &Default ($EventFlag,         $Cfg->{'waypoint'}->{'DefaultEventFlag'} );
    $EventChannelValue = &Default ($EventChannelValue, $Cfg->{'waypoint'}->{'DefaultEventChannelValue'} );
    $Speed             = &Default ($Speed,             $Cfg->{'waypoint'}->{'DefaultSpeed'}, 10 );
    $CamAngle          = &Default ($CamAngle,          $Cfg->{'waypoint'}->{'DefaultCamAngle'}, -1 );

    # x/y and/or Lat/Lon must be passed
    if ( $Wp_x eq ""  and  $Wp_y eq "" )
        {
        ($Wp_x, $Wp_y) = &MapGps2XY($Lat, $Lon);
        }
    if ( $Lat eq ""  and  $Lon eq "" )
        {
        ($Lat, $Lon) = &MapXY2Gps($Wp_x, $Wp_y);
        }

    # kind of unique Tag for this Wp    
    my ($t0_s, $t0_us) = gettimeofday;
    my $Tag = sprintf "WP-%d.%d", $t0_s, $t0_us;

    # save Wp-Hash in Waypoint-Array
    my $Wp = {};        
    $Wp->{'Tag'}               = $Tag;
    $Wp->{'MapX'}              = $Wp_x;
    $Wp->{'MapY'}              = $Wp_y;
    $Wp->{'_MapX_Rel'}         = $Wp_x / $MapSizeX;
    $Wp->{'_MapY_Rel'}         = $Wp_y / $MapSizeY;
    $Wp->{'Pos_Lat'}           = $Lat;
    $Wp->{'Pos_Lon'}           = $Lon;
    $Wp->{'Pos_Alt'}           = $Alt;
    $Wp->{'AltRate'}           = $AltRate;
    $Wp->{'Heading'}           = $Heading;
    $Wp->{'ToleranceRadius'}   = $ToleranceRadius;
    $Wp->{'Holdtime'}          = $Holdtime;
    $Wp->{'Event_Flag'}        = $EventFlag;
    $Wp->{'EventChannelValue'} = $EventChannelValue;
    $Wp->{'Speed'}             = $Speed;
    $Wp->{'CamAngle'}          = $CamAngle;

    push @Waypoints, $Wp;

    &Log ("Add WP Tag: $Tag, Lat: $Lat, Lon: $Lon", "A");

    # refresh WP list editor
    &WpEditRefresh();
    }


# Delete Waypoint from @Waypoints List
sub WpDelete ()
    {
    my ($WpIndex) = @_;

    &Log ("Delete WP $WpIndex", "A");

    # delete Wp in Waypoint-Array
    splice @Waypoints, $WpIndex, 1;

    # refresh WP list editor
    &WpEditRefresh();
    }


# Delete all Waypoints
sub WpDeleteAll ()
    {
    &Log ("Delete all WP", "A");

    undef @Waypoints;
    $WpPlayerIndex = 0;
    $WpPlayerHoldtime = -1;
    $SpdPlayerTime = 0;
    $SpdPlayerIndex = 0;

    # remove all Wp-Icons and Wp-Number on canvas
    &WpHide();

    # refresh WP list editor
    &WpEditRefresh();
    }

# Add current MK-Position to @Waypoints
sub WpCapture()
    {
    my $Lat = 0;
    my $Lon = 0;
    my $Alt = 0;
    my $Cycles = 2;
    for my $i ( 1..$Cycles)       # average over n cycles
        {
        $Lat += $MkOsd{'CurPos_Lat'};
        $Lon += $MkOsd{'CurPos_Lon'};
        $Alt += &Altitude();
        usleep (250000);  # 0.25s
        }
    $Lat /= $Cycles;
    $Lon /= $Cycles;
    $Alt /= $Cycles;

    &Log ("Capture WP Lat: $Lat, Lon: $Lon, Alt: $Alt", "A");

    &WpAdd (-lat => $Lat,
            -lon => $Lon,
            -alt => $Alt,
           );
    }

# Load @Waypoints from file
sub WpLoadFile ()
    {
    my ($WpFile) = @_;

    &Log ("Load WP file $WpFile", "A");

    if ( $WpFile =~ /.wpl$/i )
        {
        # load Mikrokopter Tool WP List *.wpl

        my $PointCnt   = 0;
        my $PointIndex = 0;

        my $PoiIndex   = 0;
        my @WpWpl;

        open WPL, "<$WpFile";
        my @Wpl = <WPL>;
        close WPL;
        foreach my $Line (@Wpl)
            {
            chomp $Line;
            if ( $Line =~ /FileVersion\s*=\s*(\d*)/i )
                {
                $FileVersion = $1;
                }
            elsif ( $Line =~ /NumberOfWaypoints\s*=\s*(\d*)/i )
                {
                $PointCnt = $1;
                }
            elsif ( $Line =~ /\[Point(\d*)\]/i )
                {
                $PointIndex = $1 -1;
                }
            elsif ( $Line =~ /(\S*)\s*=\s*(\S*)/i )
                {
                my $Key = $1;
                my $Value = $2;

                $WpWpl[$PointIndex]{$Key} = $Value;
                }
            }

        # WPL Array in Waypoints- und POI-Array umkopieren
        undef @Waypoints;

        for ( $Index=0; $Index < $PointCnt; $Index++)
            {
            my $Type = $WpWpl[$Index]{'Type'};
            if ( $Type == 1 )
                {
                # WP

                my $Wp = {};

                my $Tag = sprintf "Waypoint-%d.%d", time, $Index + 1;   # kind of unique Tag for this WP
    
                my $Lat = $WpWpl[$Index]{'Latitude'};
                my $Lon = $WpWpl[$Index]{'Longitude'};
   
                ($MapX, $MapY) = &MapGps2XY($Lat, $Lon);
                $Wp->{'Tag'}               = $Tag;
                $Wp->{'MapX'}              = $MapX;
                $Wp->{'MapY'}              = $MapY;
                $Wp->{'_MapX_Rel'}         = $MapX / $MapSizeX;
                $Wp->{'_MapY_Rel'}         = $MapY / $MapSizeY;
                $Wp->{'Pos_Lat'}           = $Lat;
                $Wp->{'Pos_Lon'}           = $Lon;
                $Wp->{'Pos_Alt'}           = &Default ($WpWpl[$Index]{'Altitude'},               &Altitude() );
                $Wp->{'AltRate'}           = &Default ($WpWpl[$Index]{'ClimbRate'},              $Cfg->{'waypoint'}->{'DefaultAltRate'}, 5);
                $Wp->{'Heading'}           = &Default ($WpWpl[$Index]{'Heading'},                $Cfg->{'waypoint'}->{'DefaultHeading'},   0);
                $Wp->{'ToleranceRadius'}   = &Default ($WpWpl[$Index]{'Radius'},                 $Cfg->{'waypoint'}->{'DefaultToleranceRadius'}, 5);
                $Wp->{'Holdtime'}          = &Default ($WpWpl[$Index]{'DelayTime'},              $Cfg->{'waypoint'}->{'DefaultHoldtime'}, 2);
                $Wp->{'Event_Flag'}        = &Default ($WpWpl[$Index]{'EventFlag'},              $Cfg->{'waypoint'}->{'DefaultEventFlag'}, 0);
                $Wp->{'EventChannelValue'} = &Default ($WpWpl[$Index]{'WP_Event_Channel_Value'}, $Cfg->{'waypoint'}->{'DefaultEventChannelValue'}, 0);
                $Wp->{'Speed'}             = &Default ($WpWpl[$Index]{'Speed'} / 10 * 3.6,       $Cfg->{'waypoint'}->{'DefaultSpeed'}, 10);
                $Wp->{'CamAngle'}          = &Default ($WpWpl[$Index]{'CAM-Nick'},               $Cfg->{'waypoint'}->{'DefaultCamAngle'}, -1);

                push @Waypoints, $Wp;
                }

            elsif ( $Type == 2 )
                {
                # POI

                $Poi[$PoiIndex]{'Poi_Lat'}       = $WpWpl[$Index]{'Latitude'};
                $Poi[$PoiIndex]{'Poi_Lon'}       = $WpWpl[$Index]{'Longitude'};
                $Poi[$PoiIndex]{'Poi_Alt'}       = $WpWpl[$Index]{'Altitude'};
                $Poi[$PoiIndex]{'Poi_Eventflag'} = $WpWpl[$Index]{'EventFlag'};

                $PoiIndex ++;
                }
            }
        }
    else
        {
        # load Mission Cockpit XML

        # XML in Hash-Ref lesen
        my $Wp = XMLin($WpFile, ForceArray => 1);

        # XML Hash-Ref in Wp-Array umkopieren
        undef @Waypoints;

        foreach $key (sort keys %$Wp)
            {
            my $Point = $Wp->{$key}->[0];

            # relative Pixelkoordinaten auf Bildgroesse umrechnen
            if ( $Point->{'MapX'} <= 1  and  $Point->{'MapY'} <= 1 )
                {
                $Point->{'MapX'} = int ( $Point->{'MapX'} * $MapSizeX + 0.5 );
                $Point->{'MapY'} = int ( $Point->{'MapY'} * $MapSizeY + 0.5 );
                }

            # abs. pixel koordinates not needed
            delete $Point->{'MapX_Pixel'};
            delete $Point->{'MapY_Pixel'};

            $Point->{'_MapX_Rel'} = $Point->{'MapX'} / $MapSizeX;
            $Point->{'_MapY_Rel'} = $Point->{'MapY'} / $MapSizeY;

            $Point->{'Pos_Alt'}           = &Default ($Point->{'Pos_Alt'},           &Altitude());
            $Point->{'AltRate'}           = &Default ($Point->{'AltRate'},           $Cfg->{'waypoint'}->{'DefaultAltRate'},           5);
            $Point->{'Heading'}           = &Default ($Point->{'Heading'},           $Cfg->{'waypoint'}->{'DefaultHeading'},           0);
            $Point->{'ToleranceRadius'}   = &Default ($Point->{'ToleranceRadius'},   $Cfg->{'waypoint'}->{'DefaultToleranceRadius'},   5);
            $Point->{'Holdtime'}          = &Default ($Point->{'Holdtime'},          $Cfg->{'waypoint'}->{'DefaultHoldtime'},          2);
            $Point->{'Event_Flag'}        = &Default ($Point->{'Event_Flag'},        $Cfg->{'waypoint'}->{'DefaultEventFlag'},         0);
            $Point->{'EventChannelValue'} = &Default ($Point->{'EventChannelValue'}, $Cfg->{'waypoint'}->{'DefaultEventChannelValue'}, 0);
            $Point->{'Speed'}             = &Default ($Point->{'Speed'},             $Cfg->{'waypoint'}->{'DefaultSpeed'},            10);
            $Point->{'CamAngle'}          = &Default ($Point->{'CamAngle'},          $Cfg->{'waypoint'}->{'DefaultCamAngle'},         -1);

            # GPS Koordinaten fr die aktuelle Karte neu aus Map x/y berechnen
            my ($Lat, $Lon) = &MapXY2Gps($Point->{'MapX'}, $Point->{'MapY'});
            $Point->{'Pos_Lat'} = $Lat;
            $Point->{'Pos_Lon'} = $Lon;

            push @Waypoints, $Point;
            }
        }
    
    # Start with 1st WP
    &WpTargetFirst();

    # refresh WP list editor
    &WpEditRefresh();
    }


# Save @Waypoints to file
sub WpSaveFile()
    {
    my ($WpFile) = @_;

    &Log ("Save WP file $WpFile", "A");

    if ( $WpFile =~ /.wpl$/i )
        {
        # save Mikrokopter Tool WP List *.wpl

        open WPL, ">$WpFile";

        my $Points = scalar @Poi + scalar @Waypoints;
        my $PointCnt = 1;

        print WPL "[General]\n";
        print WPL "FileVersion=3\n";
        print WPL "NumberOfWaypoints=$Points\n";
        print WPL "UsePOI=1\n";

        for $i ( 0 .. $#Poi )
            {
            printf WPL "[Point%d]\n", $PointCnt;
            print WPL "Latitude=$Poi[$i]{'Poi_Lat'}\n";
            print WPL "Longitude=$Poi[$i]{'Pos_Lon'}\n";
            print WPL "Radius=5\n";
            print WPL "Altitude=$Poi[$i]{'Pos_Alt'}\n";
            print WPL "ClimbRate=20\n";
            print WPL "DelayTime=2\n";
            print WPL "WP_Event_Channel_Value=0\n";
            print WPL "Heading=0\n";
            print WPL "Speed=30\n";
            print WPL "CAM-Nick=-1\n";
            print WPL "Type=2\n";

            $PointCnt ++;
            }

        for $i (0 .. $#Waypoints )
            {
            printf WPL "[Point%d]\n", $PointCnt;
            print WPL "Latitude=$Waypoints[$i]{'Pos_Lat'}\n";
            print WPL "Longitude=$Waypoints[$i]{'Pos_Lon'}\n";
            print WPL "Radius=$Waypoints[$i]{'ToleranceRadius'}\n";
            print WPL "Altitude=$Waypoints[$i]{'Pos_Alt'}\n";
            print WPL "ClimbRate=$Waypoints[$i]{'AltRate'}\n";
            print WPL "DelayTime=$Waypoints[$i]{'Holdtime'}\n";
            print WPL "WP_Event_Channel_Value=$Waypoints[$i]{'EventChannelValue'}\n";
            print WPL "Heading=$Waypoints[$i]{'Heading'}\n";
            print WPL "Speed=$Waypoints[$i]{'Speed'}\n";
            print WPL "CAM-Nick=$Waypoints[$i]{'CamAngle'}\n";
            print WPL "Type=1\n";

            $PointCnt ++;
            }
        close WPL;
        }
    else
        {
        # save Mission Cockpit XML

        # Waypoint-Array in Hash umkopieren
        for  $i ( 0 .. $#Waypoints )
            {
            my $key = sprintf ("WP-%04d", $i);
            my $Wp = {%{$Waypoints[$i]}};        # copy of Hash-content
            $WpOut{$key} = $Wp;

            # Pixelkoordinaten relativ zur Bildgroesse speichern
            $WpOut{$key}{'MapX'} /= $MapSizeX;
            $WpOut{$key}{'MapY'} /= $MapSizeY;
            }

        # WP-Hash als XML speichern
        &XMLout (\%WpOut,
                 'OutputFile' => $WpFile,
                 'AttrIndent' => '1',
                 'RootName' => 'Waypoints',
                );
        }
    }

# Recalc  WP coordinates from relative coordinates
sub WpRecalc()
    {
    for $i (0 .. $#Waypoints)
        {
        my $Wp = $Waypoints[$i];

        # new pixel coordinates from relative coordinates
        $Wp->{'MapX'} = int ( $Wp->{'_MapX_Rel'} * $MapSizeX + 0.5 );
        $Wp->{'MapY'} = int ( $Wp->{'_MapY_Rel'} * $MapSizeY + 0.5 );

        # new GPS coordinates from pixel coordinates
        my ($Lat, $Lon) = &MapXY2Gps($Wp->{'MapX'}, $Wp->{'MapY'});
        $Wp->{'Pos_Lat'} = $Lat;
        $Wp->{'Pos_Lon'} = $Lon;
        }
    }


# Get Wp Index from Canvas Id
sub WpGetIndexFromId()
    {
    my ($id) = @_;

    my @Tags = $map_canvas->gettags($id);
    my $WpTag = $Tags[1];

    for $i (0 .. $#Waypoints)
        {
        my $Wp = $Waypoints[$i];
        if ( $Wp->{'Tag'} eq $WpTag )
            {
            # got it
            return $i;
            }
        }
    return -1;
    }

	
# Resend all Waypoints to MK
sub WpSendAll()
    {
    &Log ("Resend all WPs", "A");

    # OSD/Debug Abfragefrequenz verringern, sonst kommen nicht alle Wp im MK an
    # Sicherheitshalber doppelt senden
    $MkSendWp = 1;       # verhindert ueberschreiben im Timer

    $MkSendQueue->enqueue( "o", "$AddrNC", pack ("C", 1000) );   # Frequenz OSD Datensatz, * 10ms
    $MkSendQueue->enqueue( "d", "$AddrNC", pack ("C", 1000) );   # Frequenz MK Debug Datensatz, * 10ms
    usleep (200000);
    $MkSendQueue->enqueue( "o", "$AddrNC", pack ("C", 1000) );   # Frequenz OSD Datensatz, * 10ms
    $MkSendQueue->enqueue( "d", "$AddrNC", pack ("C", 1000) );   # Frequenz MK Debug Datensatz, * 10ms
    usleep (200000);

    # Alte WP-Liste im MK lschen
    my $Wp = $Waypoints[0];
    &MkFlyTo ( '-lat'  => $Wp->{'Pos_Lat'},
               '-lon'  => $Wp->{'Pos_Lon'},
               '-mode' => "Waypoint Delete"
             );
		
    for $i (0 .. $#Waypoints)
        {
        my $Wp = $Waypoints[$i];
        &MkFlyTo ( '-lat'               => $Wp->{'Pos_Lat'},
                   '-lon'               => $Wp->{'Pos_Lon'},
                   '-alt'               => $Wp->{'Pos_Alt'},
                   '-altrate'           => $Wp->{'AltRate'},
                   '-heading'           => $Wp->{'Heading'},
                   '-toleranceradius'   => $Wp->{'ToleranceRadius'},
                   '-holdtime'          => $Wp->{'Holdtime'},
                   '-eventflag'         => $Wp->{'Event_Flag'},
                   '-eventchannelvalue' => $Wp->{'EventChannelValue'},
                   '-altrate'           => $Wp->{'AltRate'},
                   '-speed'             => $Wp->{'Speed'},
                   '-camangle'          => $Wp->{'CamAngle'},
                   '-mode'              => "Waypoint",
                   '-index'             => $i,
                 );
		
        usleep (150000)  # NC Zeit zum Verarbeiten geben
        }

    $MkSendWp = 0;  # normale OSD/Debug Abfragefrequenz wird automatisch im 5s Timer wieder eingestellt

    # grey connectors: Wp are sent to MK
    $map_canvas->itemconfigure('Waypoint-Connector',
                               '-fill' => $Cfg->{'mkcockpit'}->{'ColorWpConnector'},
                              );

    # MK ist nun synchron mit @Waypoints
    $WaypointsModified = 0;
    }		

	
# Redraw Waypoint Icons
sub WpRedrawIcons()
    {
    if ( $PlayerWptKmlMode eq 'WPT' and ($PlayerRandomMode eq 'STD' or $PlayerRandomMode eq 'RND')  or
         $PlayerWptKmlMode eq 'SPD' )
        { 
        # delete old icons and Wp-Number from canvas
        $map_canvas->delete('Waypoint');
        $map_canvas->delete('WaypointNumber');

        # create new icons
        for $i (0 .. $#Waypoints)
           {
            my $Wp = $Waypoints[$i];
            my $x = $Wp->{'MapX'};
            my $y = $Wp->{'MapY'};
            my $Tag = $Wp->{'Tag'};

            # Waypoint Icon
            my $IconHeight = 48;
            my $IconWidth = 24;
            $map_canvas->createImage($x-$IconWidth/2, $y-$IconHeight,
                                     '-tags' => ['Waypoint', $Tag],
                                     '-anchor' => 'nw',
                                     '-image'  => 'Waypoint-Photo',
                                    );
            # Waypoint Number
            my $WpNumber = $i + 1;
            $map_canvas->createText ( $x+3, $y-$IconHeight/2+12,
                                      '-tags' => ['WaypointNumber', $Tag],
                                      '-text' => $WpNumber,
                                      '-font' => '-*-Arial-Bold-R-Normal--*-100-*',
                                      '-fill' => $Cfg->{'mkcockpit'}->{'ColorWpNumber'},
                                      '-anchor' => 'w',
                                     );
            }
        $map_canvas->lower('Waypoint', 'Target');           # waypoint below Target
        $map_canvas->lower('WaypointNumber', 'Waypoint');   # waypoint-number below waypoint
        }
    }

	
# Redraw Waypoint+spline connectors, calc Spline coefficients
sub WpRedrawLines()
    {
    # delete old connectors from canvas
    $map_canvas->delete('Waypoint-Connector');  

    # Delete old spline
    $map_canvas->delete('Waypoint-Spline');

    if ( scalar(@Waypoints) < 2 )
        {
        # need at least 2 WPs
        return;
        }

    if ( $PlayerWptKmlMode eq 'SPD')
        {
        #
        # draw spline polygon
        #

        my @WpX;
        my @WpY;
        for $i (0 .. $#Waypoints)
            {
            my $Wp = $Waypoints[$i];
            push @WpX, $Wp->{'MapX'};
            push @WpY, $Wp->{'MapY'};
            }

        # Calc spline coefficients
        # global - also unsed in SPD-Timer
        @SplineCoeffX = &SplineCalcNaturalCubic(@WpX);
        @SplineCoeffY = &SplineCalcNaturalCubic(@WpY);

        # calc spline polygon
        # very crude technique - just break each segment up into Steps lines
        my $Steps = 16;

        my @Polygon;
        push @Polygon, &SplineEval(0, %{$SplineCoeffX[0]});
        push @Polygon, &SplineEval(0, %{$SplineCoeffY[0]});

        for $i (0 .. $#SplineCoeffX)
            {
            for ($j = 1; $j <= $Steps; $j++)
                {
                $u = $j / $Steps;
                push @Polygon, &SplineEval($u, %{$SplineCoeffX[$i]});
                push @Polygon, &SplineEval($u, %{$SplineCoeffY[$i]});
                }
            }

        # draw polygon
        $map_canvas->createLine ( @Polygon,
                                  -tags => 'Waypoint-Spline',
                                  -fill => &Default ($Cfg->{'mkcockpit'}->{'ColorWpConnector'}, 'white'),
                                  -width => 1,
                                );

        $map_canvas->raise('Waypoint-Spline', 'Map');      # Spline above map
        }

    elsif ( $PlayerWptKmlMode eq 'WPT'  and  $PlayerRandomMode eq 'STD' )
        {
        #
        # draw WP-connectors
        #
        my $Color = $Cfg->{'mkcockpit'}->{'ColorWpConnector'};
        if ( $WaypointsModified )
            {
            $Color = $Cfg->{'mkcockpit'}->{'ColorWpResend'};
            }

        my $Wp = $Waypoints[0];
        my $x_0 = $Wp->{'MapX'};
        my $y_0 = $Wp->{'MapY'};
        my $x_last = $x_0;
        my $y_last = $y_0;

        for $i (1 .. $#Waypoints)
            {
            my $Wp = $Waypoints[$i];
            my $x = $Wp->{'MapX'};
            my $y = $Wp->{'MapY'};

            $map_canvas->createLine ( $x_last, $y_last, $x, $y,
                                      '-tags' => 'Waypoint-Connector',
                                      '-arrow' => 'last',
                                      '-arrowshape' => [10, 10, 3 ],
                                      '-fill' => $Color,
                                      '-width' => 1,
                                    );
            $x_last = $x;
            $y_last = $y;
            }
        # Connector last WP to 1st WP
        $map_canvas->createLine ( $x_last, $y_last, $x_0, $y_0,
                                  '-tags' => 'Waypoint-Connector',
                                  '-arrow' => 'last',
                                  '-arrowshape' => [10, 10, 3 ],
                                  '-fill' => $Color,
                                  '-width' => 1,
                                );
		
        $map_canvas->raise('Waypoint-Connector', 'Map');   # connector above map
        }
    }


# Hide Waypoints and connectors on Canvas
sub WpHide()
   {
   $map_canvas->delete('Waypoint');
   $map_canvas->delete('WaypointNumber');
   $map_canvas->delete('Waypoint-Connector');
   $map_canvas->delete('Waypoint-Spline');
   }
   

# Convert WP polygon to Border polygon
sub BorderWpConvert()
    {
    my $Wpts = scalar @Waypoints;
    if ( $Wpts >= 3 )
        {
        my @Border;

        for (my $i=0; $i < $Wpts; $i++)
            {
            my $Wp = $Waypoints[$i];
            my $x = $Wp->{'MapX'};
            my $y = $Wp->{'MapY'};
            push @Border, $x, $y;
           }
        @{$Maps{'Current'}->{'Border'}} = @Border;
        }
    }

# Save Border polygon
sub BorderSave()
    {
    my $Border = "";
    my $Sep = "";
    foreach $i (@{$Maps{'Current'}->{'Border'}})
        {
        $Border = $Border . "$Sep" .  "$i";
        $Sep = ",";
        }
    $Xml{'Border'} = $Border;

    # Delete old border file
    my $BorderFile = &BorderFileDelete();

    # Save Border as XML
    if ( $Border ne "" )
        {
        &XMLout (\%Xml,
                 'OutputFile' => $BorderFile,
                 'AttrIndent' => '1',
                 'RootName' => 'Border',
                );
        }
    }


# Delete Border polygon
sub BorderDelete()
    {
    undef @{$Maps{'Current'}->{'Border'}};

    # delete optional border XML
    &BorderFileDelete();
    }


# Hide Kml-Track on Canvas
sub KmlHide()
   {
   $map_canvas->delete('KML-Track');
   }


# Load @KmlTargets from file
sub KmlLoadFile()
    {
    my ($File) = @_;

    &Log ("Load KML file $File", "A");

    # XML in Hash-Ref lesen
    my $Kml = XMLin($File);

    # init state maschine
    undef @KmlTargets;
    $KmlPlayerIndex = 0;

    my $Coordinates = $Kml->{Document}->{Placemark}->{LineString}->{coordinates};
    foreach $Line (split "\n", $Coordinates)
        {
        chomp $Line;
        $Line =~ s/\s//g;       # remove white space
        if ( $Line ne "" )
            {
            my ($Lon, $Lat, $Alt) = split ",", $Line;
            $Lon = sprintf ("%f", $Lon);
            $Lat = sprintf ("%f", $Lat);
            $Alt = sprintf ("%f", $Alt);

            push @KmlTargets, {'Lat'  => $Lat,
                               'Lon'  => $Lon,
                               'Alt'  => $Alt,  # relative to ground
                               'Gier' => 0,     # unknown heading, no heading change
                              };
            }
        }
    }

# Redraw KML track
sub KmlRedraw()
    {

    # delete old Track from canvas
    $map_canvas->delete('KML-Track');

    my @Track;

    foreach $Target ( @KmlTargets )
        {
        my $Lat = $Target->{'Lat'};
        my $Lon = $Target->{'Lon'};
        my $Alt = $Target->{'Alt'};
        my ($x, $y) = &MapGps2XY($Lat, $Lon);
        push @Track, $x, $y;
        }

    if ( scalar @Track >= 4 )  # at least 2 Koordinaten-Paare
        {
        $map_canvas->createLine ( @Track,
                                  '-tags' => 'KML-Track',
                                  '-fill' => $Cfg->{'mkcockpit'}->{'ColorKmlTrack'},
                                  '-width' => 1,
                                );	 

        $map_canvas->lower('KML-Track', 'Target');        # Track below Target
        }
    }


# Redraw Footprint
sub FootprintRedraw()
    {
    # delete old Footprint from canvas
    $map_canvas->delete('Footprint');  

    if ( scalar @Footprint >= 4 )  # at least 2 Koordinaten-Paare
        {
        $map_canvas->createLine ( @Footprint,
                                  '-tags' => 'Footprint',
                                  '-fill' => $Cfg->{'mkcockpit'}->{'ColorFootprint'},
                                  '-width' => 1,
                                );	 
        }
		
    $map_canvas->lower('Footprint', 'Target');
    }


# Waypoint Player: Set Waypoint - sequence or random
sub WpTargetSet()
    {
    my ($Index) = @_;

    my $WpCnt = scalar @Waypoints;
    if ( $Index < 0  or  $Index >= $WpCnt )
        {
        # invalid WP number
        return 1;
        }

    &Log ("Set target WP $Index", "A");

    my $Wp         = $Waypoints[$Index];
    my $Wp_x       = $Wp->{'MapX'};
    my $Wp_y       = $Wp->{'MapY'};
    my $Wp_Heading = $Wp->{'Heading'};
    my $Wp_Alt     = $Wp->{'Pos_Alt'};

    # is Wp reachable?
    if ( ! &IsTargetReachable($Wp_x, $Wp_y) )
        {
        # new Wp-Target is not reachable
        &Log ("Target WP $Index not reachable", "A");
        return 1;
        }

    # set new Wp-Target 
    $WpPlayerIndex = $Index;
    $WpPlayerHoldtime = -1;

    # Gier Control
    $System{'GierSetpoint'} = &CompassCalc( $Wp_Heading );

    # Alt Control
    $System{'AltSetpoint'} = $Wp_Alt;
    
    return 0;
    }


# Waypoint Player: Goto next Waypoint - sequence or random
sub WpTargetNext()
    {
    &Log ("Set next target WP", "A");

    my $WpCnt = scalar @Waypoints;

    # Std- or Random Waypoint sequence
    if ( $PlayerRandomMode =~ /STD/i  or
         $PlayerRandomMode =~ /RND/i )
        {
        $NewIndex = $WpPlayerIndex;

        # get next Wp
        for ( $i=0; $i<10; $i++)        # avoid deadlock, if no WP reachable
            {
            for ( $j=0; $j<10; $j++ )   # avoid deadlock, if only 1 WP
                {
   
                if ( $PlayerRandomMode =~ /STD/i )
                    {
                    $NewIndex ++;
                    if ( $NewIndex >= $WpCnt )
                        {
                        # Restart with 1st Wp
                        $NewIndex = 0;
                        }
                    }
  
                if ( $PlayerRandomMode =~ /RND/i )
                    {
                    $NewIndex = int (rand($WpCnt));
                    }

                # want to have different Wp 
                if ( $NewIndex ne $WpPlayerIndex )
                    {
                    last;
                    }
                }

            # Set new Target 
            if ( &WpTargetSet ($NewIndex) == 0 )
                {
                # new Wp-Target found
                last;
                }
            }
        }

    # Random Map sequence
    if ( $PlayerRandomMode =~ /MAP/i )
        {
        $RandomTarget_x = $System{'CurPos_x'};
        $RandomTarget_y = $System{'CurPos_y'};

        for ( $i=0; $i<50; $i++)        # avoid deadlock, if target not reachable
            {
            # don't use 10% around the map
            my $New_x = int (rand($MapSizeX - 2 * $MapSizeX/10)); 
            my $New_y = int (rand($MapSizeY - 2 * $MapSizeY/10)); 
            $New_x += $MapSizeX/10; 
            $New_y += $MapSizeY/10; 

            # is Target reachable?
            if ( &IsTargetReachable($New_x, $New_y) )
                {
                # new Target found
                &Log ("Set random target WP x: $New_x, y: $New_y", "A");

                $RandomTarget_x = $New_x;
                $RandomTarget_y = $New_y;
                last;
                }
            }
        }
    
    &TtsSpeak ('MEDIUM', $Translate{'TtsNextTarget'});

    $WpPlayerHoldtime = -1;

    # Altitude Control
    &AltCurPosSave()
    }


# Waypoint Player: Goto previous Waypoint
sub WpTargetPrev()
    {
    &Log ("Set previous target WP", "A");

    if ( $PlayerRandomMode =~ /STD/i )
        {
        $WpPlayerIndex --;
        if ( $WpPlayerIndex < 0 )
            {
            # Restart with last Wp
            $WpPlayerIndex = $#Waypoints;
            }
        }
    else
        {
        # Next Random Target
        &WpTargetNext();
        }

    $WpPlayerHoldtime = -1;

    # Altitude Control
    &AltCurPosSave();
    if ( scalar @Waypoints > 0 )
        {
        $System{'AltSetpoint'} = $Waypoints[$WpPlayerIndex]{'Pos_Alt'};
        }
    }


# Waypoint Player: Goto first Waypoint
sub WpTargetFirst()
    {
    &Log ("Set first target WP", "A");

    $WpPlayerIndex = 0;
    $WpPlayerHoldtime = -1;

    # Altitude Control
    &AltCurPosSave();
    if ( scalar @Waypoints > 0 )
        {
        $System{'AltSetpoint'} = $Waypoints[$WpPlayerIndex]{'Pos_Alt'};
        }
    }

# Waypoint Player: Goto last Waypoint
sub WpTargetLast()
    {
    &Log ("Set last target WP", "A");

    $WpPlayerIndex = $#Waypoints;
    $WpPlayerHoldtime = -1;

    # Altitude Control
    &AltCurPosSave();
    if ( scalar @Waypoints > 0 )
        {
        $System{'AltSetpoint'} = $Waypoints[$WpPlayerIndex]{'Pos_Alt'};
        }
    }


# Waypoint Player: Waypoint Target reached?
sub WpCheckTargetReached()
    {
    if ( $WpPlayerHoldtime == -1 )
        {
        lock (%MkOsd);              # until end of block

        if ( &CurPosIsValid() and  &HomePosIsValid() and  &MkIsWptMode() )
            {
            # Gueltige SAT Daten

            # for Wp mode
            my $Wp = $Waypoints[$WpPlayerIndex];
            my $WpTarget_Lat = $Wp->{'Pos_Lat'};
            my $WpTarget_Lon = $Wp->{'Pos_Lon'};
            my $WpTarget_Alt = $Wp->{'Pos_Alt'};
            my $WpTolerance  = $Wp->{'ToleranceRadius'};
            my $WpHoldtime   = $Wp->{'Holdtime'};

            # Random-Map Mode
            if ( $PlayerRandomMode =~ /MAP/i )
                {
                ($WpTarget_Lat, $WpTarget_Lon) = &MapXY2Gps ($RandomTarget_x, $RandomTarget_y);
                $WpTolerance  = &Default ($Cfg->{'waypoint'}->{'DefaultToleranceRadius'}, 5);
                $WpHoldtime   = &Default ($Cfg->{'waypoint'}->{'DefaultHoldtime'}, 10);
                $WpTarget_Alt = $System{'Alt'};
                }

            # Operation Radius pruefen
            my ($HomeDist, $HomeBearing) = &MapGpsTo($MkOsd{'HomePos_Lat'}, $MkOsd{'HomePos_Lon'}, $WpTarget_Lat, $WpTarget_Lon );
            if ( $HomeDist > $MkOsd{'OperatingRadius'} )
                {
                # Target entsprechend Operation Radius neu berechnen
                $HomeDist = $MkOsd{'OperatingRadius'};
                ($WpTarget_Lat, $WpTarget_Lon) = &MapGpsAt($MkOsd{'HomePos_Lat'}, $MkOsd{'HomePos_Lon'}, $HomeDist, $HomeBearing);
                }

            # Abstand/Hoehe zum Ziel pruefen
            my ($Dist, $Bearing) = &MapGpsTo($MkOsd{'CurPos_Lat'}, $MkOsd{'CurPos_Lon'}, $WpTarget_Lat, $WpTarget_Lon );
            my $AltDist = 0;
            if ( $System{'TargetSetpoint_AltRate'} != 0)
                {
                # AltCtrl is active
                $AltDist = abs ($WpTarget_Alt - $System{'Alt'});
                }

            $Dist = int ($Dist + 0.5);
            $AltDist = int ($AltDist + 0.5);
            if ( $Dist <= $WpTolerance  and  $AltDist <= $WpTolerance )
                {
                # Target reached - count down Holdtime
                $WpPlayerHoldtime = 2 * $WpHoldtime;      # 0..2n - decrement im 0.5s timer

                &TtsSpeak ('MEDIUM', $Translate{'TtsTargetReached'});

                &Log ("Target reached Lat: $WpTarget_Lat, Lon: $WpTarget_Lon", "A");
                }
            }
        }

    if ( $WpPlayerHoldtime == 0 )  # wird im 0.5s timer runtergezaehlt
        {
        # Target reached - Holdtime is over
        $WpPlayerHoldtime = -1;

        &Log ("Target holdtime over", "A");

        return 1;
        }

    # Target NOT reached
    return 0;
    }


# KML Player: 10s forward
sub KmlTargetNext()
    {
    &Log ("KML 10s forward", "A");

    $KmlPlayerIndex += int (10 / $Cfg->{waypoint}->{'KmlTimeBase'} + 0.5);
    if ( $KmlPlayerIndex > $#KmlTargets )
        {
        # Next loop
        $KmlPlayerIndex -= $#KmlTargets;
        }
    }

# KML Player: 10s backward
sub KmlTargetPrev()
    {
    &Log ("KML 10s backward", "A");

    $KmlPlayerIndex -= int (10 / $Cfg->{waypoint}->{'KmlTimeBase'} + 0.5);
    if ( $KmlPlayerIndex < 0 )
        {
        # Next loop
        $KmlPlayerIndex += $#KmlTargets;
        }
    }

# KML Player: Goto first Target
sub KmlTargetFirst()
    {
    &Log ("KML goto start", "A");

    $KmlPlayerIndex = 0;
    }

# KML Player: Goto last Target
sub KmlTargetLast()
    {
    &Log ("KML goto end", "A");

    $KmlPlayerIndex = $#KmlTargets;
    }


# SPD Player: 10s forward
sub SpdTargetNext()
    {
    &Log ("SPD 10s forward", "A");

    $SpdPlayerTime += 10;

    # Altitude Control
    &AltCurPosSave()
    }

# SPD Player: 10s backward
sub SpdTargetPrev()
    {
    &Log ("KML 10s backward", "A");

    $SpdPlayerTime -= 10;
    if ( $SpdPlayerTime < 0 )
        {
        $SpdPlayerTime = 0;
        $System{'AltWpIndex'} = 0;
        }

    # Altitude Control
    &AltCurPosSave()
    }

# SPD Player: Goto first Target
sub SpdTargetFirst()
    {
    &Log ("SPD goto start", "A");

    $SpdPlayerTime = 0;
    $SpdPlayerIndex = 0;
    $System{'AltWpIndex'} = 0;

    # Altitude Control
    &AltCurPosSave();
    }

# SPD Player: Goto last Target
sub SpdTargetLast()
    {
    &Log ("SPD goto end", "A");

    $SpdPlayerTime = 0;
    $SpdPlayerIndex = 0;
    $System{'AltWpIndex'} = 0;

    # Altitude Control
    &AltCurPosSave();
    }


#
# Set Player modes
#

# set Player mode
sub PlayerModeSet()
    {
    my ($Mode) = @_;

    &Log ("Set player mode: $Mode", "A");

    if    ( $Mode =~ /play/i )  { &PlayerPlay(); }
    elsif ( $Mode =~ /pause/i ) { &PlayerPause(); }
    elsif ( $Mode =~ /home/i )  { &PlayerHome(); }
    elsif ( $Mode =~ /stop/i )  { &PlayerStop(); }
    }


# set player to "Play" mode
sub PlayerPlay()
    {
    &Log ("Set player mode play", "A");

    $PlayerMode = 'Play';
    $WpPlayerHoldtime = -1;

    # Play/Pause-Icon loeschen und neu anzeigen
    $map_canvas->delete('Wp-PlayPause');
    $map_canvas->createImage($MapSizeX/2+150, $MapSizeY-48,
                             '-tags' => 'Wp-PlayPause',
                             '-anchor' => 'nw',
                             '-image'  => 'WpPause-Foto',
                             );
    &FoxHide();
    &CrosshairHide();

    # Altitude control
    &AltCurPosSave();
    if ( scalar @Waypoints > 0 )
        {
        $System{'AltSetpoint'} = $Waypoints[$WpPlayerIndex]{'Pos_Alt'};
        }
    }


# set player to "Pause" mode
sub PlayerPause()
    {
    &Log ("Set player mode pause", "A");

    $PlayerMode = 'Pause';
    $WpPlayerHoldtime = -1;

    # Play/Pause-Icon loeschen und neu anzeigen
    $map_canvas->delete('Wp-PlayPause');
    $map_canvas->createImage($MapSizeX/2+150, $MapSizeY-48,
                             '-tags' => 'Wp-PlayPause',
                             '-anchor' => 'nw',
                             '-image'  => 'WpPlay-Foto',
                             );

    # momentane Position merken und im Player-Timer senden
    $PlayerPause_Lon = "";
    $PlayerPause_Lat = "";

    lock (%MkOsd);              # until end of block
    if ( &CurPosIsValid() )
        {
        $PlayerPause_Lon = $MkOsd{'CurPos_Lon'};
        $PlayerPause_Lat = $MkOsd{'CurPos_Lat'};
        }

    &FoxShow();

    # restart crosshair timer
    $CrosshairTimerCnt = 0;

    # Gier Control - set current Heading
    $System{'GierSetpoint'} = $MkOsd{'CompassHeading'};

    # Altitude Control - set current altitude
    &AltCurPosSave();
    $System{'AltSetpoint'}  = &Altitude();
    }


# set player to "Home" mode
sub PlayerHome()
    {
    &Log ("Set player mode home", "A");

    $PlayerMode = 'Home';
    &WpTargetFirst();

    # Play/Pause-Icon loeschen und neu anzeigen
    $map_canvas->delete('Wp-PlayPause');
    $map_canvas->createImage($MapSizeX/2+150, $MapSizeY-48,
                             '-tags'   => 'Wp-PlayPause',
                             '-anchor' => 'nw',
                             '-image'  => 'WpPlay-Foto',
                             );
    &FoxHide();
    &CrosshairHide();

    # Altitude Control - set current altitude
    &AltCurPosSave();
    $System{'AltSetpoint'}  = &Altitude();

    # Gier Control - set current Heading
    $System{'GierSetpoint'} = $MkOsd{'CompassHeading'};
    }


# set player to "Stop" mode
sub PlayerStop()
    {
    &Log ("Set player mode stop", "A");

    $PlayerMode = 'Stop';
    &WpTargetFirst();

    # set Play/Pause Icon to "Play
    $map_canvas->delete('Wp-PlayPause');
    $map_canvas->createImage($MapSizeX/2+150, $MapSizeY-48,
                             '-tags'   => 'Wp-PlayPause',
                             '-anchor' => 'nw',
                             '-image'  => 'WpPlay-Foto',
                             );

    # switch player to Wp Mode
    &PlayerWpt();

    &FoxHide();
    &CrosshairHide();

    # Altitude control
    &AltCurPosSave();
    $System{'AltSetpoint'}  = &Altitude();
    }


# set player Random Mode to "STD"
sub PlayerRandomStd()
    {
    &Log ("Set player random mode STD", "A");

    if ( $PlayerWptKmlMode ne 'WPT' )
        {
        # not in WPT Mode
        return;
        }

    $PlayerRandomMode = "STD";

    # Set Icon
    $map_canvas->delete('Wp-WptRandom');
    $map_canvas->createImage($MapSizeX/2-200, $MapSizeY-48,
                             '-tags'   => 'Wp-WptRandom',
                             '-anchor' => 'nw',
                             '-image'  => 'WpRandomOn-Foto',
                            );

    # redraw connectors and Icons on canvas
    &WpRedrawLines();
    &WpRedrawIcons();

    # Altitude control
    &AltCurPosSave();
    if ( scalar @Waypoints > 0 )
        {
        $System{'AltSetpoint'} = $Waypoints[$WpPlayerIndex]{'Pos_Alt'};
        }
    }


# set player Random Mode to "RND"
sub PlayerRandomRnd()
    {
    &Log ("Set player random mode RND", "A");

    if ( $PlayerWptKmlMode ne 'WPT' )
        {
        # not in WPT Mode
        return;
        }

    $PlayerRandomMode = "RND";

    # Set Icon
    $map_canvas->delete('Wp-WptRandom');
    $map_canvas->createImage($MapSizeX/2-200, $MapSizeY-48,
                             '-tags' => 'Wp-WptRandom',
                             '-anchor' => 'nw',
                             '-image'  => 'WpRandomMap-Foto',
                            );

    # delete Wp-connectors from canvas
    $map_canvas->delete('Waypoint-Connector');
    $map_canvas->delete('Waypoint-Spline');

    # Altitude control
    &AltCurPosSave();
    if ( scalar @Waypoints > 0 )
        {
        $System{'AltSetpoint'} = $Waypoints[$WpPlayerIndex]{'Pos_Alt'};
        }
    }


# set player Random Mode to "MAP"
sub PlayerRandomMap()
    {
    &Log ("Set player random mode MAP", "A");

    if ( $PlayerWptKmlMode ne 'WPT' )
        {
        # not in WPT Mode
        return;
        }

    $PlayerRandomMode = "MAP";

    # Set Icon
    $map_canvas->delete('Wp-WptRandom');
    $map_canvas->createImage($MapSizeX/2-200, $MapSizeY-48,
                             '-tags' => 'Wp-WptRandom',
                             '-anchor' => 'nw',
                             '-image'  => 'WpRandomOff-Foto',
                            );

    # Get 1st Target
    &WpTargetNext();

    # hide WP and connectors on canvas
    &WpHide();

    # Altitude control
    &AltCurPosSave();
    $System{'AltSetpoint'} = &Altitude();
    }


# set player Pause Mode to "MAP", "MK"
sub PlayerPauseMode()
    {
    ($PlayerPauseMode) = @_;

    &Log ("Set player pause mode $PlayerPauseMode", "A");
    }


# set player to KML mode
sub PlayerKml()
    {
    &Log ("Set player KML mode", "A");

    $PlayerWptKmlMode = 'KML';

    # Wpt/Kml-Player-Icon loeschen und neu anzeigen
    $map_canvas->delete('Wp-WptKml');
    $map_canvas->createImage($MapSizeX/2-250, $MapSizeY-48,
                             '-tags' => 'Wp-WptKml',
                             '-anchor' => 'nw',
                             '-image'  => 'WpKml-Foto',
                             );

    # delete Waypoints from canvas
    &WpHide();

    # show KML Track
    &KmlRedraw();

    # Altitude control
    &AltCurPosSave();
    }


# set player to WPT mode
sub PlayerWpt()
    {
    &Log ("Set player WPT mode", "A");

    $PlayerWptKmlMode = 'WPT';

    # Wpt/Kml-Player-Icon loeschen und neu anzeigen
    $map_canvas->delete('Wp-WptKml');
    $map_canvas->createImage($MapSizeX/2-250, $MapSizeY-48,
                             '-tags' => 'Wp-WptKml',
                             '-anchor' => 'nw',
                             '-image'  => 'WpWpt-Foto',
                             );

    # delete Kml-Track from canvas
    &KmlHide();

    # remove all Wp-Icons and Wp-Number on canvas
    &WpHide(); 

    # Show waypoints, WP resend required
    $WaypointsModified = 1;

    if ( $PlayerRandomMode ne 'MAP' )
        {
        &WpRedrawIcons()
        }
    if ( $PlayerRandomMode eq 'STD' )
        {
        &WpRedrawLines()
        }

    # Altitude control
    &AltCurPosSave();
    if ( scalar @Waypoints > 0 )
        {
        $System{'AltSetpoint'} = $Waypoints[$WpPlayerIndex]{'Pos_Alt'};
        }
    }


# set player to SPD mode
sub PlayerSpd()
    {
    &Log ("Set player SPD mode", "A");

    $PlayerWptKmlMode = 'SPD';

    # Wpt/Spd/Kml-Player-Icon loeschen und neu anzeigen
    $map_canvas->delete('Wp-WptKml');
    $map_canvas->createImage($MapSizeX/2-250, $MapSizeY-48,
                             '-tags' => 'Wp-WptKml',
                             '-anchor' => 'nw',
                             '-image'  => 'WpSpd-Foto',
                             );

    # delete Kml-Track from canvas
    &KmlHide();

    # remove all Wp-Icons and Wp-Number on canvas
    &WpHide(); 

    # Show waypoints, WP resend required
    $WaypointsModified = 1;

    # show icon and spline
    &WpRedrawIcons();
    &WpRedrawLines();

    # Altitude control
    &AltCurPosSave();
    }


# Activate Recording mode
sub PlayerRecordOn
    {
    &Log ("Start recording", "A");

    $PlayerRecordMode = "REC";
    $map_canvas->itemconfigure ('MK-OSD-Rec-Value', '-text' => "Recording" );

    # Record new KML-Track
    undef @KmlTargets;
    $KmlPlayerIndex = 0;

    # delete Kml-Track from canvas
    &KmlHide();
    }

# Deactivate Recording mode
sub PlayerRecordOff
    {
    &Log ("Stop recording", "A");

    $PlayerRecordMode = "";
    $map_canvas->itemconfigure ('MK-OSD-Rec-Value', '-text' => "" );
    }


# Hide Fox icon on canvas
sub FoxHide()
    {
    &Log ("Hide Fox", "A");

    $map_canvas->lower('Fox', 'Map');
    }

# Show Fox icon on canvas
sub FoxShow()
    {
    &Log ("Show Fox", "A");

    $map_canvas->raise('Fox', 'Target');
    }

# Hide POI icon on canvas
sub PoiHide()
    {
    &Log ("Hide POI", "A");

    $map_canvas->lower('POI', 'Map');

    # disable gier to POI
    $System{'GierSetpoint'} = $MkOsd{'CompassHeading'};
    }

# Show POI icon on canvas
sub PoiShow()
    {
    &Log ("Show POI Lat: $Poi[$PoiIndex]{'Poi_Lat'}, Lon: $Poi[$PoiIndex]{'Poi_Lon'}", "A");

    my ($x, $y) = &MapGps2XY($Poi[$PoiIndex]{'Poi_Lat'}, $Poi[$PoiIndex]{'Poi_Lon'});
    $Poi[$PoiIndex]{'Poi_x'} = $x;
    $Poi[$PoiIndex]{'Poi_y'} = $y;

    $map_canvas->coords ('POI', $x-24, $y-48 );
    $map_canvas->raise('POI', 'Track-Antenna');
    }

# Show Grid on canvas
sub GridShow()
    {
    &Log ("Show Grid", "A");

    my $Dist  = &Default ($Cfg->{map}->{'GridDist'},  50);
    my $Color = &Default ($Cfg->{map}->{'GridColor'}, "#909090");

    my $xmin = 0;
    my $ymin = 0;
    my $xmax = $MapSizeX;
    my $ymax = $MapSizeY;


    my $PhiRef = &MapAngel();  
    my ($Lat1, $Lon1) = &MapXY2Gps($xmin, $ymin);
    my ($Lat2, $Lon2) = &MapGpsAt($Lat1, $Lon1, $Dist, $PhiRef);
    my ($x, $y) = &MapGps2XY($Lat2, $Lon2);
    my $dpix = int ($x - $xmin + 0.5);

    lock (%MkOsd);              # until end of block
    my $x0 = $MapSizeX / 2;
    my $y0 = $MapSizeY / 2;
    if ( &HomePosIsValid() )
        {
        ($x0, $y0) = &MapGps2XY ($MkOsd{'HomePos_Lat'}, $MkOsd{'HomePos_Lon'});
        }

    for ($x = $xmin + $x0 % $dpix; $x < $xmax; $x +=$dpix)
        {
        $map_canvas->createLine ( $x, $ymin, $x, $ymax,
                                  '-tags' => 'Map-Grid',
                                  '-arrow' => 'none',
                                  '-fill' => $Color,
                                  '-width' => 1,
                                );
        }

    for ($y = $ymin + $y0 % $dpix; $y < $ymax; $y +=$dpix)
        {
        $map_canvas->createLine ( $xmin, $y, $xmax, $y,
                                  '-tags' => 'Map-Grid',
                                  '-arrow' => 'none',
                                  '-fill' => $Color,
                                  '-width' => 1,
                                );
        }

    # Beschriftung x
    for ( $x = xmin + $x0 % $dpix; $x < $xmax; $x += $dpix)
        {
        my $ScaleX = int (($x - $x0) / $dpix * $Dist + 0.5);
        if ( $ScaleX < 0 )
            {
            $ScaleX = int (($x - $x0) / $dpix * $Dist - 0.5);
            }
        $map_canvas->createText ( $x - 2, $y0 - 8,
                                  '-tags' => 'Map-Grid',
                                  '-text' => sprintf ("%d", $ScaleX),
                                  '-font' => '-*-Arial-Bold-R-Normal--*-150-*',
                                  '-fill' => $Color,
                                  '-anchor' => 'e',
                                );
        }
    # Beschriftung y
    for ( $y = ymin + $y0 % $dpix; $y < $ymax; $y += $dpix)
        {
        my $ScaleY = int (($y - $y0) / $dpix * $Dist + 0.5);
        if ( $ScaleY < 0 )
            {
            $ScaleY = int (($y - $y0) / $dpix * $Dist - 0.5);
            }
        $map_canvas->createText ( $x0 + 4, $y - 8,
                                  '-tags' => 'Map-Grid',
                                  '-text' => sprintf ("%d", $ScaleY * -1),
                                  '-font' => '-*-Arial-Bold-R-Normal--*-150-*',
                                  '-fill' => $Color,
                                  '-anchor' => 'w',
                                );
        }

    $map_canvas->raise('Map-Grid', 'Map');
    }


# Hide Grid on canvas
sub GridHide()
    {
    &Log ("Hide Grid", "A");

    $map_canvas->delete('Map-Grid'); 
    }


# Show Crosshair on canvas
sub CrosshairShow()
    {
    my ($Lat, $Lon, $Heading) = @_;

    my ($x, $y) = &MapGps2XY ($Lat, $Lon);
    if ( ($x != $LastCrosshairX  and  $y != $LastCroshairY)  or
         ($Heading != $LastCrosshairHeading) )
        {
        # Only update, if coords changed - CPU consuming!
        $map_canvas->coords ('Map-Crosshair-X', 0, $y, $MapSizeX, $y);
        $map_canvas->coords ('Map-Crosshair-Y', $x, 0, $x, $MapSizeY);

        my $Angle = deg2rad (&MapAngel() - $Heading);
        my $x2 = $x + $MapSizeX * cos($Angle);
        my $y2 = $y - $MapSizeX * sin($Angle);
        $map_canvas->coords ('Map-Crosshair-Heading', $x, $y, $x2 , $y2 );

        $map_canvas->raise('Map-Crosshair', 'Target');
        }

    $LastCrosshairX = $x;
    $LastCrosshairY = $y;
    $LastCrosshairHeading = $Heading;
    }


# Hide Crosshair on canvas
sub CrosshairHide()
    {
    if ( $CrosshairTimerCnt > 0 )
        {
        $map_canvas->lower('Map-Crosshair', 'Map');  # hide below map
        $LastCrosshairX = -1;
        $LastCrosshairY = -1;
        }

    $CrosshairTimerCnt = -10;
    }


#
# System Messages
#

# Init Messages for a Subsystem/timer
sub MkMessageInit ()
    {
    my ($Id) = @_;

    $MkMessages{$Id} = [];
    }


# Register message
sub MkMessage ()
    {
    my ($Message, $Id) = @_;

    push @{$MkMessages{$Id}}, $Message;
    }


# show registered messages
sub MkMessageShow()
    {
    my @Messages;
    my $MsgLines = 0;
    my $MaxMsgLen = 0;

    # Collect Messages of each category
    foreach my $Id (keys %MkMessages)
        {
        foreach $i ( 0 .. $#{$MkMessages{$Id}} )
            {
            my $Msg = $MkMessages{$Id}[$i];
            push @Messages, $Msg;

            $MsgLines ++;
            
            my $Len = length $Msg;
            if ( $Len > $MaxMsgLen )
                {
                $MaxMsgLen = $Len;
                }
            }
        }

    $map_canvas->delete('Message-Balloon');  # delete old Balloon

    if ( $MsgLines > 0 )
        {
        # draw Balloon
        my $MkPos_x = $System{'CurPos_x'};
        my $MkPos_y = $System{'CurPos_y'};

        my @MsgBalloon = ( $MkPos_x ,                       $MkPos_y,
                           $MkPos_x + 30 ,                  $MkPos_y + 40,
                           $MkPos_x + 30 + $MaxMsgLen * 11, $MkPos_y + 40,
                           $MkPos_x + 30 + $MaxMsgLen * 11, $MkPos_y + 44 + $MsgLines * 20,
                           $MkPos_x + 20,                   $MkPos_y + 44 + $MsgLines * 20,
                           $MkPos_x + 20,                   $MkPos_y + 40,
                           $MkPos_x,                        $MkPos_y,
                          );

        $map_canvas->createPolygon( @MsgBalloon,
                                    '-tags' => ['Message-Balloon', 'Message-BalloonBubble'],
                                    '-fill' => 'yellow',
                                    '-outline' => 'yellow',
                                    '-width' => 1,
                                  );
        # draw Messages
        my $MsgLine = 1;
        foreach my $Msg (@Messages)
            {
            $map_canvas->createText ( $MkPos_x + 25, $MkPos_y + 32 + $MsgLine * 20 ,
                                      '-tags' => ['Message-Balloon', 'Message-BalloonText'],
                                      '-text' => $Msg,
                                      '-font' => '-*-Arial-Bold-R-Normal--*-200-*',
                                      '-fill' => 'blue',
                                      '-anchor' => 'w',
                                        );
            $MsgLine ++;
            }


        $map_canvas->lower('Message-Balloon', 'MK-Arrow');
        }

    }


# Show Balloon, when aproaching Target
sub TargetMessageShow ()
    {
    $map_canvas->delete('Target-Balloon');  # delete old Balloon

    if ( $OperationMode ne "Free" and $MkOsd{'TargetPos_Stat'} == 1  and $MkOsd{'TargetPosDev_Dist'} /10 < 25 )
        {
        my $BalloonLines = 0;
        $ColorBalloon = "blue";
        my ($T_x, $T_y) = &MapGps2XY($MkOsd{'TargetPos_Lat'}, $MkOsd{'TargetPos_Lon'});
        my $Wp = $Waypoints[$MkOsd{'WaypointIndex'}];

        # Holdtime Wp-Player Mode
        if ( $WpPlayerHoldtime >= 0  and  $PlayerWptKmlMode  eq "WPT" )
            {
            # Holdtime
            $ColorBalloon = 'red';
            my $HoldTime = sprintf ("%5s %3d s", "HLD:", int ($WpPlayerHoldtime / 2  + 0.5) );
            $map_canvas->createText ( $T_x + 25, $T_y - 40,
                                      '-tags' => ['Target-Balloon', 'Target-BalloonText'],
                                      '-text' => $HoldTime,
                                      '-font' => '-*-Arial-Bold-R-Normal--*-200-*',
                                      '-fill' => $ColorBalloon,
                                      '-anchor' => 'w',
                                    );
            $BalloonLines ++;
            }

        # Holdtime WPT-Mode
        if ( &MkTargetReached()  and  $OperationMode eq "WPT" )
            {
            # Holdtime from MK
            $ColorBalloon = 'red';
            my $HoldTime = sprintf ("%5s %3d s", "HLD:", int ($MkOsd{'TargetHoldTime'} + 0.5) );
            $map_canvas->createText ( $T_x + 25, $T_y - 40,
                                      '-tags' => ['Target-Balloon', 'Target-BalloonText'],
                                      '-text' => $HoldTime,
                                      '-font' => '-*-Arial-Bold-R-Normal--*-200-*',
                                      '-fill' => $ColorBalloon,
                                      '-anchor' => 'w',
                                    );
            $BalloonLines ++;
            }

        # Tolerance Radius Player Mode
        if ( &MkIsWptMode()  and  $OperationMode eq "Play" and $PlayerWptKmlMode eq "WPT" )
            {
            my $WpTolerance  = sprintf ("%5s %3d m", "TOL:", $Wp->{'ToleranceRadius'});
            $map_canvas->createText ( $T_x + 25, $T_y - 60,
                                      '-tags' => ['Target-Balloon', 'Target-BalloonText'],
                                      '-text' => $WpTolerance,
                                      '-font' => '-*-Arial-Bold-R-Normal--*-200-*',
                                      '-fill' => $ColorBalloon,
                                      '-anchor' => 'w',
                                    );
            $BalloonLines ++;
            }

        # Tolerance WPT-Mode
        if ( &MkIsWptMode()  and  $OperationMode eq "WPT" )
            {
            my $WpTolerance  = sprintf ("%5s %3d m", "TOL:", $Wp->{'ToleranceRadius'} );
            $map_canvas->createText ( $T_x + 25, $T_y - 60,
                                      '-tags' => ['Target-Balloon', 'Target-BalloonText'],
                                      '-text' => $WpTolerance,
                                      '-font' => '-*-Arial-Bold-R-Normal--*-200-*',
                                      '-fill' => $ColorBalloon,
                                      '-anchor' => 'w',
                                    );
            $BalloonLines ++;
            }

        # Distance to Target
        my $Dist = int ($MkOsd{'TargetPosDev_Dist'} /10 + 0.5);
        $map_canvas->createText ( $T_x + 25, $T_y - 80,
                                  '-tags' => ['Target-Balloon', 'Target-BalloonText'],
                                  '-text' => sprintf ("%5s %3d m", "DST:", $Dist) ,
                                  '-font' => '-*-Arial-Bold-R-Normal--*-200-*',
                                  '-fill' => $ColorBalloon,
                                  '-anchor' => 'w',
                                );
        $BalloonLines ++;

        if ( $BalloonLines >= 1 )
            {
            # draw Balloon
            my @TargetBalloon = ( $T_x ,      $T_y,
                                  $T_x + 30,  $T_y - (3 - $BalloonLines) * 20 -27,
                                  $T_x + 150, $T_y - (3 - $BalloonLines) * 20 -27 ,
                                  $T_x + 150, $T_y - 93,
                                  $T_x + 20,  $T_y - 93,
                                  $T_x + 20,  $T_y - (3 - $BalloonLines) * 20 -27,
                                  $T_x,       $T_y,
                                );

            $map_canvas->createPolygon( @TargetBalloon,
                                        '-tags' => ['Target-Balloon', 'Target-BalloonBubble'],
                                        '-fill' => 'lightgray',
                                        '-outline' => 'yellow',
                                        '-width' => 1,
                                       );
            }


        $map_canvas->lower('Target-Balloon', 'MK-Home-Line');
        $map_canvas->lower('Target-BalloonBubble', 'Target-BalloonText');
        }
    }


#
# Airfield border
#

# Are two segments A(a1/a2), B(b1/b2) and C(c1/c2), D(d1/d2) crossing ?
sub SegmentCross()
    {
    my ( $a1, $a2, $b1, $b2, $c1, $c2, $d1, $d2) = @_;

    # segment C/D ist vertical, avoid div/0
    if ( $c1 == $d1 )
        {
        $d1 += 0.00001;
        }

    my $n = ($b1 - $a1) * ($d2 - $c2) - ($b2 - $a2) * ($d1 - $c1);
    if ( $n == 0.0 )
        {
        # AB und CD sind parallel
        return 0;
        }

    my $s = ( ($c1 - $a1) * ($d2 - $c2) - ($c2 - $a2) * ($d1 - $c1) ) / $n;
    my $t = ( $a1 - $c1 + $s * ($b1 - $a1) ) / ( $d1 - $c1 );
    if ( $s >= 0.0  and  $s <= 1.0  and  $t >= 0.0  and  $t <= 1.0 )
        {
        # beide Strecken kreuzen sich

        # Schnittpunkt: s_x, s_y
        my $s_x = $a1 + $s * ( $b1 - $a1 );
        my $s_y = $a2 + $s * ( $b2 - $a2 );

        return 1;
        }

    # beide Strecken kreuzen sich nicht
    return 0;
    }


# How often does a segment A(a1,a2), B(b1,b2) cross the polygon?
sub SegmentPolygonCross()
    {
    my ( $a1, $a2, $b1, $b2, $Polygon) = @_;

    my $Cross = 0;
    my $PolyCnt = scalar @{$Polygon};
    my $PolyPointCnt = $PolyCnt / 2;

    my $i = 0;
    for ( $p=0; $p < $PolyPointCnt; $p++ )
        {
        my $c1 = ${$Polygon}[$i++];
        my $c2 = ${$Polygon}[$i++];

        if ( $i >= $PolyCnt ) { $i = 0; }

        my $d1 = ${$Polygon}[$i];
        my $d2 = ${$Polygon}[$i+1];

        # map calibration offsets
        $c1 -= $Map{'Offset_x'};
        $c2 += $Map{'Offset_y'};
        $d1 -= $Map{'Offset_x'};
        $d2 += $Map{'Offset_y'};    

        if ( &SegmentCross($a1, $a2, $b1, $b2, $c1, $c2, $d1, $d2) )
            {
            $Cross ++;
            }
        }

    return $Cross;
    }


# Is point A inside airfield border?
sub IsInsideBorder()
    {
    my ($a1, $a2) = @_;

    # need at least 3 border points
    if ( scalar @{$Map{'Border'}} < 3*2 )
        {
        # no border defined, always inside
        return 1;
        }

    my $Cross = &SegmentPolygonCross (-10, -10, $a1, $a2, @Map{'Border'} );

    # Ungerade Anzahl Kreuzungen: Inside
    return ( $Cross % 2 );
    }



# Is segment A, B crossing the airfield border?
sub IsCrossingBorder()
    {
    my ($a1, $a2, $b1, $b2) = @_;

    # need at least 3 boorder points
    if ( scalar @{$Map{'Border'}} < 3*2 )
        {
        # no border defined, always not crossing
        return 0;
        }

    my $Cross = &SegmentPolygonCross ($a1, $a2, $b1, $b2, @Map{'Border'} );

    return ( $Cross > 0 );
    }


# How often is segment A, B crossing the airfield border?
sub CrossingBorderCount()
    {
    my ($a1, $a2, $b1, $b2) = @_;

    # need at least 3 border points
    if ( scalar @{$Map{'Border'}} < 3*2 )
        {
        # no border defined, not crossing
        return 0;
        }

    my $Cross = &SegmentPolygonCross ($a1, $a2, $b1, $b2, @Map{'Border'} );

    return ( $Cross );
    }


# check, if Target is reachable my MK
sub IsTargetReachable()
    {
    my ($T_x, $T_y) = @_;

    my $MkIsInside = &IsInsideBorder($System{'CurPos_x'}, $System{'CurPos_y'});
    my $TargetIsInside = &IsInsideBorder($T_x, $T_y);
    my $MkTargetCrossingCount = &CrossingBorderCount($System{'CurPos_x'}, $System{'CurPos_y'}, $T_x, $T_y);

    if ( ($MkIsInside  and  $MkTargetCrossingCount == 0 )  or
         (! $MkIsInside  and  $TargetIsInside  and  $MkTargetCrossingCount == 1) )
        {
        # Target is reachable
        return 1;
        }

    # Target is not reachable
    return 0;
    }


#
# Closed natural cubic spline
#
# Derived from: http://www.cse.unsw.edu.au/~lambert/splines
# @author Tim Lambert
#
# calculates the closed natural cubic spline that interpolates x[0], x[1], ... x[n]
# The first segment is returned as
# C[0].a + C[0].b*u + C[0].c*u^2 + C[0].d*u^3 0<=u<1
# the other segments are in C[1], C[2], ...  C[n]

sub SplineCalcNaturalCubic()
    {
    my (@x) = @_;

    # We solve the equation
    #   [4 1      1] [D[0]]   [3(x[1] - x[n])  ]
    #   |1 4 1     | |D[1]|   |3(x[2] - x[0])  |
    #   |  1 4 1   | | .  | = |      .         |
    #   |    ..... | | .  |   |      .         |
    #   |     1 4 1| | .  |   |3(x[n] - x[n-2])|
    #   [1      1 4] [D[n]]   [3(x[0] - x[n-1])]
       
    #   by decomposing the matrix into upper triangular and lower matrices
    #   and then back sustitution.  See Spath "Spline Algorithms for Curves
    #   and Surfaces" pp 19--21. The D[i] are the derivatives at the knots.

    my @C, @D, @v, @w, @y, $F, $G, $H, $k, $z;
    my $n = $#x;

    $w[1] = $v[1] = $z = 1.0/4.0;
    $y[0] = $z * 3 * ($x[1] - $x[$n]);
    $H = 4;
    $F = 3 * ($x[0] - $x[$n-1]);
    $G = 1;
    for ( $k = 1; $k < $n; $k++)
        {
        $v[$k+1] = $z = 1/(4 - $v[$k]);
        $w[$k+1] = -$z * $w[$k];
        $y[$k] = $z * (3*($x[$k+1]-$x[$k-1]) - $y[$k-1]);
        $H = $H - $G * $w[$k];
        $F = $F - $G * $y[$k-1];
        $G = -$v[$k] * $G;
        }
    $H = $H - ($G+1)*($v[$n]+$w[$n]);
    $y[$n] = $F - ($G+1)*$y[$n-1];
    
    $D[$n] = $y[$n]/$H;
    $D[$n-1] = $y[$n-1] - ($v[$n]+$w[$n])*$D[$n];
    for ( $k = $n-2; $k >= 0; $k--)
        {
        $D[$k] = $y[$k] - $v[$k+1]*$D[$k+1] - $w[$k+1]*$D[$n];
        }

    # now compute the coefficients of the cubics
    for ( $k = 0; $k < $n; $k++)
        {
        $C[$k]{'a'} = $x[$k];
        $C[$k]{'b'} = $D[$k];
        $C[$k]{'c'} = 3*($x[$k+1] - $x[$k]) - 2*$D[$k] - $D[$k+1];
        $C[$k]{'d'} = 2*($x[$k] - $x[$k+1]) + $D[$k] + $D[$k+1];
        }

    $C[$n]{'a'} = $x[$n];
    $C[$n]{'b'} = $D[$n];
    $C[$n]{'c'} = 3*($x[0] - $x[$n]) - 2*$D[$n] - $D[0];
    $C[$n]{'d'} = 2*($x[$n] - $x[0]) + $D[$n] + $D[0];

    return @C;
    }

# Eval interpolated spline value
sub SplineEval()
    {
    my ($u, %C) = @_;
 
    $a = $C{'a'};
    $b = $C{'b'};
    $c = $C{'c'};
    $d = $C{'d'};

    return ((($d*$u) + $c)*$u + $b)*$u + $a;
    }

# Get pixel position on WP-Spline after elapesd time
sub SplinePosFromTime()
    {
    my ($Time) = @_;

    if ( scalar(@Waypoints) >= 2 )
        {
        # eval x/y target according to Time and Speed
        my $TrackTime = 0;      # time/s needed to fly spline-track
        my $WpIndex = 0;
        my $Steps = 16;

        while (1)
            {
            my $Wp = $Waypoints[$WpIndex];
            my $x0 = $Wp->{'MapX'};
            my $y0 = $Wp->{'MapY'};

            my $WpIndex1 = $WpIndex + 1;
            if ( $WpIndex1 >= scalar (@Waypoints) )
                {
                $WpIndex1 = 0;
                }
            my $Wp = $Waypoints[$WpIndex1];
            my $Wp1_Speed = &Default ($Wp->{'Speed'}, $Cfg->{'waypoint'}->{'DefaultSpeed'}, 10);  # km/h
            $Wp1_Speed /= 3.6;                     # m/s

            for (my $j = 1; $j <= $Steps; $j++)
                {
                my $u = $j / $Steps;
                my $x = &SplineEval($u, %{$SplineCoeffX[$WpIndex]});
                my $y = &SplineEval($u, %{$SplineCoeffY[$WpIndex]});

                my $SegmentLen = &MapXYTo($x0, $y0, $x, $y);
                my $SegmentTime = $SegmentLen / $Wp1_Speed;
                $TrackTime += $SegmentTime;

                if ( $TrackTime >= $SpdPlayerTime )
                    {
                    # position found
                    # linear interpolation between spline polygon points
                    my $DiffTime = $TrackTime - $SpdPlayerTime;
                    my $Ratio = 1 - $DiffTime / $SegmentTime;
                    my $P_x = $x0 + ($x - $x0) * $Ratio;
                    my $P_y = $y0 + ($y - $y0) * $Ratio;

                    return ($P_x, $P_y, $WpIndex1);
                    }

                $x0 = $x;
                $y0 = $y;
                }

            # next WPT
            $WpIndex ++;
            if ( $WpIndex >= scalar (@Waypoints) )
                {
                # next loop: reset counter
                $WpIndex = 0;
                $TrackTime = 0;
                $SpdPlayerTime = 0;
                }
            }
        }

    # "position not found" return
    return ("", "", "");
    }

# save current position as starting point for Altitude control
sub AltCurPosSave()
    {
    my ($WpIndex) = @_;

    if ( $WpIndex ne "" )
        {
        # take WP as start position
        my $Wp = $Waypoints[$WpIndex];
        $System{'AltStartPos_x'}   = $Wp->{'MapX'};
        $System{'AltStartPos_y'}   = $Wp->{'MapY'};
        $System{'AltStartPos_Alt'} = $Wp->{'Pos_Alt'};
        }
    else
        {
        # take current position as start position
        $System{'AltStartPos_x'} = $System{'CurPos_x'};
        $System{'AltStartPos_y'} = $System{'CurPos_y'};
        $System{'AltStartPos_Alt'} = &Altitude();
        }
    }


# Altitude of way from Startpoint $System{'AltStartPos_Xy'} to destination WP
sub AltCalc()
    {
    # A: Start Point; B: Dest Point; C: Current Position
    my $Ax = $System{'AltStartPos_x'};
    my $Ay = $System{'AltStartPos_y'};

    my $DestWp = $System{'AltWpIndex'};
    my $Wp = $Waypoints[$DestWp];
    my $Bx = $Wp->{'MapX'};
    my $By = $Wp->{'MapY'};

    my $Cx = $System{'CurPos_x'};
    my $Cy = $System{'CurPos_y'};

    my $c = sqrt( ($Cx - $Ax)**2 + ($Cy - $Ay)**2 );
    my $d = sqrt( ($Cx - $Bx)**2 + ($Cy - $By)**2 );
    my $e = sqrt( ($Bx - $Ax)**2 + ($By - $Ay)**2 );

    my $a = ($c*$c - $d*$d + $e*$e) / (2 * $e);

    my $AltA = $System{'AltStartPos_Alt'};
    my $AltB = &Default ($Wp->{'Pos_Alt'}, $Cfg->{'waypoint'}->{'DefaultAltitude'}, 0);
    my $Alt = $a/$e * ($AltB - $AltA) + $AltA;

    # Alt Limits between A and B
    if ($AltA < $AltB  and  $Alt < $AltA ) { $Alt = $AltA; };
    if ($AltA > $AltB  and  $Alt > $AltA ) { $Alt = $AltA; };
    if ($AltB < $AltA  and  $Alt < $AltB ) { $Alt = $AltB; };
    if ($AltB > $AltA  and  $Alt > $AltB ) { $Alt = $AltB; };

    return $Alt;
    }


# norm bearing to 1..360 degree
sub NormBearing()
    {
    my ($Bearing) = @_;

    while ( $Bearing > 360 ) { $Bearing -= 360; }
    while ( $Bearing < 0 )   { $Bearing += 360; }
    if ( $Bearing < 1 )
        {
        $Bearing = 360;
        }

    return $Bearing;
    }


# compass bearing
sub CompassCalc()
    {
    my ( $Bearing) = @_;

    my $Setpoint = 0;

    # <0: POI index in WP-List
    if ( $Bearing < 0 )
        {
        # not supported
        $Setpoint = 0;
        }

    # 0: no change
    elsif ( $Bearing < 1 )
        {
        # No direction change
        $Setpoint = 0;
        }

    # 1..360: Absolute direction
    elsif ( $Bearing <= 360 )
        {
        $Setpoint = $Bearing;
        }

    # 361: Invalid
    elsif ( $Bearing <= 361 )
        {
        # No direction change
        $Setpoint = 0;
        }

    # >= 362: calc direction to target
    else
        {
        $Setpoint = $System{'TargetBearing'};
        if ( $System{'TargetDist'} <= 5 )
            {
            # no direction change, if target is too close
            $Setpoint = 0;
            }
        } 

    return $Setpoint;
    }


1;

__END__
