#!/usr/bin/perl
#!/usr/bin/perl -d:ptkdb

###############################################################################
#
# libmktimer.pl -  MK Mission Cockpit - Timer for GUI Frontend
#
# Copyright (C) 2009  Rainer Walther  (rainerwalther-mail@web.de)
#
# Creative Commons Lizenz mit den Zusaetzen (by, nc, sa)
#
# Es ist Ihnen gestattet: 
#     * das Werk vervielfltigen, verbreiten und ffentlich zugnglich machen
#     * Abwandlungen bzw. Bearbeitungen des Inhaltes anfertigen
# 
# Zu den folgenden Bedingungen:
#     * Namensnennung.
#       Sie mssen den Namen des Autors/Rechteinhabers in der von ihm festgelegten Weise nennen.
#     * Keine kommerzielle Nutzung.
#       Dieses Werk darf nicht fr kommerzielle Zwecke verwendet werden.
#     * Weitergabe unter gleichen Bedingungen.
#       Wenn Sie den lizenzierten Inhalt bearbeiten oder in anderer Weise umgestalten,
#       verndern oder als Grundlage fr einen anderen Inhalt verwenden,
#       drfen Sie den neu entstandenen Inhalt nur unter Verwendung von Lizenzbedingungen
#       weitergeben, die mit denen dieses Lizenzvertrages identisch oder vergleichbar sind.
# 
# Im Falle einer Verbreitung mssen Sie anderen die Lizenzbedingungen, unter welche dieses
# Werk fllt, mitteilen. Am Einfachsten ist es, einen Link auf diese Seite einzubinden.
# 
# Jede der vorgenannten Bedingungen kann aufgehoben werden, sofern Sie die Einwilligung
# des Rechteinhabers dazu erhalten.
# 
# Diese Lizenz lsst die Urheberpersnlichkeitsrechte unberhrt.
# 
# Weitere Details zur Lizenzbestimmung gibt es hier:
#   Kurzform: http://creativecommons.org/licenses/by-nc-sa/3.0/de/
#   Komplett: http://creativecommons.org/licenses/by-nc-sa/3.0/de/legalcode
#
###############################################################################
# 2009-08-09 0.2.5 rw Timer moved from mkcockpit.pl
#                     Optional Player home-pos in map configuration
# 2009-08-23 0.2.6 rw Tracking-Antenna Icon
#                     POI heading control
# 2009-10-11 0.2.7 rw Tracker control changed
# 2010-02-10 0.4.0 rw Altitude average changed
#                     Event Engine timer
#                     Serial channel timer
#                     External Control rimer
#                     Crosshair Timer
#                     5s timer changed to 3s
#                     Current, UsedCapacity, Power added
# 2010-02-13 0.4.1 rw bugfix JoystickButton min-value
# 2010-02-14 0.4.2 rw Input chontrol parser added
#                     Request LCD Stick/Poti
#                     Renew OSD-Abo every 1.5s
#                     NC-Debug abo frequency 100ms -> 200ms
#                     Low Battery check from MKFLAGS
# 2010-07-01 0.5.0 rw TTS system messages closer to current situation
# 2010-07-29 0.5.1 rw Start Tracker at motor start (was calibration)
# 2010-10-16 0.5.4 rw Spline speed controlled mode
# 2010-11-01 0.6.0 rw Target not reachable bugfix
#                     NC 0.21
# 2011-03-06 0.7.0 rw Switch UART to NC
#                     Read MK Settings
#                     NC POI/Gier/Altitude control
#                     split 100ms timer to 200ms timer for less CPU load
# 2011-04-17 0.7.1 rw Use Altitude for ODO calculation
#                     Logging added
# 2011-10-15 0.8.1 rw Disable Gier after POI off in FREE or PH
#                     Move Pause-Position manually with manual RC
#                     Switch player to Pause, if manual RC in play mode
#                     Send POI only if player is not stopped
# 2011-12-24 0.9.0 rw NC 0.28
#                     Requst Motor data
#
###############################################################################

$Version{'libmktimer.pl'}  = "0.9.0 - 2011-12-24";

use Math::Trig;
use Time::HiRes qw(gettimeofday);   # http://search.cpan.org/~jhi/Time-HiRes-1.9719/HiRes.pm


#
# Timer: 1.5s
#
my $NcConnectionCounter = 0;

$main->repeat (1500, sub 
    {
    if ( ! $MkSendWp )
        {
        &MkSwitchToNC();

        # Abfragefrequenz OSD und Debug regelmig neu einstellen
        $MkSendQueue->enqueue( "o", "$AddrNC", pack ("C", 10) );   # Frequenz OSD Datensatz, * 10ms
        $MkSendQueue->enqueue( "d", "$AddrNC", pack ("C", 20) );   # Frequenz MK Debug Datensatz, * 10ms
        $MkSendQueue->enqueue( "k", "$AddrNC", pack ("C", 33) );   # Request Motor Data, * 10ms


        $MkSendQueue->enqueue( "v", "$AddrNC", "");                # Version
        $MkSendQueue->enqueue( "e", "$AddrNC", "");                # Error Text Request
        $MkSendQueue->enqueue( "l", "$AddrNC", pack ("C", 1) );    # Request LCD-Screen 1 - Type of GPS fix

        $MkSendQueue->enqueue( "z", "$AddrNC", pack ("S", $NcConnectionCounter) );
        $NcConnectionCounter ++;
        }

    lock (%MkOsd);              # until end of block

    # Draw Operation Radius Border
    $map_canvas->delete('Map-Border-OperatingRadius');
    if ( &HomePosIsValid() )
        {
        my $Radius = $MkOsd{'OperatingRadius'};
        my $H_Lat = $MkOsd{'HomePos_Lat'};
        my $H_Lon = $MkOsd{'HomePos_Lon'};
        my $Angel = &MapAngel();

        my ($T_Lat, $T_Lon) = &MapGpsAt ($H_Lat, $H_Lon, $Radius, $Angel -90); 
        my ($R_Lat, $R_Lon) = &MapGpsAt ($H_Lat, $H_Lon, $Radius, $Angel); 
        my ($B_Lat, $B_Lon) = &MapGpsAt ($H_Lat, $H_Lon, $Radius, $Angel + 90); 
        my ($L_Lat, $L_Lon) = &MapGpsAt ($H_Lat, $H_Lon, $Radius, $Angel + 180); 

        my ($T_x, $T_y) = &MapGps2XY ($T_Lat, $T_Lon);
        my ($R_x, $R_y) = &MapGps2XY ($R_Lat, $R_Lon);
        my ($B_x, $B_y) = &MapGps2XY ($B_Lat, $B_Lon);
        my ($L_x, $L_y) = &MapGps2XY ($L_Lat, $L_Lon);

        $map_canvas->createArc ( $L_x, $B_y, $R_x, $T_y,
                                 '-tags' => 'Map-Border-OperatingRadius',
                                 '-extent' => '359',
                                 '-start' => '0', 
                                 '-style' => 'chord',
                                 '-outline' => $Cfg->{'mkcockpit'}->{'ColorAirfield'},
                                 '-width' => '1',
                               );
        $map_canvas->raise('Map-Border-OperatingRadius', 'Map');  # Border above Map
        }

    });


#
# Timer: 0.1s - update Map Overlay - fast objects
#
$frame_map_top->repeat (100, sub
    {
    # performance profiling
    my ($t0_s, $t0_us) = gettimeofday;

    # Clear old messages from this timer
    &MkMessageInit ("Timer-MapOverlay");

    lock (%MkOsd);              # until end of block
    lock (%MkNcDebug);          # until end of block

    # Aktuell gltige Karte
    %Map = %{$Maps{'Current'}};

    if ( &MkOsdIsValid() )
        {
        # Gueltige OSD Daten

        if ( &CurPosIsValid()  and  &HomePosIsValid() )
            {
            # ausreichender GPS Empfang

            # get x,y map coords of current position
            my ($C_x, $C_y, $C_Angel) = &MapGps2XY($MkOsd{'CurPos_Lat'}, $MkOsd{'CurPos_Lon'}, $MkOsd{'CompassHeading'});

            $System{'CurPos_x'} = $C_x;
            $System{'CurPos_y'} = $C_y;

            # rotate MK arrow
            my $dy = sin (deg2rad $C_Angel) * ($MapMkLen/2);
            my $dx = cos (deg2rad $C_Angel) * ($MapMkLen/2);
            my $x0 = $C_x - $dx;
            my $y0 = $C_y - $dy;
            my $x1 = $C_x + $dx;
            my $y1 = $C_y + $dy;
            $map_canvas->coords ('MK-Arrow', $x0, $y0, $x1, $y1);

            # Update speed vector
            my $MapAngel = &MapAngel();   # North to Map-Horizont
            $PhiMapSpeed = $MkOsd{'Heading'} - $MapAngel;

            # 555 cm/s ~ 20 km/h -> Zeigerlnge = $MkSpeedLen bei 20 km/h
            my $dy = sin (deg2rad $PhiMapSpeed) * $MapMkSpeedLen * $MkOsd{'GroundSpeed'} / 555;
            my $dx = cos (deg2rad $PhiMapSpeed) * $MapMkSpeedLen * $MkOsd{'GroundSpeed'} / 555;
            my $x0 = $C_x;
            my $y0 = $C_y;
            my $x1 = $C_x + $dx;
            my $y1 = $C_y + $dy;
            $map_canvas->coords ('MK-Speed', $x0, $y0, $x1, $y1);
 
            # Update Line between Home and MK
            my ($H_x, $H_y) = &MapGps2XY($MkOsd{'HomePos_Lat'}, $MkOsd{'HomePos_Lon'});
            $map_canvas->coords ('MK-Home-Line', $H_x, $H_y, $C_x, $C_y);

            # Update Distance between Home and MK
            my ($Dist, $Bearing) = &MapGpsTo($MkOsd{'CurPos_Lat'}, $MkOsd{'CurPos_Lon'},
		                                    $MkOsd{'HomePos_Lat'}, $MkOsd{'HomePos_Lon'} );
            my $x = ($C_x - $H_x) / 2 + $H_x + 8;
            my $y = ($C_y - $H_y) / 2 + $H_y + 8;
            $map_canvas->coords ('MK-Home-Dist', $x, $y);
            $map_canvas->itemconfigure ('MK-Home-Dist',
                                        '-text' => sprintf ("%4d m", int ($Dist + 0.5) ),
                                       );
            $System{'HomePos_y'}   = $H_x;
            $System{'HomePos_x'}   = $H_y;
            $System{'HomeDist'}    = $Dist;
            $System{'HomeBearing'} = $Bearing;

            if ( &TargetIsValid() )
                {
                # Valid Target

                # Update Line between Target and MK
                my ($T_x, $T_y) = &MapGps2XY($MkOsd{'TargetPos_Lat'}, $MkOsd{'TargetPos_Lon'});
                $map_canvas->coords ('MK-Target-Line', $C_x, $C_y, $T_x, $T_y);

                # Update Distance between Target and MK
                my ($Dist, $Bearing) = &MapGpsTo($MkOsd{'CurPos_Lat'}, $MkOsd{'CurPos_Lon'},
		                                         $MkOsd{'TargetPos_Lat'}, $MkOsd{'TargetPos_Lon'} );
                $System{'TargetPos_x'}   = $T_x;
                $System{'TargetPos_y'}   = $T_y;
                $System{'TargetDist'}    = $Dist;
                $System{'TargetBearing'} = $Bearing;

                if ( $Dist >= 25 )   
                    {
                    my $x = ($C_x - $T_x) / 2 + $T_x - 8;
                    my $y = ($C_y - $T_y) / 2 + $T_y + 8;
                    $map_canvas->coords ('MK-Target-Dist', $x, $y);
                    $map_canvas->itemconfigure ('MK-Target-Dist',
                                                '-text' => sprintf ("%4d m", int ($Dist + 0.5) ),
                                               );
                    }
                else
                    {
                    # Don't show distance < 25m
                    $map_canvas->coords ('MK-Target-Dist', 0, -100);
                    }
                }
            else
                {
                #
                # No valid Target
                #

                # Move target line out of sight/canvas
                $map_canvas->coords ('MK-Target-Line', 0, -100, 0, -100);
                $map_canvas->coords ('MK-Target-Dist', 0, -100);
                }

            if ( $PoiMode )
                {
                # Update Line between MK and POI
                my ($P_x, $P_y) = &MapGps2XY($Poi[$PoiIndex]{'Poi_Lat'}, $Poi[$PoiIndex]{'Poi_Lon'});
                $map_canvas->coords ('MK-POI-Line', $P_x, $P_y, $C_x, $C_y);

                $System{'PoiPos_x'} = $P_x;
                $System{'PoiPos_y'} = $P_y;
                }
            else
                {
                $map_canvas->coords ('MK-POI-Line', 0, -200, 0, -200);
                }
            }
        }

    # Wp-Number input from keyboard
    $KbTimer++;
    if ( $CbPlayerKey ne "" )
        {
        # Key pressed
        $KbNum = "$KbNum" . "$CbPlayerKey";

        $CbPlayerKey = "";
        $KbTimer = 0;
        }
    if ( $KbTimer > 7  and $KbNum ne "" )
        {
        # number complete after 700ms
        my $Number = sprintf ("%d", $KbNum);

        # prepare for next number
        $KbNum = "";

        if ( $PlayerMode eq "Pause" )
            {
            # set Gier Control - absolute heading
            &Log ("Set absolute heading $Number", "A");

            $System{'GierSetpoint'} = &CompassCalc($Number);
            }
        else
            {
            # set target wp
            &WpTargetSet ($Number - 1);

            # Altitude Control
            &AltCurPosSave();
            }
        }

    # performance profiling
    my ($t1_s, $t1_us) = gettimeofday;
    $Profiler{'Timer_Mapoverlay'} = ($t1_s - $t0_s) * 1000000 + $t1_us - $t0_us;
    });


#
# Timer: 0.2s - update Map Overlay - slow objects
#
$frame_map_top->repeat (200, sub
    {
    # performance profiling
    my ($t0_s, $t0_us) = gettimeofday;

    # Clear old messages from this timer
    &MkMessageInit ("Timer-MapOverlay-2");

    lock (%MkOsd);              # until end of block

    if ( &MkOsdIsValid() )
        {
        # Gueltige OSD Daten

        $System{'NoDataLink'} = 0;

        # Operation Mode
        $OperationMode = "";
        if ( &MkIsWptMode() )         { $OperationMode = "WPT"; }
        if ( $PlayerMode eq "Play" )  { $OperationMode = "Play"; }
        if ( $PlayerMode eq "Pause" ) { $OperationMode = "Paus"; }
        if ( $PlayerMode eq "Home" )  { $OperationMode = "Home"; }
        if ( &MkIsPhMode() )          { $OperationMode = "PH"; }
        if ( &MkIsFreeMode() )        { $OperationMode = "Free"; }

        if ( &CurPosIsValid()  and  &HomePosIsValid() )
            {
            # ausreichender GPS Empfang
            # Update OSD - Sat dependent values

            # Alt = average Luftdruck und Sat
            $System{'Alt'} = int (&Altitude() + 0.5);

            my $Alt = sprintf ("%3d m", $System{'Alt'});
            if ( &MkIsWptMode()  and  $System{'TargetSetpoint_AltRate'} != 0 )
                {
                $Alt = $Alt . sprintf " (%d)", $System{'AltSetpoint'};
                }
            $map_canvas->itemconfigure ('MK-OSD-Alt-Value', '-text' => $Alt );
            $map_canvas->itemconfigure ('MK-OSD-Spd-Value', '-text' => sprintf ("%3d km/h", $MkOsd{'GroundSpeed'} * 0.036) );

            if ( &TargetIsValid() )
                {
                # show target icon
                my $IconHeight = 48;
                my $IconWidth = 48;
                $map_canvas->coords('Target', $System{'TargetPos_x'} - $IconWidth/2, $System{'TargetPos_y'} - $IconHeight );

                # Border check
                $System{'CrossingBorder'} = 0;
                if ( &MkIsFlying()  and  &IsCrossingBorder($System{'CurPos_x'},    $System{'CurPos_y'},
                                                           $System{'TargetPos_x'}, $System{'TargetPos_y'}) )
                    {
                    # only, if MK is flying
                    $System{'CrossingBorder'} = 1;
                    &MkMessage ($Translate{'MsgCrossingBorder'}, "Timer-MapOverlay-2");
                    }
                }
            else
                {
                # no target, hide target icon
                $map_canvas->coords('Target', 0, -100, );
                }

            # POI direction, camera elevation
            if ( $PoiMode)
                {
                # direction to POI
                my ($PoiDist, $PoiBearing) = &MapGpsTo ( $MkOsd{'CurPos_Lat'}, $MkOsd{'CurPos_Lon'},
                                                         $Poi[$PoiIndex]{'Poi_Lat'},   $Poi[$PoiIndex]{'Poi_Lon'} );
                $System{'PoiDist'}    = $PoiDist;
                $System{'PoiBearing'} = $PoiBearing;
 
                # camera elevation to POI
                my $AltDiff = $System{'Alt'} - $Poi[$PoiIndex]{'Poi_Alt'};
                my $Dist = sqrt ($AltDiff * $AltDiff + $PoiDist * $PoiDist);
                if ( $Dist > 5 )
                    {
                    $System{'PoiElevation'} = rad2deg atan2 ($AltDiff, $PoiDist);
                    }
                }
            }
        else
            {
            # kein ausreichender Sat-Empfang
            $map_canvas->itemconfigure ('MK-OSD-Spd-Value', '-text' => sprintf ("%3d km/h", 0 ) );
            }

        #
        # Update OSD - non Sat dependent values
        #

        $map_canvas->itemconfigure ('MK-OSD-Odo-Value', '-text' => sprintf ("%3.3f km", $OdoMeter / 1000) );
        $map_canvas->itemconfigure ('MK-OSD-Tim-Value', '-text' => sprintf ("%02d:%02d", $MkFlyingTime / 60, $MkFlyingTime % 60) );
        $map_canvas->itemconfigure ('MK-OSD-Cur-Value', '-text' => sprintf ("%.1f A", $MkOsd{'Current'}) );

        # GPS OK?
        my $ColorOsdSat = $Cfg->{'mkcockpit'}->{'ColorOsd'};
        if ( ! &MkGpsOk() )
            {
            $ColorOsdSat = 'red';
            }
        $map_canvas->itemconfigure ('MK-OSD-Sat-Value', '-fill' => $ColorOsdSat);
        my $ValueOsdSat = $MkOsd{'SatsInUse'};
        if ( $Mk{'GpsFix'} =~ /DGPS/i )
            {
            $ValueOsdSat .= "     DGPS";
            }
        $map_canvas->itemconfigure ('MK-OSD-Sat-Value', '-text' => $ValueOsdSat );

        # Used Capacity
        my $UsedCapacityFactor = &Default ($Cfg->{'map'}->{'UsedCapacityFactor'}, "1.0");
        $System{'UsedCapacity'} = $MkOsd{'UsedCapacity'} * $UsedCapacityFactor;
        $map_canvas->itemconfigure ('MK-OSD-Cap-Value', '-text' => sprintf ("%.2f Ah", $System{'UsedCapacity'} / 1000) );

        # Power
        $System{'Power'} = int ($MkOsd{'Current'} * $MkOsd{'UBat'} + 0.5);
        $map_canvas->itemconfigure ('MK-OSD-Pow-Value', '-text' => sprintf ("%d W", $System{'Power'} ) );

        # battery - OSD and warning
        my $UBat = sprintf ("%3.1f", $MkOsd{'UBat'});
        $System{'UBat'} = $UBat;
        $map_canvas->itemconfigure ('MK-OSD-Bat-Value', '-text' => "$UBat V" );

        $System{'BatWarning'} = 0;
        $map_canvas->itemconfigure ('MK-OSD-Bat-Value', '-fill' => $Cfg->{'mkcockpit'}->{'ColorOsd'});
        if ( &MkIsLowVoltage() )
            {
            if ( time %2 )
                {
                $map_canvas->itemconfigure ('MK-OSD-Bat-Value', '-fill' => 'red');
                }

            &MkMessage ($Translate{'MsgBatWarning'}, "Timer-MapOverlay-2");
            $System{'BatWarning'} = 1;
            }

        # Display Player Operation Mode
        my $DisplayMode = $OperationMode;
        if ( &MkIsWptMode()  and  $OperationMode eq "Play" )
            {
            my %ModeMatrix = 
               (
               "KML-STD" => "Play KML",
               "KML-RND" => "Play KML",
               "KML-MAP" => "Play KML",
               "WPT-STD" => "Play WPT",
               "WPT-RND" => "Rand WPT",
               "WPT-MAP" => "Rand MAP",
               "SPD-STD" => "Play SPD",
               "SPD-RND" => "Play SPD",
               "SPD-MAP" => "Play SPD",
               );
            my $Key = "${PlayerWptKmlMode}-${PlayerRandomMode}";
            $DisplayMode = $ModeMatrix{$Key};
            }

        if ( &MkIsWptMode()  and  $OperationMode eq "Paus" )
            {
            $DisplayMode = $DisplayMode . " " . $PlayerPauseMode;
            }
        $map_canvas->itemconfigure ('MK-OSD-Mode-Value', '-text' => $DisplayMode );


        # Waypoints abhaengig vom Modus NC/Player
        my $WpValue = "-- / --";
        if ( $MkOsd{'WaypointNumber'} > 0)
            {
            $WpValue = sprintf ("%d / %d", $MkOsd{'WaypointIndex'}, $MkOsd{'WaypointNumber'});
            }
        if ($PlayerMode ne "Stop" and $PlayerWptKmlMode eq "WPT")
            {
            $WpValue = sprintf ("%d / %d", $WpPlayerIndex +1, scalar @Waypoints);
            }
        if ($PlayerMode ne "Stop" and $PlayerWptKmlMode eq "KML" )
            {
            my $KmlTimeBase = &Default ($Cfg->{'waypoint'}->{'KmlTimeBase'}, 1.0);
            my $CurrTime = int ($KmlPlayerIndex * $KmlTimeBase + 0.5);
            my $TotTime = int (scalar @KmlTargets * $KmlTimeBase + 0.5);
            $WpValue = sprintf ("%02d:%02d / %02d:%02d", $CurrTime / 60, $CurrTime % 60, $TotTime / 60, $TotTime % 60);
            }
        if ($PlayerMode ne "Stop" and $PlayerWptKmlMode eq "SPD" )
            {
            $WpValue = sprintf ("%02d:%02d (%d/%d)", $SpdPlayerTime / 60, $SpdPlayerTime % 60,
                                                     $SpdPlayerIndex + 1, scalar @Waypoints);
            }
        $map_canvas->itemconfigure ('MK-OSD-Wp-Value',  '-text' => "$WpValue");

        # Recording Mode
        my $RecordText = "";
        if ( $PlayerRecordMode =~ /REC/i )
            {
            my $KmlTimeBase = &Default ($Cfg->{'waypoint'}->{'KmlTimeBase'}, 1.0);
            my $TotTime = int (scalar @KmlTargets * $KmlTimeBase + 0.5);
            $RecordText = sprintf ("Recording %02d:%02d", $TotTime / 60, $TotTime % 60);
            }
        $map_canvas->itemconfigure ('MK-OSD-Rec-Value', '-text' => $RecordText );

        # Farbe MK-Zeiger abhngig vom GPS Empfang
        my $MkCol= $Cfg->{'mkcockpit'}->{'ColorMkSatGood'};
        if ( ! &MkGpsOk() )
            {
            $MkCol= $Cfg->{'mkcockpit'}->{'ColorMkSatNo'};
            }
        elsif ( $MkOsd{'SatsInUse'} < 6 )
            {
            $MkCol = $Cfg->{'mkcockpit'}->{'ColorMkSatLow'} ;
            }
        $map_canvas->itemconfigure ('MK-Arrow', '-fill' => $MkCol);

        # Show/Hide SatFix Icon
        if ( &MkGpsOk() )
            {
            $map_canvas->coords('Satellite', $MapSizeX/4-10, 55, );
            }
        else
            {
            # move icon out of sight
            $map_canvas->coords('Satellite', 0, -100, );
            }

        # Variometer Pointer
        my $dy = -$MkOsd{'Variometer'} * 2;
        $map_canvas->coords('Map-Variometer-Pointer', 5, $MapSizeY/2+$dy, 20, $MapSizeY/2+10+$dy, 20, $MapSizeY/2-10+$dy);

        #
        # System checks
        #

        if ( ! &MkIsMotorOn() )      { &MkMessage ($Translate{'MsgMotorOff'},         "Timer-MapOverlay-2"); }
        if ( ! &MkIsFlying() )       { &MkMessage ($Translate{'MsgNotFlying'},        "Timer-MapOverlay-2"); }
        if ( &MkIsCalibrating() )    { &MkMessage ($Translate{'MsgCalibrate'},        "Timer-MapOverlay-2"); }
        if ( &MkIsMotorStarting() )  { &MkMessage ($Translate{'MsgStart'},            "Timer-MapOverlay-2"); }
        if ( &MkEmergencyLanding() ) { &MkMessage ($Translate{'MsgEmergencyLanding'}, "Timer-MapOverlay-2"); }
        if ( &MkRangeLimit() )       { &MkMessage ($Translate{'MsgRangeLimit'},       "Timer-MapOverlay-2"); }
        if ( &MkManualControl() )    { &MkMessage ($Translate{'MsgManualControl'},    "Timer-MapOverlay-2"); }

        # RC range check
        my $RcQuality = $MkOsd{'RC_Quality'};
        $System{'RCQuality'} = "";

        if ( $RcQuality < 100 )
            {
            $System{'RCQuality'} = "NO";
            &MkMessage ($Translate{'MsgRcError'}, "Timer-MapOverlay-2");
            }
        elsif ( $RcQuality < 150 )
            {
            $System{'RCQuality'} = "WEAK";
            &MkMessage ($Translate{'MsgRcWarning'}, "Timer-MapOverlay-2");
            }

        # Sat reception quality
        if ( ! &MkGpsOk() )
            {
            &MkMessage ($Translate{'MsgNoSatReception'}, "Timer-MapOverlay-2");
            }
        elsif ( $MkOsd{'SatsInUse'} < 6 )
            {
            &MkMessage ($Translate{'MsgWeakSatReception'}, "Timer-MapOverlay-2");
            }

        # MK Border check
        $System{'OutsideBorder'} = "0";
        if ( &MkIsFlying() and ! &IsInsideBorder($System{'CurPos_x'}, $System{'CurPos_y'}) )
            {
            # only, if MK is flying
            $System{'OutsideBorder'} = "1";
            &MkMessage ($Translate{'MsgOutsideBorder'}, "Timer-MapOverlay-2");
            }

        # Range Warning
        $System{'RangeWarning'} = 0;
        if ( &MkRangeLimit() )
            {
            $System{'RangeWarning'} = 1;
            }

        # Show Balloon, when aproaching Target
        &TargetMessageShow();
        }
    else
        {
        # keine aktuellen OSD Daten vom MK verfgbar
        $System{'NoDataLink'} = 1;
        &MkMessage ($Translate{'MsgNoData'}, "Timer-MapOverlay-2");
        }


    # display transmit status in upper status line
    my $Color  = 'lightgray';
    my $Relief = 'flat';
    if ( $Cfg->{'serialchannel'}->{'SerialChannelSend'} =~ /y/i )
        {
        $Color = 'green';
        $Relief = 'sunken';
        }
    if ($TxExtOn == 0 )
        {
        $Color = 'red';
        $Relief = 'sunken';
        }
    $map_status_top[0]->configure (-text       => "Tx:$Translate{'serialchannel'}",
                                   -background => $Color,
                                   -relief     => $Relief,
                                  );
    my $Color = 'lightgray';
    my $Relief = 'flat';
    if ( $Cfg->{'externcontrol'}->{'ExternControlSend'} =~ /y/i )
        {
        $Color  = 'green';
        $Relief = 'sunken';
        }
    if ($TxExtOn == 0 )
        {
        $Color  = 'red';
        $Relief = 'sunken';
        }
    $map_status_top[1]->configure (-text       => "Tx:$Translate{'externcontrol'}",
                                   -background => $Color,
                                   -relief     => $Relief,
                                  );

    # FC operating status in upper status line
    &DispayUpperStatus (2, &MkIsCarefree(), 'CareFree', 'green');
    &DispayUpperStatus (3, &MkIsAltCtrl(),  'AltCtrl',  'green');
    &DispayUpperStatus (4, &MkIsOut2(),  'Out2',  'green');
    &DispayUpperStatus (5, &MkIsOut1(),  'Out1',  'green');

    # display event status in lower status line, max. 10 fields
    my @Event = sort keys %{$Event};
    my $EventIndex = 0;
    while ( $EventIndex < 10 )
        {
        my $Key = pop @Event;
        my $Color = 'lightgray';
        my $Relief = 'flat';

        if ( $Key eq "" )
            {
            # clear unused fields
            $map_status_event[$EventIndex]->configure (-text       => $Key,
                                                       -background => $Color,
                                                       -relief     => $Relief,
                                                      );
            $EventIndex ++;
            }

        elsif ( ref $Event->{$Key} ne ""  and  $Event->{$Key}->{'Active'} =~ /y/i )
            {
            if ( $EventStat{$Key}{'Status'} eq "ON" )
                {
                $Color  = 'green';
                $Relief = 'sunken';
                }
            elsif ( $EventStat{$Key}{'Status'} eq "OFF" )
                {
                $Color  = 'red';
                $Relief = 'sunken';
                }

            $map_status_event[$EventIndex]->configure (-text       => $Key,
                                                       -background => $Color,
                                                       -relief     => $Relief,
                                                      );
            $EventIndex ++;
            }
         else
            {
            # don't display disabled events
            }
        }

    # Show System Messages
    &MkMessageShow();

    # performance profiling
    my ($t1_s, $t1_us) = gettimeofday;
    $Profiler{'Timer_Mapoverlay2'} = ($t1_s - $t0_s) * 1000000 + $t1_us - $t0_us;
    });


#	
# Timer: 0.25s - request Poti LCD-Info Screen 9,10  (8= Sticks)
#
my $LcdScreen = 9;
if ( $Cfg->{'mkcomm'}->{'RequestRcChannel'} =~ /y/i )
    {
    $frame_map_top->repeat (250, sub
        {
        # performance profiling
        my ($t0_s, $t0_us) = gettimeofday;

        # request only one screen at each timer run to avoid data link overload
        $MkSendQueue->enqueue( "l", "$AddrNC", pack ("C", $LcdScreen) );

        $LcdScreen ++;
        if ( $LcdScreen > 10 )
            {
            $LcdScreen = 9;
            }

        # performance profiling
        my ($t1_s, $t1_us) = gettimeofday;
        $Profiler{'Timer_RequestPoti'} = ($t1_s - $t0_s) * 1000000 + $t1_us - $t0_us;
        });
    }

#	
# Timer: 0.1s - update Tracking overlay
#
if ( $Cfg->{'track'}->{'Active'} =~ /y/i )
    {
    $TrackMkCalibrateEdge = 0;  # calibrate edge detection

    $frame_map_top->repeat (100, sub
        {
        # performance profiling
        my ($t0_s, $t0_us) = gettimeofday;

        # Clear old messages from this timer
        &MkMessageInit ("Timer-Tracking");     

        lock (%MkOsd);              # until end of block
        lock (%MkTrack);            # until end of block

        # Aktuell gltige Karte
        %Map = %{$Maps{'Current'}};

        # Zeiger neu zeichnen
        my $AngelPan  = @ServoPos[$MkTrack{'ServoPan'}];
        my $AngelTilt = @ServoPos[$MkTrack{'ServoTilt'}];
        if ( $AngelPan ne "" and $AngelTilt ne "" )
            {
            my $Angel = $AngelPan;
            if ( $AngelTilt > 90 )
                {
                $Angel += 180;
                }

            # Pan
            my $x0 = $TrackSizeX/2;     
            my $y0 = $TrackSizeY - 0 - $TrackOffY;
            my $x1 = $TrackSizeX/2 - ($TrackPtrLen-22) * cos( deg2rad $Angel);
            my $y1 = $TrackSizeY   - ($TrackPtrLen-22) * sin (deg2rad $Angel) - $TrackOffY;
            $track_canvas->coords ('Track-Ptr-Pan', $x0, $y0, $x1, $y1);

            # Tilt
            my $x0 = $TrackSizeX/2;     
            my $y0 = $TrackSizeY - 0 - $TrackOffY;
            my $x1 = $TrackSizeX/2 - ($TrackPtrLen-22) * cos( deg2rad 180 - $AngelTilt);
            my $y1 = $TrackSizeY   - ($TrackPtrLen-22) * sin (deg2rad 180 - $AngelTilt) - $TrackOffY;
            $track_canvas->coords ('Track-Ptr-Tilt', $x0, $y0, $x1, $y1);
            }

        # Farbe Zeiger abhngig vom GPS Empfang
        my $TrackPtrCol= 'green';
        if ( ! &MkGpsOk() )
            {
            $TrackPtrCol = 'red';
            }
        elsif ( $MkOsd{'SatsInUse'} < 6 )
            {
            $TrackPtrCol = 'orange';
            }
        $track_canvas->itemconfigure ('Track-Ptr-Pan', '-fill' => $TrackPtrCol);

        # tracker coldstart condition at MK Motor start, rising edge of signal
        if ( &MkIsMotorStarting() )
            {
            if ( $TrackMkColdStartEdge == 0 )
                {
                $TrackMkColdStartEdge = 1;

                # send coldstart command to tracker
                $TrackQueue->enqueue("COLDSTART");
                }
            }
        else
            {
            $TrackMkColdStartEdge = 0;
            }

        # performance profiling
        my ($t1_s, $t1_us) = gettimeofday;
        $Profiler{'Timer_Tracking'} = ($t1_s - $t0_s) * 1000000 + $t1_us - $t0_us;
        });
    }


#	
# Timer: 0.5s - Waypoint Player
#
$System{'AltWpIndex'}   = 0;     # for altitude control
$System{'GierSetpoint'} = 0;     # Gier bearing control

$frame_map_top->repeat (500, sub
    {
    # performance profiling
    my ($t0_s, $t0_us) = gettimeofday;

    # Clear old messages from this timer
    &MkMessageInit ("Timer-Player");         

    lock (%MkOsd);              # until end of block

    if ( &MkIsWptMode() )
        {
        # NC is in WPT Mode

        # Switch to pause, if manual RC control in play-mode
        if ( $PlayerMode eq "Play" and  &MkManualControl() )
            {
            &PlayerPause();
            }

        # Altitude Control
        if ( $System{'AltStartPos_x'} eq ""  or  $System{'AltStartPos_y'} eq "" )
            {
            &AltCurPosSave();             
            }

        # POI Gier control
        if ( $PlayerMode ne "Stop"  and  $PoiMode  and  &CurPosIsValid()  and  $System{'PoiDist'} > 5 )
            {
            $System{'GierSetpoint'} = -2;   # direction to POI
            ## $System{'GierSetpoint'} = &CompassCalc ($System{'PoiBearing'});

            # send POI
            &MkFlyTo ( '-lat'       => $Poi[$PoiIndex]{'Poi_Lat'},
                       '-lon'       => $Poi[$PoiIndex]{'Poi_Lon'},
                       '-alt'       => $Poi[$PoiIndex]{'Poi_Alt'},
                       '-eventflag' => $Poi[$PoiIndex]{'Poi_EventFlag'},
                       '-type'      => 1,          # POI
                       '-index'     => 1,          # Slot 2
                       '-mode'      => 'Waypoint',
                     );
            }

        # player pause control
        if ( $PlayerMode eq "Pause" and  $PlayerPause_Lat ne ""  and  $PlayerPause_Lon ne "" )
            {
            # AltRate depending on altitude difference
            my $AltRate = 0;  # disable
            my $MinAltRate = &Default ($Cfg->{'map'}->{'ControlMinAltRate'}, 10); # 0.1 m/s
            if ( $MinAltRate > 0 )
                {
                $AltRate = abs ($System{'AltSetpoint'} - &Altitude());
                if ( $AltRate < $MinAltRate ) { $AltRate = $MinAltRate; }   # min. m/s
                if ( $AltRate > 60 )          { $AltRate = 60; }            # max. 6 m/s
                }

            # Gespeicherte Pausen-Pos senden
            &MkFlyTo ( '-lat'      => $PlayerPause_Lat,
                       '-lon'      => $PlayerPause_Lon,
                       '-holdtime' => "60",
                       '-heading'  => $System{'GierSetpoint'},
                       '-alt'      => $System{'AltSetpoint'},
                       '-altrate'  => $AltRate,
                       '-mode'     => 'Target',
                       # '-speed'    => '',
                       # '-camangle' => '',
                     );
            }

        # player home control
        if ( $PlayerMode eq "Home" and &HomePosIsValid() )
            {
            # Gespeicherte oder eingestellte Home-Pos senden

            my $Home_Lat = &Default ($Map{'Home_Lat'}, $MkOsd{'HomePos_Lat'});
            my $Home_Lon = &Default ($Map{'Home_Lon'}, $MkOsd{'HomePos_Lon'});

            &MkFlyTo ( '-lat'      => $Home_Lat,
                       '-lon'      => $Home_Lon,
                       '-holdtime' => "60",
                       '-heading'  => $System{'GierSetpoint'},
                       '-alt'      => $System{'AltSetpoint'},
                       '-altrate'  => 0,     # alt disabled
                       '-mode'     => 'Target',
                       # '-speed'    => '',
                       # '-camangle' => '',
                     );
            }


        if ( $PlayerWptKmlMode ne 'WPT' )
            {
            # not in Wp mode
            return;
            }


        if ( $PlayerMode eq "Play"  )
            {
            if ( $PlayerRandomMode =~ /RND/i  or $PlayerRandomMode =~ /STD/i )
                {
                # Altitude control
                # $System{'AltSetpoint'} calculated once when switching to next WP

                $System{'TargetNotReachable'} = 0;
                $map_canvas->delete('TargetSetpoint');

                my $WpCnt = scalar @Waypoints;
                if ( $WpCnt > 0  and  $WpPlayerIndex < $WpCnt )
                    {
                    $System{'AltWpIndex'} = $WpPlayerIndex;

                    # Target WP-Pos senden
                    my $Wp = $Waypoints[$WpPlayerIndex];
                    my $Wp_Lon               = $Wp->{'Pos_Lon'};
                    my $Wp_Lat               = $Wp->{'Pos_Lat'};
                    my $Wp_Alt               = $Wp->{'Pos_Alt'};
                    my $Wp_AltRate           = $Wp->{'AltRate'};
                    my $Wp_Heading           = $Wp->{'Heading'};
                    my $Wp_EventFlag         = $Wp->{'Event_Flag'};
                    my $Wp_EventChannelValue = $Wp->{'EventChannelValue'};
                    my $Wp_Speed             = $Wp->{'Speed'};
                    my $Wp_CamAngle          = $Wp->{'CamAngle'};

                    # Gier Control
                    if ( $Wp_Heading >= 362  and  ! $PoiMode )
                        {
                        # gier to flight direction
                        $System{'GierSetpoint'} = &CompassCalc($Wp_Heading);
                        }

                    if ( $Wp_Lat ne ""  and  $Wp_Lon ne "" )
                        {
                        &MkFlyTo ( '-lat'               => $Wp_Lat,
                                   '-lon'               => $Wp_Lon,
                                   '-holdtime'          => "60",
                                   '-heading'           => $System{'GierSetpoint'},
                                   '-alt'               => $System{'AltSetpoint'},
                                   '-altrate'           => $Wp_AltRate,
                                   '-eventflag'         => $Wp_EventFlag,
                                   '-eventchannelvalue' => $Wp_EventChannelValue,
                                   '-speed'             => $Wp_Speed,
                                   '-camangle'          => $Wp_CamAngle,
                                   '-mode'              => "Target",
                                 );
                        }
                    }
                }

            if ( $PlayerRandomMode =~ /MAP/i )
                {
                # Target Map-Pos senden
                &MkFlyTo ( '-x'                 => $RandomTarget_x ,
                           '-y'                 => $RandomTarget_y ,
                           '-holdtime'          => "60",
                           '-heading'           => $System{'GierSetpoint'},
                           '-alt'               => $System{'AltSetpoint'},
                           '-altrate'           => $Cfg->{'waypoint'}->{'DefaultAltRate'},
                           '-speed'             => $Cfg->{'waypoint'}->{'DefaultSpeed'},
                           '-camangle'          => $Cfg->{'waypoint'}->{'DefaultCamAngle'},
                           '-mode'              => 'Target',
                         );
                }
 
            # Ziel erreicht?
            if ( &WpCheckTargetReached() )
                {
                &WpTargetNext();

                # Altitude control
                &AltCurPosSave();
                $System{'AltWpIndex'} = $WpPlayerIndex;
                }
            }
        }

    # WP Player Holdtime count down
    if ( $WpPlayerHoldtime > 0  )
        {
        $WpPlayerHoldtime --;
        }

    # performance profiling
    my ($t1_s, $t1_us) = gettimeofday;
    $Profiler{'Timer_Player'} = ($t1_s - $t0_s) * 1000000 + $t1_us - $t0_us;
    });


#	
# Timer: 0.5s - Speed (SPD) controlled Player
#
$frame_map_top->repeat (500, sub
    {
    # performance profiling
    my ($t0_s, $t0_us) = gettimeofday;

    # Clear old messages from this timer
    &MkMessageInit ("Timer-SpdPlayer");     

    lock (%MkOsd);              # until end of block

    if ( &MkIsWptMode() and  $PlayerMode eq "Play"  and  $PlayerWptKmlMode eq 'SPD')
        {
        # Play SPD
        # Pause, Home is handled in WPT-Timer

        # Altitude control
        $System{'AltSetpoint'} = &AltCalc();

        $System{'TargetNotReachable'} = 0;
        $map_canvas->delete('TargetSetpoint');

        my ($x, $y, $TargetWpIndex) = &SplinePosFromTime ($SpdPlayerTime);
        if ( $x ne ""  and  $y ne "" )
            {
            if ( &IsTargetReachable($x, $y) )
                {
                # Gier Control
                my $Wp = $Waypoints[$TargetWpIndex];
                my $Wp_Heading = $Wp->{'Heading'};
                if ( $Wp_Heading >= 362  and  ! $PoiMode )
                    {
                    # gier to WP-target direction
                    $System{'GierSetpoint'} = &CompassCalc($Wp_Heading);
                    }

                # Fly to target
                &MkFlyTo ( '-x'          => $x,
                           '-y'          => $y,
                           '-holdtime'   => "60",
                           '-heading'    => $System{'GierSetpoint'},  # calculated in WPT-Player
                           '-alt'        => $System{'AltSetpoint'},
                           '-altrate'    => $Wp->{'AltRate'},
                           '-speed'      => 0,                        # max. speed ($Wp->{'Speed'},)
                           '-camangle'   => $Wp->{'CamAngle'},
                           '-mode'       => "Target",
                         );
                }
            else
                {
                # Target not reachable
                $System{'TargetNotReachable'} = 1;
                &MkMessage ( $Translate{'MsgTrackOutsideBorder'}, "Timer-SpdPlayer");

                $map_canvas->createOval ( $x-4, $y+4, $x+4, $y-4,
                                         '-tags'    => 'TargetSetpoint',
                                         '-fill'    => 'green',
                                         '-outline' => "white",
                                        );
                }
            }
        else
            {
            # No Target available
            $System{'TargetNotReachable'} = 1;
            &MkMessage ( $Translate{'MsgTrackOutsideBorder'}, "Timer-SpdPlayer");
            }

        # Ziel-WP erreicht?
        if ( $SpdPlayerIndex != $TargetWpIndex  and  $SpdPlayerTime > 0 )
            {
            &Log ("SPD Player target $SpdPlayerIndex reached", "A");

            # Altitude control
            &AltCurPosSave($SpdPlayerIndex);

            $SpdPlayerIndex = $TargetWpIndex;

            # Gier Control
            my $Wp = $Waypoints[$SpdPlayerIndex];
            $System{'GierSetpoint'} = &CompassCalc($Wp->{'Heading'});

            &TtsSpeak ('MEDIUM', $Translate{'TtsTargetReached'});
            }
        $System{'AltWpIndex'} = $SpdPlayerIndex;

        # Operation Radius pruefen
        my $T_Lat = $System{'TargetSetpoint_Lat'};
        my $T_Lon = $System{'TargetSetpoint_Lon'};
        my ($HomeDist, $HomeBearing) = &MapGpsTo($MkOsd{'HomePos_Lat'}, $MkOsd{'HomePos_Lon'}, $T_Lat, $T_Lon );
        if ( $HomeDist > $MkOsd{'OperatingRadius'} )
            {
            # TargetSetpoint entsprechend Operation Radius neu berechnen
            $HomeDist = $MkOsd{'OperatingRadius'};
            ($T_Lat, $T_Lon) = &MapGpsAt($MkOsd{'HomePos_Lat'}, $MkOsd{'HomePos_Lon'}, $HomeDist, $HomeBearing);
            }

        # wait, if distance to TargetSetpoint is too big
        my ($Dist, $Bearing) = &MapGpsTo($MkOsd{'CurPos_Lat'}, $MkOsd{'CurPos_Lon'}, $T_Lat, $T_Lon );
        my $Wp = $Waypoints[$SpdPlayerIndex];
        my $Speed = &Default ($Wp->{'Speed'}, 10);        # km/h
        my $MaxDist = $Speed / 3.6 * 6;          # max n sec. distance
        if ( $MaxDist < 10 ) { $MaxDist = 10; }
        if ( $MaxDist > 30 ) { $MaxDist = 30; }
        if ( $Dist < $MaxDist )
            {
            # proceed to next Target
            $SpdPlayerTime += 0.5;
            }
        }

    # performance profiling
    my ($t1_s, $t1_us) = gettimeofday;
    $Profiler{'Timer_SpdPlayer'} = ($t1_s - $t0_s) * 1000000 + $t1_us - $t0_us;
    });


#	
# Timer: variabel - KML Player
#
my $KmlTimeBase = &Default ($Cfg->{'waypoint'}->{'KmlTimeBase'}, 1.0);
$KmlTimeBase *= 1000;

$frame_map_top->repeat ($KmlTimeBase, sub
    {
    # performance profiling
    my ($t0_s, $t0_us) = gettimeofday;

    # Clear old messages from this timer
    &MkMessageInit ("Timer-KMLPlayer");     

    lock (%MkOsd);              # until end of block

    if ( &CurPosIsValid() and $PlayerRecordMode =~ /REC/i )
        {
        # record current position
        push @KmlTargets, {
                          'Lat'  => $MkOsd{'CurPos_Lat'},
                          'Lon'  => $MkOsd{'CurPos_Lon'},
                          'Alt'  => &Altitude(),
                          'Gier' => $MkOsd{'CompassHeading'}, 
                          }; 
        }

    if ( &MkIsWptMode() and  $PlayerMode eq "Play"  and  $PlayerWptKmlMode eq 'KML')
        {
        # Play KML
        # Pause, Home is handled in WPT-Timer

        $System{'TargetNotReachable'} = 0;
        $map_canvas->delete('TargetSetpoint');

        my $KmlCnt = scalar @KmlTargets;
        if ( $KmlCnt > 0  and  $KmlPlayerIndex < $KmlCnt )
            {
            my $Lat     = $KmlTargets[$KmlPlayerIndex]->{'Lat'};
            my $Lon     = $KmlTargets[$KmlPlayerIndex]->{'Lon'};
            my $Alt     = $KmlTargets[$KmlPlayerIndex]->{'Alt'};
            my $AltRate = &Default($Cfg->{'waypoint'}->{'KmlAltRate'}, 20);
            my $Gier    = $KmlTargets[$KmlPlayerIndex]->{'Gier'};

            my ($T_x, $T_y) = &MapGps2XY ($Lat, $Lon);
            if ( &IsTargetReachable($T_x, $T_y) )
                {
                # Altitude control
                $System{'AltSetpoint'} = $Alt;

                # Gier Control
                if ( $Gier ne "" )
                    {
                    $System{'GierSetpoint'} = $Gier;
                    }

                &MkFlyTo ( '-lat'      => $Lat,
                           '-lon'      => $Lon,
                           '-holdtime' => "60",
                           '-heading'  => $System{'GierSetpoint'},
                           '-alt'      => $Alt,
                           '-altrate'  => $AltRate,
                           '-speed'    => 0,
                           # '-camangle' => '',
                           '-mode'     => "Target",
                         );
                }
            else
                {
                $System{'TargetNotReachable'} = 1;
                $map_canvas->createOval ( $T_x-4, $T_y+4, $T_x+4, $T_y-4,
                                         '-tags'    => 'TargetSetpoint',
                                         '-fill'    => 'green',
                                         '-outline' => "white",
                                        );

                &MkMessage ( $Translate{'MsgTrackOutsideBorder'}, "Timer-KMLPlayer");
                }

            # wait, if distance to target is too big
            if ( $System{'TargetDist'} < 20 )
                {
                # proceed to next Target
                $KmlPlayerIndex ++;
                if ( $KmlPlayerIndex >= $KmlCnt )
                    {
                    $KmlPlayerIndex = 0;
                    }
                }
            }
        }

    # performance profiling
    my ($t1_s, $t1_us) = gettimeofday;
    $Profiler{'Timer_KmlPlayer'} = ($t1_s - $t0_s) * 1000000 + $t1_us - $t0_us;
    });


#	
# Timer: 0.1s - POI Timer for Free and PH mode
#
$frame_map_top->repeat (100, sub
    {
    # performance profiling
    my ($t0_s, $t0_us) = gettimeofday;

    # Clear old messages from this timer
    &MkMessageInit ("Timer-POI");         

    lock (%MkOsd);              # until end of block

    if ( $PlayerMode ne "Stop" )
        {
        if ( $PoiMode  and  &MkIsCarefree()  and  (&MkIsFreeMode()  or &MkIsPhMode()) )
            {
            &NcSetParam(1, $System{'PoiBearing'});
            &NcSetParam(2, $System{'PoiElevation'} * 10 );
            }
        }

    #
    # move Pause-Position manually with RC
    #
    if ( &MkIsWptMode()  and  $PlayerMode eq 'Pause'  and  &MkManualControl() )
        {
        $PlayerPause_Lat = $MkOsd{'CurPos_Lat'};
        $PlayerPause_Lon = $MkOsd{'CurPos_Lon'};

        # restart crosshair display timer
        $CrosshairTimerCnt = 0;
        }

    # performance profiling
    my ($t1_s, $t1_us) = gettimeofday;
    $Profiler{'Timer_POI'} = ($t1_s - $t0_s) * 1000000 + $t1_us - $t0_us;
    });


#	
# Timer: 1s
#
$frame_map_top->repeat (1000, sub
    {
    # performance profiling
    my ($t0_s, $t0_us) = gettimeofday;

    # Clear old messages from this timer
    &MkMessageInit ("Timer-Misc-1s");     

    # Refresh Waypoint List Window
    &WpEditRefresh();

    lock (%MkOsd);              # until end of block

    # Aktuell gltige Karte
    %Map = %{$Maps{'Current'}};

    if ( &MkOsdIsValid() )
        {

        # Heartbeat MK Datenbertragung
        if ( time %2 )
            {
            $map_canvas->itemconfigure('Heartbeat', '-image' => 'HeartbeatLarge', );
            }
        else
            {
            $map_canvas->itemconfigure('Heartbeat', '-image' => 'HeartbeatSmall', );
            }

        # Flugzeit aktualisieren
        # Flugzeit selber mitzhlen, da $MkOsd{'FlyingTime'} immer 0 (0.14b)
        if ( &MkIsFlying() )
            {
            $MkFlyingTime += 1;
            }
        $System{'MkFlyingTime'} = $MkFlyingTime;

        # Update ODO-Meter
        if ( &CurPosIsValid() )
            {
            my $C_Lat = $MkOsd{'CurPos_Lat'};
            my $C_Lon = $MkOsd{'CurPos_Lon'};
            my $C_Alt = &Altitude();

            if ( $OdoFirst ne "" )
                {
                my ($Dist, $Bearing) = &MapGpsTo($C_Lat, $C_Lon, $OdoPos_Lat, $OdoPos_Lon );
                $AltDiff = $C_Alt - $OdoPos_Alt;
                $OdoMeter += sqrt ($Dist * $Dist  + $AltDiff * $AltDiff);
                }
            $OdoPos_Lat = $C_Lat;
            $OdoPos_Lon = $C_Lon;
            $OdoPos_Alt = $C_Alt;
            $OdoFirst = "1";
            }

        # Footprint
        if ( $Cfg->{'map'}->{'FootprintLength'} > 0 )
            {
            if ( &CurPosIsValid() )
                {
                # neuen Footprint hinten anhaengen
                my ($x, $y) = &MapGps2XY($MkOsd{'CurPos_Lat'}, $MkOsd{'CurPos_Lon'});
                push @Footprint, $x, $y;
                }

            while ( $#Footprint / 2  >  $Cfg->{'map'}->{'FootprintLength'} )
                {
                # alte Footprints entfernen
                splice @Footprint, 0, 2;
                }

            &FootprintRedraw();
            }

        # show tracking antenna icon
        if ( $MkTrack{'HomePos_Lat'} ne ""  and  $MkTrack{'HomePos_Lon'} ne ""  )
            {
            # show antenna icon
            my ($x, $y) = &MapGps2XY($MkTrack{'HomePos_Lat'}, $MkTrack{'HomePos_Lon'});

            my $IconHeight = 48;
            my $IconWidth = 48;
            $map_canvas->coords('Track-Antenna', $x - $IconWidth/2, $y - $IconHeight );
            }
        else
            {
            # move icon out of sight
            $map_canvas->coords('Track-Antenna', 0, -50 );
            }

        if ( $GridIsOn )
            {
            # redraw Grid on canvas
            &GridHide();
            &GridShow();
            }

        # fun :-)
        if ( int rand (100) == 43 )
            {
            &TtsSpeak ('LOW', $Translate{'TtsFun'});
            }
        }

    # performance profiling
    my ($t1_s, $t1_us) = gettimeofday;
    $Profiler{'Timer_Misc'} = ($t1_s - $t0_s) * 1000000 + $t1_us - $t0_us;
    });


#	
# Timer: 1s - TTS Sprachausgabe
#
$frame_map_top->repeat (1000, sub
    {
    # performance profiling
    my ($t0_s, $t0_us) = gettimeofday;

    # Aktuell gltige Karte
    %Map = %{$Maps{'Current'}};

    lock (%MkOsd);              # until end of block

    my $StatusInterval = &Default ($Cfg->{'tts'}->{'StatusInterval'}, "30");   
    $SpeechTimer ++;
    my @TtsMsg;

    if ( &MkOsdIsValid() )
        {
        # Gueltige OSD Daten

        if ( $SpeechTimer % $StatusInterval == 0 )
            {
            #
            # give system status, low prio messages
            #

            # clear old low prio messages
            &TtsClear('LOW');

            # get messages from configuration
            for ( $Message = 1; $Message < 10; $Message ++)
                {
                my $MsgId = sprintf ("Message%d", $Message);
                my $Msg = $Cfg->{'tts'}{$MsgId};
   
                if ( $Msg =~ /FLIGHT_TIME/i )
                    {
                    # Flight time
                    &TtsSpeak ('LOW', "&TtsFlightTime()");
                    }

                elsif ( $Msg =~ /BATTERY/i )
                    {
                    # Battery
                    &TtsSpeak ('LOW', "&TtsUBat()");
                    }

                elsif ( $Msg =~ /ALTITUDE/i )
                    {
                    # Altitude
                    &TtsSpeak ('LOW', "&TtsAltitude()");
                    }

                elsif ( $Msg =~ /SATELLITES/i )
                    {
                    # Satellites
                    &TtsSpeak ('LOW', "&TtsSatsInUse()");
                    }

                elsif ( $Msg =~ /HOME_DIST/i )
                    {
                    # Home Distance
                    &TtsSpeak ('LOW', "&TtsHomeDist()");
                    }

                elsif ( $Msg =~ /TARGET_DIST/i )
                    {
                    # Target Distance
                    &TtsSpeak ('LOW', "&TtsTargetDist()");
                    }

                elsif ( $Msg ne "" )
                    {
                    # text or perl code
                    &TtsSpeak ('LOW', eval "$Msg" );
                    }
                }
            }

       # high prio messages
       if ( $SpeechTimer % 5 == 0 )
            {
            if ( $System{'BatWarning'} )          { push @TtsMsg, $Translate{'TtsBatteryWarning'}; }
            if ( $System{'RCQuality'} eq "WEAK" ) { push @TtsMsg, $Translate{'TtsRcWeak'}; }
            if ( $System{'RCQuality'} eq "NO" )   { push @TtsMsg, $Translate{'TtsRcNo'}; }
            if ( $System{'CrossingBorder'} )      { push @TtsMsg, $Translate{'TtsCrossingBorder'}; }
            if ( $System{'OutsideBorder'} )       { push @TtsMsg, $Translate{'TtsOutsideAirfield'}; }
            if ( $System{'RangeWarning'} )        { push @TtsMsg, $Translate{'TtsRange'}; }
            if ( $System{'TargetNotReachable'} )  { push @TtsMsg, $Translate{'TtsTargetNotReachable'}; }
            }
        }
    else
        {
        # no data link
        if ( $SpeechTimer % 5 == 0 )
            {
            push @TtsMsg, $Translate{'TtsNoDataLink'};
            }
        }

    # speak high prio messages 
    if ( scalar @TtsMsg > 0 )
        {
        # Clear pending messsages
        &TtsClear('HIGH');

        # Speak collected messages
        foreach $Msg (@TtsMsg)
            {
            &TtsSpeak ('HIGH', $Msg);
            }
        }

    # performance profiling
    my ($t1_s, $t1_us) = gettimeofday;
    $Profiler{'Timer_TTS'} = ($t1_s - $t0_s) * 1000000 + $t1_us - $t0_us;
    });


#	
# Timer: 0.5s - Log change of values
#

my @LogState = ( # Variable               # Current Value                  # Last Value                       # Bitmask
                 'NoDataLink',            \$System{'NoDataLink'},          $LogState{'NoDataLink'},            "",
                 'BatWarning',            \$System{'BatWarning'},          $LogState{'BatWarning'},            "",
                 'RCQuality',             \$System{'RCQuality'},           $LogState{'RCQuality'},             "",
                 'CrossingBorder',        \$System{'CrossingBorder'},      $LogState{'CrossingBorder'},        "",
                 'OutsideBorder',         \$System{'OutsideBorder'},       $LogState{'OutsideBorder'},         "",
                 'RangeWarning',          \$System{'RangeWarning'},        $LogState{'RangeWarning'},          "",
                 'TargetNotReachable',    \$System{'TargetNotReachable'},  $LogState{'TargetNotReachable'},    "",
                 'JoystickMoveX',         \$System{'JoystickMoveX'},       $LogState{'JoystickMoveX'},         "",
                 'JoystickMoveY',         \$System{'JoystickMoveY'},       $LogState{'JoystickMoveY'},         "",
                 'JoystickGier',          \$System{'JoystickGier'},        $LogState{'JoystickGier'},          "",
                 'JoystickAlt',           \$System{'JoystickAlt'},         $LogState{'JoystickAlt'},           "",
                 'CurPos_Stat',           \$MkOsd{'CurPos_Stat'},          $LogState{'CurPos_Stat'},           "",
                 'TargetPos_Stat',        \$MkOsd{'TargetPos_Stat'},       $LogState{'TargetPos_Stat'},        "",
                 'TargetPos_Lat',         \$MkOsd{'TargetPos_Lat'},        $LogState{'TargetPos_Lat'},         "",
                 'TargetPos_Lon',         \$MkOsd{'TargetPos_Lon'},        $LogState{'TargetPos_Lon'},         "",
                 'HomePos_Stat',          \$MkOsd{'HomePos_Stat'},         $LogState{'HomePos_Stat'},          "",
                 'SatsInUse',             \$MkOsd{'SatsInUse'},            $LogState{'SatsInUse'},             "",

                 'MKFlags.MotorRun',      \$MkOsd{'MKFlags'},              $LogState{'MKFlags.MotorRun'},      0x01,
                 'MKFlags.Fly',           \$MkOsd{'MKFlags'},              $LogState{'MKFlags.Fly'},           0x02,
                 'MKFlags.Calibrate',     \$MkOsd{'MKFlags'},              $LogState{'MKFlags.Calibrate'},     0x04,
                 'MKFlags.Start',         \$MkOsd{'MKFlags'},              $LogState{'MKFlags.Fly'},           0x08,
                 'MKFlags.Emergency',     \$MkOsd{'MKFlags'},              $LogState{'MKFlags.Emergency'},     0x10,
                 'MKFlags.LowBat',        \$MkOsd{'MKFlags'},              $LogState{'MKFlags.LowBat'},        0x20,
                 'MKFlags.VarioTrimUp',   \$MkOsd{'MKFlags'},              $LogState{'MKFlags.VarioTrimUp'},   0x40,
                 'MKFlags.VarioTrimDown', \$MkOsd{'MKFlags'},              $LogState{'MKFlags.VarioTrimDown'}, 0x80,

                 'MKFlags2.Carefree',     \$MkOsd{'MKFlags2'},             $LogState{'MKFlags.Carefree'},      0x01,
                 'MKFlags2.AltCtrl',      \$MkOsd{'MKFlags2'},             $LogState{'MKFlags.AltCtrl'},       0x02,

                 'NCFlags.Free',          \$MkOsd{'NCFlags'},              $LogState{'NCFlags.Free'},          0x01,
                 'NCFlags.PH',            \$MkOsd{'NCFlags'},              $LogState{'NCFlags.PH'},            0x02,
                 'NCFlags.CH',            \$MkOsd{'NCFlags'},              $LogState{'NCFlags.CH'},            0x04,
                 'NCFlags.RangeLimit',    \$MkOsd{'NCFlags'},              $LogState{'NCFlags.RangeLimit'},    0x08,
                 'NCFlags.NoSerialLink',  \$MkOsd{'NCFlags'},              $LogState{'NCFlags.NoSerialLink'},  0x10,
                 'NCFlags.TargetReached', \$MkOsd{'NCFlags'},              $LogState{'NCFlags.TargetReached'}, 0x20,
                 'NCFlags.ManualControl', \$MkOsd{'NCFlags'},              $LogState{'NCFlags.ManualControl'}, 0x40,
                 'NCFlags.GpsOK',         \$MkOsd{'NCFlags'},              $LogState{'NCFlags.GpsOK'},         0x80,
               );

$frame_map_top->repeat (500, sub
    {
    # performance profiling
    my ($t0_s, $t0_us) = gettimeofday;

    lock (%MkOsd);  # until end of Block

    # Check variable change
    my $i = 0;
    for $State (0 .. $#LogState/4)
        {
        my $bDefined = 0;
        if ( defined $LogState[$i+2] )
            {
            $bDefined = 1;
            }

        $Variable =  $LogState[$i];
        $CurValue =  ${$LogState[$i+1]};
        $LastValue = $LogState[$i+2];
        $BitMask =   $LogState[$i+3];

        $LogState[$i+2] = $CurValue;

        if ( $BitMask ne "" )
            {
            $CurValue  &= $BitMask;
            $LastValue &= $BitMask;
            }

        if ( $bDefined  and  $CurValue ne $LastValue )
            {
            &Log ("ValueChange $Variable: $LastValue -> $CurValue", "V");
            }

        $i += 4;
        }

    # performance profiling
    my ($t1_s, $t1_us) = gettimeofday;
    $Profiler{'Timer_Log'} = ($t1_s - $t0_s) * 1000000 + $t1_us - $t0_us;
    });


#
#  Event engine
#

$frame_map_top->repeat (50, sub     # 50ms
    {
    # performance profiling
    my ($t0_s, $t0_us) = gettimeofday;

    foreach $Key (keys %{$Event})
        {
        if ( ref $Event->{$Key} eq "" )
            {
            # ignore: CreationDate, Version
            next;
            }

        my $Active      = $Event->{$Key}->{'Active'};
        my $Trigger     = $Event->{$Key}->{'Trigger'};       # RISE, FALL, TOGGLE_RISE, TOGGLE_FALL, TRUE, FALSE
        my $Condition   = $Event->{$Key}->{'Condition'};     # Perl commands, Return 1, if Condition is true
        my $Action      = $Event->{$Key}->{'Action'};        # Perl commands
        my $ActionElse  = $Event->{$Key}->{'ActionElse'};    # Perl commands
        my $Delay       = $Event->{$Key}->{'Delay'};         # Action rearm delay in ms
        my $Repeat      = $Event->{$Key}->{'Repeat'};        # Action repeat time in  ms
        my $RepeatElse  = $Event->{$Key}->{'RepeatElse'};    # ActionElse repeat time in ms

        my $DoAction = 0;

        if ( $Active =~ /y/i  and  $Condition ne "" )
            {

            # Provide info about current Event
            %MyEvent = %{$Event->{$Key}};
            $MyEvent{'EventName'} = $Key;

            # lock all shared hashes, which might me used in eval-code
            lock (%MkOsd);           # until end of block
            lock (%MkSerialChannel); # until end of block
            lock (%Stick);           # until end of block
            lock (%MkNcDebug);       # until end of block

            # compile and execute condition at runtime. Return 1, if Condition is true
            my $CondStat = eval "$Condition";    
            if ( $@ ne "" )
                {
                # print compiler message to STDOUT
                print "Event $Key: Condition: $@";
                }

            # current time for Delay, Repeat timing
            my ($t0_s, $t0_us) = gettimeofday;

            # for Condition edge detection
            my $LastCondStat = $EventStat{$Key}{'LastCondValue'};
            $EventStat{$Key}{'LastCondValue'} = $CondStat;

            # some flags
            my $Rise  = (! $LastCondStat  and    $CondStat);
            my $Fall  = (  $LastCondStat  and  ! $CondStat);
            my $True  =   $CondStat;
            my $False = ! $CondStat;

            # Delay time check
            my $DelayCheck = 1;
            my $t1_s  =  $EventStat{$Key}{'DelayTime_s'};
            my $t1_us =  $EventStat{$Key}{'DelayTime_us'};
            if ( $Delay ne "" and  $t1_s ne ""  and  $t1_us ne "" )
                {
                my $Diff_ms = ($t0_s - $t1_s) * 1000 + ($t0_us - $t1_us) / 1000;
                if ( $Diff_ms < $Delay )
                    {
                    $DelayCheck = 0;
                    }
                }

            #
            # rising or falling edge
            #
            if ( ($Trigger eq "RISE" and  $Rise)  or
                 ($Trigger eq "FALL" and  $Fall) )
                {
                $DoAction = $DelayCheck;

                if ( $DoAction )
                    {
                    $EventStat{$Key}{'DelayTime_s'}  = $t0_s;
                    $EventStat{$Key}{'DelayTime_us'} = $t0_us;

                    # reset Repeat time when event gets active
                    $EventStat{$Key}{'RepeatTime_s'}  = 0;
                    $EventStat{$Key}{'RepeatTime_us'} = 0;

                    # reset event counter
                    $EventStat{$Key}{'EventCnt'} = 0;
                    }
                }

            #
            # toggle on rising/falling edge
            #
            elsif ( $Trigger =~ /TOGGLE/i )
                {
                $DoAction = 0;
                if ( ($Rise  and  $Trigger eq "TOGGLE_RISE")  or
                     ($Fall  and  $Trigger eq "TOGGLE_FALL") )
                    {
                    $DoAction = $DelayCheck;

                    if ( $DoAction )
                        {
                        $EventStat{$Key}{'DelayTime_s'}  = $t0_s;
                        $EventStat{$Key}{'DelayTime_us'} = $t0_us;

                        # reset Repeat time when event gets active
                        $EventStat{$Key}{'RepeatTime_s'}  = 0;
                        $EventStat{$Key}{'RepeatTime_us'} = 0;

                        # reset event counter
                        $EventStat{$Key}{'EventCnt'} = 0;
                        }
                    }

                if ( $DoAction )
                    {
                    $EventStat{$Key}{'ToggleOnOff'} ^= 1;
                    }

                $DoAction = $EventStat{$Key}{'ToggleOnOff'};
                }


            #
            # TRUE Condition
            #
            elsif ( $Trigger eq "TRUE" )
                {
                if ( $True )
                    {
                    $DoAction = $DelayCheck;

                    if ( $DoAction and $Rise)
                        {
                        # reset Repeat time when event gets active
                        $EventStat{$Key}{'RepeatTime_s'}  = 0;
                        $EventStat{$Key}{'RepeatTime_us'} = 0;

                        # reset event counter
                        $EventStat{$Key}{'EventCnt'} = 0;
                        }
                    }

                if ( $Fall )
                    {
                    # set Delay reference from falling edge

                    if ( $DelayCheck )
                        {
                        # timestamp of last falling Condition edge change
                        $EventStat{$Key}{'DelayTime_s'}  = $t0_s;
                        $EventStat{$Key}{'DelayTime_us'} = $t0_us;

                        # reset event counter
                        $EventStat{$Key}{'EventCnt'} = 0;
                        }
                    }
                }

            #
            # FALSE Condition
            #
            elsif ( $Trigger eq "FALSE" )
                {
                if ( $False )
                    {
                    $DoAction = $DelayCheck;

                    if ( $DoAction and $Fall)
                        {
                        # reset Repeat time when event gets active
                        $EventStat{$Key}{'RepeatTime_s'}  = 0;
                        $EventStat{$Key}{'RepeatTime_us'} = 0;

                        # reset event counter
                        $EventStat{$Key}{'EventCnt'} = 0;
                        }
                    }

                if ( $Rise )
                    {
                    # set Delay reference from rising edge

                    if ( $DelayCheck )
                        {
                        # timestamp of last rising Condition edge change
                        $EventStat{$Key}{'DelayTime_s'}  = $t0_s;
                        $EventStat{$Key}{'DelayTime_us'} = $t0_us;

                        # reset event counter
                        $EventStat{$Key}{'EventCnt'} = 0;
                        }
                    }
                }

            else
                {
                # unkown, fall through
                }


            # undefined Status, neither ON, nor OFF
            $EventStat{$Key}{'Status'} = "";

            #
            # Process Action
            #
            if ( $DoAction  and  $Action ne "" )
                {
                $EventStat{$Key}{'Status'} = "ON";
                my $Execute = 1;

                my $t1_s  =  $EventStat{$Key}{'RepeatTime_s'};
                my $t1_us =  $EventStat{$Key}{'RepeatTime_us'};
                if ( $Repeat ne "" and  $t1_s ne ""  and  $t1_us ne "" )
                    {
                    my $Diff_ms = ($t0_s - $t1_s) * 1000 + ($t0_us - $t1_us) / 1000;
                    if ( $Diff_ms < $Repeat )
                        {
                        $Execute = 0;
                        }
                    }

                if ( $Execute )
                    {
                    # Accounting
                    $MyEvent{'EventCnt'} = $EventStat{$Key}{'EventCnt'} ++;

                    # compile and execute action at runtime
                    eval "$Action";
                    if ( $@ ne "" )
                        {
                        # print compiler message to STDOUT
                        print "Event $Key: Action: $@";
                        }

                    # Timestamp, when Action was executed
                    $EventStat{$Key}{'RepeatTime_s'}  = $t0_s;
                    $EventStat{$Key}{'RepeatTime_us'} = $t0_us;

                    # reset ActionElse repeat timing
                    $EventStat{$Key}{'RepeatTimeElse_s'}  = 0;
                    $EventStat{$Key}{'RepeatTimeElse_us'} = 0;
                    }
                }

            #
            # Process ActionElse
            #
            if (  ! $DoAction  and  $ActionElse ne "" )
                {
                $EventStat{$Key}{'Status'} = "OFF";
                my $Execute = 1;

                # check repeat time
                my $t1_s  =  $EventStat{$Key}{'RepeatTimeElse_s'};
                my $t1_us =  $EventStat{$Key}{'RepeatTimeElse_us'};

                if ( $RepeatElse ne "" and  $t1_s ne ""  and  $t1_us ne "" )
                    {
                    my $Diff_ms = ($t0_s - $t1_s) * 1000 + ($t0_us - $t1_us) / 1000;
                    if ( $Diff_ms < $RepeatElse )
                        {
                        $Execute = 0;
                        }
                    }

                if ( $Execute )
                    {
                    # Accounting
                    $MyEvent{'EventCnt'} = $EventStat{$Key}{'EventCnt'} ++;

                    # compile and execute action at runtime
                    eval "$ActionElse";
                    if ( $@ ne "" )
                        {
                        # print compiler message to STDOUT
                        print "Event $Key: ActionElse: $@";
                        }

                    # Timestamp, when ActionElse was executed
                    $EventStat{$Key}{'RepeatTimeElse_s'}  = $t0_s;
                    $EventStat{$Key}{'RepeatTimeElse_us'} = $t0_us;
                    }
                }
            }

        else
            {
            $EventStat{$Key}{'Status'} = "DISABLED";
            }
        }

    # performance profiling
    my ($t1_s, $t1_us) = gettimeofday;
    $Profiler{'Timer_Event'} = ($t1_s - $t0_s) * 1000000 + $t1_us - $t0_us;
    });


#
# Send Serial Channel + stick button/analog handling
#

my $SerialChannelTiming = &Default ($Cfg->{'serialchannel'}->{'SerialChannelFrquency'}, 20);
$SerialChannelTiming = int (1 / $SerialChannelTiming * 1000);
if ( $SerialChannelTiming < 10 )
    {
    $SerialChannelTiming = 10;
    }

# Timer: variable timing
$frame_map_top->repeat ($SerialChannelTiming, sub
    {

    # performance profiling
    my ($t0_s, $t0_us) = gettimeofday;

    # 12 serial Channel
    for ( my $ChannelInd = 0; $ChannelInd < 12; $ChannelInd ++)
        {
        my $Channel = sprintf ("SerialChannel%02d", $ChannelInd + 1);  # key for Cfg-hash

        my $Control = $Cfg->{'serialchannel'}->{$Channel};
        if ( $Control ne "" )
            {
            my $Expo    = $Cfg->{'serialchannel'}->{$Channel . "Expo"};
            my $Limit   = $Cfg->{'serialchannel'}->{$Channel . "Limit"};
            my $ChannelVal = &ParseControls($Channel, $Control, $Expo, $Limit, $SerialChannelTiming);

            &SerialChannel($ChannelInd, $ChannelVal);
            }
        }

    $MkSerialChannel{'SerialChannelSend'}   = $Cfg->{'serialchannel'}->{'SerialChannelSend'};
    $MkSerialChannel{'SerialChannelTiming'} = $SerialChannelTiming;

    # send serial Channel to MK, nicht wenn deaktiviert  oder WP uebertragen werden
    if ( ! $MkSendWp  and  $Cfg->{'serialchannel'}->{'SerialChannelSend'} =~ /y/i  and  $TxExtOn == 1)
        {
        &SendSerialChannel();  # send values in %MkSerialChannel
        }

    # performance profiling
    my ($t1_s, $t1_us) = gettimeofday;
    $Profiler{'Timer_SerialChannel'} = ($t1_s - $t0_s) * 1000000 + $t1_us - $t0_us;
    });


#
# Extern Control
#

my $ExternControlTiming = &Default ($Cfg->{'externcontrol'}->{'ExternControlFrquency'}, 20);
$ExternControlTiming = int (1 / $ExternControlTiming * 1000);
if ( $ExternControlTiming < 10 )
    {
    $ExternControlTiming = 10;
    }

# Timer: variable timing
$frame_map_top->repeat ($ExternControlTiming, sub
    {
    # performance profiling
    my ($t0_s, $t0_us) = gettimeofday;

    foreach my $Channel ( qw(ExternControlRoll ExternControlNick ExternControlGier ExternControlGas ExternControlHeight) )
        {
        my $Control = $Cfg->{'externcontrol'}->{$Channel};
        if ( $Control ne "" )
            {
            my $Expo    = $Cfg->{'externcontrol'}->{$Channel . "Expo"};
            my $Limit   = $Cfg->{'externcontrol'}->{$Channel . "Limit"};
            $MkExternControl{$Channel} = &ParseControls($Channel, $Control, $Expo, $Limit, $ExternControlTiming);
            }
        }

    $MkExternControl{'ExternControlSend'}   = $Cfg->{'externcontrol'}->{'ExternControlSend'};
    $MkExternControl{'ExternControlTimimg'} = $ExternControlTiming;

    # send extern control to MK, nicht wenn deaktiviert oder WP uebertragen werden
    if ( ! $MkSendWp  and  $Cfg->{'externcontrol'}->{'ExternControlSend'} =~ /y/i  and  $TxExtOn == 1 )
        {
        &SendExternalControl ( '-nick'   => $MkExternControl{'ExternControlNick'},
                               '-roll'   => $MkExternControl{'ExternControlRoll'},
                               '-gier'   => $MkExternControl{'ExternControlGier'},
                               '-gas'    => $MkExternControl{'ExternControlGas'},
                               '-height' => $MkExternControl{'ExternControlHeight'},
                               '-remotebuttons' => 0,
                               '-config' => 1,
                               '-frame'  => 1,
                             );

        $MkExternControl{'_Timestamp'} = time;   # time when command was sent
        }

    # performance profiling
    my ($t1_s, $t1_us) = gettimeofday;
    $Profiler{'Timer_ExternControl'} = ($t1_s - $t0_s) * 1000000 + $t1_us - $t0_us;
    });


#
# mouse/joystick crosshair move/gier
#

$frame_map_top->repeat (50, sub     # 50ms
    {
    # performance profiling
    my ($t0_s, $t0_us) = gettimeofday;

    # crosshair center is current position
    my $Crosshair_Lat = $MkOsd{'CurPos_Lat'};
    my $Crosshair_Lon = $MkOsd{'CurPos_Lon'};

    if ( $PlayerMode eq 'Pause'  and 
         $PlayerPause_Lat ne ""  and  $PlayerPause_Lon ne "" )
        {
        my $ControlX = $Cfg->{'map'}->{'CrosshairMoveX'};
        my $ControlY = $Cfg->{'map'}->{'CrosshairMoveY'};

        if ( $ControlX ne ""  and $ControlY ne "" )
            {
            my $StickRange = 250;
            my $Bias = $StickRange /2;

            my $ExpoX    = $Cfg->{'map'}->{'CrosshairMoveXExpo'};
            my $LimitX   = $Cfg->{'map'}->{'CrosshairMoveXLimit'};
            my $SpeedX  = &ParseControls('CrosshairMoveX', $ControlX, $ExpoX, $LimitX, 50);

            my $ExpoY    = $Cfg->{'map'}->{'CrosshairMoveYExpo'};
            my $LimitY   = $Cfg->{'map'}->{'CrosshairMoveYLimit'};
            my $SpeedY  = &ParseControls('CrosshairMoveY', $ControlY, $ExpoY, $LimitY, 50);

            $System{'JoystickMoveX'} = $SpeedX;
            $System{'JoystickMoveY'} = $SpeedY;

            my $BearingTop = &MapAngel() - 90.0;
            my $BearingMouse = rad2deg atan2($SpeedX, $SpeedY);
            my $Bearing = $BearingTop + $BearingMouse;

            if ( $PlayerPauseMode eq "MK" )
                {
                # MK Reference
                $Bearing = $MkOsd{'CompassHeading'} + $BearingMouse;
                }

            # max. 30m/s at 50ms frame time
            my $Speed = sqrt ($SpeedX * $SpeedX + $SpeedY * $SpeedY); 
            my $Dist = 40*50/1000 * $Speed / $Bias;
            ($PlayerPause_Lat, $PlayerPause_Lon) = &MapGpsAt($PlayerPause_Lat, $PlayerPause_Lon, $Dist, $Bearing);

            if ( $SpeedX != 0  or  $SpeedY != 0 )
                {
                $CrosshairTimerCnt = 0;    # stick move - restart timer
                }
            }

        # Crosshair center is pause position
        $Crosshair_Lat = $PlayerPause_Lat;
        $Crosshair_Lon = $PlayerPause_Lon;
        }

    #
    # show/update Crosshair for 5 sec after move or gier
    #
    if ( $CrosshairTimerCnt < 0 )
       {
       # disabled
       &CrosshairHide();
       }
    elsif ( $CrosshairTimerCnt < 5 * 1000/50 )
       {
       my $Heading = $System{'GierSetpoint'};
       if ( $Heading < 0 )
           {
           $Heading = $System{'PoiBearing'};
           }
       &CrosshairShow ($Crosshair_Lat, $Crosshair_Lon, $Heading);
       }
    elsif ( $CrosshairTimerCnt ==  5 * 1000/50 )
       {
       # timeout reached
       &CrosshairHide();
       }

    $CrosshairTimerCnt ++;

    # performance profiling
    my ($t1_s, $t1_us) = gettimeofday;
    $Profiler{'Timer_Crosshair'} = ($t1_s - $t0_s) * 1000000 + $t1_us - $t0_us;
    });


#
# mouse/joystick gier/alt control
#

$frame_map_top->repeat (100, sub     # 100ms
    {
    # performance profiling
    my ($t0_s, $t0_us) = gettimeofday;

    #
    # Gier
    #
    my $ControlGier = $Cfg->{'map'}->{'ControlGier'};
    if ( $ControlGier ne "" )
        {
        my $Gier  = &ParseControls('Gier', $ControlGier, "", "", 100);

        if ( $Gier != 0 )
            {
            # we have a stick move

            $System{'JoystickGier'} = $Gier;
        
            my $Heading = $System{'GierSetpoint'};
            if ( $Heading < 1  or  $Heading > 360 )
                {
                $Heading = $MkOsd{'CompassHeading'};
                }

            $Gier = $Gier * 90/125 * 100/1000;   # max. 90 Grad/sec
            $Gier = &NormBearing($Heading + $Gier);

            # send Heading at once, for faster response
            if ( $Gier != $TimerGierLast )
                {
                &NcSetParam(1, $Gier);
                $System{'GierSetpoint'} = $Gier;
                }
            $TimerGierLast = $Gier;

            $CrosshairTimerCnt = 0;   # show crosshair
            }
        }

    #
    # Altitude
    #
    my $ControlAlt = $Cfg->{'map'}->{'ControlAlt'};
    if ( $ControlAlt ne "" )
        {
        my $Alt  = &ParseControls('Altitude', $ControlAlt, "", "", 100);

        if ( $Alt != 0 )
            {
            # we have a stick move
            $System{'JoystickAlt'} = $Alt;
            $System{'AltSetpoint'} = $System{'AltSetpoint'} + $Alt * 20/125 * 100/1000;   # max. 20 m/sec

            $CrosshairTimerCnt = 0;   # show crosshair
            }
        }

    # performance profiling
    my ($t1_s, $t1_us) = gettimeofday;
    $Profiler{'Timer_GierAlt'} = ($t1_s - $t0_s) * 1000000 + $t1_us - $t0_us;
    });



__END__
