#!/usr/bin/perl
#!/usr/bin/perl -d:ptkdb

###############################################################################
#
# mkcomm.pl -  MK Communication Routines
#
# Copyright (C) 2009  Rainer Walther  (rainerwalther-mail@web.de)
#
# Creative Commons Lizenz mit den Zusaetzen (by, nc, sa)
#
# Es ist Ihnen gestattet: 
#     * das Werk vervielfltigen, verbreiten und ffentlich zugnglich machen
#     * Abwandlungen bzw. Bearbeitungen des Inhaltes anfertigen
# 
# Zu den folgenden Bedingungen:
#     * Namensnennung.
#       Sie mssen den Namen des Autors/Rechteinhabers in der von ihm festgelegten Weise nennen.
#     * Keine kommerzielle Nutzung.
#       Dieses Werk darf nicht fr kommerzielle Zwecke verwendet werden.
#     * Weitergabe unter gleichen Bedingungen.
#       Wenn Sie den lizenzierten Inhalt bearbeiten oder in anderer Weise umgestalten,
#       verndern oder als Grundlage fr einen anderen Inhalt verwenden,
#       drfen Sie den neu entstandenen Inhalt nur unter Verwendung von Lizenzbedingungen
#       weitergeben, die mit denen dieses Lizenzvertrages identisch oder vergleichbar sind.
# 
# Im Falle einer Verbreitung mssen Sie anderen die Lizenzbedingungen, unter welche dieses
# Werk fllt, mitteilen. Am Einfachsten ist es, einen Link auf diese Seite einzubinden.
# 
# Jede der vorgenannten Bedingungen kann aufgehoben werden, sofern Sie die Einwilligung
# des Rechteinhabers dazu erhalten.
# 
# Diese Lizenz lsst die Urheberpersnlichkeitsrechte unberhrt.
# 
# Weitere Details zur Lizenzbestimmung gibt es hier:
#   Kurzform: http://creativecommons.org/licenses/by-nc-sa/3.0/de/
#   Komplett: http://creativecommons.org/licenses/by-nc-sa/3.0/de/legalcode
#
###############################################################################
# 2009-02-21 0.0.1 rw created
# 2009-03-18 0.0.2 rw NC 0.14e
# 2009-04-01 0.1.0 rw RC1
# 2009-04-06 0.1.1 rw NC 0.15c
# 2009-05-16 0.1.2 rw External control
# 2009-08-15 0.1.3 rw SIG-Handler removed
# 2009-10-05 0.1.4 rw COM-Ports > 9 + PortSetSkip configuration
#                     Export Target-Hash to Simulator
# 2009-10-24 0.3.0 rw NC 0.17b
# 2010-02-10 0.4.0 rw Port read timout reduced from 100ms to 10ms
#                     Serial Channel
#                     Extern Control
#                     Controls via joystick or 3D-Mouse
#                     Navidata Current, Capacity
# 2010-02-18 0.4.1 rw LCD Stick/Poti
# 2010-03-07 0.4.2 rw MKFLAG: LOWBAT
# 2010-07-01 0.5.0 rw WP/POI adjustments for NC 0.19/0.20
#                     Hardware Error Codes
# 2010-10-16 0.5.4 rw Spline speed controlled mode (SPD)
# 2010-11-03 0.6.0 rw NC 0.21, DGPS
# 2011-01-01 0.7.0 rw Switch UART to NC, FC, GPS
#                     Read MK Settings
#                     WP_EventChannelValue in WP/Target command added
#                     NC 0.24
#                     Set NC-Parameter command added
#                     Gier Control
#                     WP Alt control by NC
#                     Port read timout changed from 10ms to 20ms
#                     NC Connection checker "Z" added
# 2011-07-31 0.8.0 rw NC 0.26, FC 0.86 settings structure
# 2011-10-18 0.8.1 rw NaviSpeedCompensation --> NaviAccCompensation
#                     RC_FAILSAVE_ACTIVE
# 2012-03-25 0.9.0 rw NC 0.28, FC 0.88
#                     Motor Current
#                     WP Name
# 2013-06-13 0.9.1 rw NC 0.31, FC 0.91
# 2014-01-25 0.9.2 rw FC 2.02
#
###############################################################################

$Version{'mkcomm.pl'} = "0.9.2 - 2014-01-25";

# MK Protokoll
#   http://www.mikrokopter.de/ucwiki/en/SerialCommands
#   http://www.mikrokopter.de/ucwiki/en/SerialProtocol

#
# Parameter
#

# Com Port of MK Comm-Device (BT, WI.232)
if ( ! defined $Cfg->{'mkcomm'}->{'Port'} )
    {
    # set default
    $Cfg->{'mkcomm'}->{'Port'} = "COM5";
    }

$AddrFC     = "b";
$AddrNC     = "c";
$AddrMK3MAG = "d";


if ( $0 =~ /mkcomm.pl$/i )
    {
    # Program wurde direkt aufgerufen

    # change working directory to program path
    my $Cwd = substr ($0, 0, rindex ($0, "mkcomm.pl"));
    chdir $Cwd;

    # set path for local Perl libs
    push @INC, $Cwd . "perl/lib";
    }


# Packages
use threads;                # http://search.cpan.org/~jdhedden/threads-1.72/threads.pm
                            # http://perldoc.perl.org/threads.html
use threads::shared;        # http://search.cpan.org/~jdhedden/threads-shared-1.28/shared.pm
use Thread::Queue;          # http://search.cpan.org/dist/Thread-Queue-2.11/lib/Thread/Queue.pm
use Time::HiRes qw(usleep); # http://search.cpan.org/~jhi/Time-HiRes-1.9719/HiRes.pm
if ( $^O =~ /Win32/i )
    {
    require Win32::SerialPort;  # http://search.cpan.org/dist/Win32-SerialPort
    }
else
    {
    require Device::SerialPort; # http://search.cpan.org/~cook/Device-SerialPort-1.04/SerialPort.pm
    }

require "libmap.pl";

# Hashes exported to other threads and main-program
share (%MkOsd);
share (%MkTarget);
share (%MkNcDebug);
share (%Mk);
share (%MkSetting);
share (%MkSSim);     # Target info for simulator
share (%MkSerialChannel);
share (%MkMotor);


$Mk{'_RxCrcError'} = 0;    # statistics

# Queue for Sending to MK
$MkSendQueue = Thread::Queue->new();

sub MkCommExit()
    {	
    # close COM port
    &MkClose();

    if ( defined threads->self() )
        {
        threads->exit();
        }
    exit;
    }


sub MkInit()
    {
    if ( defined $MkPort )
        {
        return;   # already open
        }

    # open COM-Port		
    my $MkComPort = $Cfg->{'mkcomm'}->{'Port'};



$MkComPort = "COM5";
print "\nWARNING: mkcomm.pl: Set MkComPort = $MkComPort\n\n";



     if ( $MkComPort =~ /^COM/i )
         {
         $MkComPort = "\\\\.\\" . $MkComPort;   # \\.\COMnn for nn > 9
         }

    undef $MkPort;
    if ( $^O =~ m/Win32/ )
        {
        $MkPort = Win32::SerialPort->new ($MkComPort) || die "Error open $MkComPort\n";
        }
    else
        {
        $MkPort = Device::SerialPort->new ($MkComPort) || die "Error open $MkComPort\n";
        }

    if ( ! ($Cfg->{'mkcomm'}->{'PortSetSkip'} =~ /y/i) )
        {
        # Set COM parameters, don't set for Bluetooth device
        $MkPort->baudrate(57600);
        $MkPort->parity("none");
        $MkPort->databits(8);
        $MkPort->stopbits(1);
        $MkPort->handshake('none');
        $MkPort->write_settings;
        }
	
    $MkPort->read_const_time(20); # total = (avg * bytes) + const (ms)
    }

# Read one line from MK
# Check send-queue
sub MkIOLine()
    {
    # Init serial port
    &MkInit();
	
    my $RxLine = "";		
    while ( 1 )
        {	
        # Check Send-Queue
        my $Items = $MkSendQueue->pending();
        if ( $Items >= 3 )  # Cmd, Addr, Data
            {
            my ($Id, $Addr, $Data) = $MkSendQueue->dequeue(3);
            &MkSend ($Id, $Addr, $Data);	
            }
		
        # Zeichenweise lesen, blockierend mit Timeout
        my ($RxLen, $RxChar) = $MkPort->read(1);   
        if ( $RxLen == 1 )
            {
            $Mk{'_BytesRx'} ++;    # Statistics

            if ( "$RxChar" eq "#" )       # 1st char of line
                {
                $RxLine = "#";
                }
            elsif ( "$RxChar" eq "\r" )   # last char of line
                {
                return ($RxLine);
                }
            else
                {
                $RxLine = "$RxLine" . "$RxChar";    # collect char
                }
            }
        }
    }


# Read and decode a command from MK
# process send queue in &MkIOLine()
sub MkIO()
    {	
    my $RxData = &MkIOLine();     # Blocking Read for complete line
	
    # Zeile decodieren
    if ( substr ($RxData, 0, 1) eq '#' )
        {			
        # Zeile decodieren
        $Header = substr($RxData, 0, 3);
        $Chksum = substr($RxData, -2);
        $Data = substr($RxData, 3, length ($RxData) -5);

        # CRC prfen
        if ( &CrcCheck ("$Header" . "$Data", $Chksum ) )
            {
            # Base64 decodieren
            $Data = &Decode64($Data);

            # Daten auswerten und in shared Hash schreiben
            if ( &ProcessRx($Header, $Data) )
                {
                return 1;    # alles OK
                }
            }
        else
            {
            $Mk{'_RxCrcError'} ++;    # Statistics
            }
        }
		
    return 0;  # keine Daten empfangen
    }


# Send a command to MK
sub MkSend()
    {
    my ($Id, $Addr, $Data) = @_;

    # Init serial port
    &MkInit();

    my $TxSend = $Data;

    if ( $Id ne "_DIRECT" )
        {
        my $Base64Data = &Encode64($Data);

        my $TxData = "#" . "$Addr" . "$Id" . "$Base64Data"; 
        my $Crc = &Crc($TxData);
        $TxSend	= "$TxData" . "$Crc" . "\r";
        }

    $Mk{'_BytesTx'} += length $TxSend;     # Statistics

    $MkPort->write($TxSend);
    }


# close COM-Port
sub MkClose()
    {
    undef $MkPort;
    }

	
# CRC Prfung
sub CrcCheck ()
    {
    my ($Data, $Crc) = @_;

    my $Check = &Crc($Data);
    if ( $Check ne $Crc )
        {
        return 0;   # CRC passt nicht
        }	   
    return (1);    # CRC OK
    }

	
# CRC berechnen
sub Crc ()
    {
    my ($Data) = @_;
    my $TmpCrc = 0;
    my $Len = length $Data;

    for ($i=0; $i<$Len; $i++)
        {
        $TmpCrc += ord(substr($Data, $i, 1));
        }
		
    $TmpCrc %= 4096;
    my $Crc1 = ord ("=") + $TmpCrc / 64;
    my $Crc2 = ord ("=") + $TmpCrc % 64;
    $Crc = pack("CC", $Crc1, $Crc2);

    return ($Crc);
    }


# Empfangene Daten decodieren, modifiziertes Base64
sub Decode64()
    {
    my ($DataIn) = @_;

    my $ptrIn  = 0;
    my $DataOut  = "";
    my $len = length ($DataIn);

    while ( $len > 0 )
        {
        $a = ord (substr ($DataIn, $ptrIn ++, 1)) - ord ("=");
        $b = ord (substr ($DataIn, $ptrIn ++, 1)) - ord ("=");
        $c = ord (substr ($DataIn, $ptrIn ++, 1)) - ord ("=");
        $d = ord (substr ($DataIn, $ptrIn ++, 1)) - ord ("=");

        $x = ($a << 2) | ($b >> 4);
        $y = (($b & 0x0f) << 4) | ($c >> 2);
        $z = (($c & 0x03) << 6) | $d;

        foreach $i ( $x, $y, $z )
            {
            if ( $len--)
                {
                my $Tmp = pack ('C1', $i);
                $DataOut = "$DataOut" . "$Tmp";
                }
            else
                {
                last;
                }
            }
        }

    return ($DataOut);
    }

	
# zu sendende Daten codieren, modifiziertes Base64
sub Encode64()
    {
    my ($Data) = @_;

    my $Length = length $Data;
    my $TxBuf = "";
    my $ptr = 0;

    while( $Length > 0 )
        {
        my $a = 0;
        my $b = 0;
        my $c = 0;
        if ($Length) {$a = ord(substr ($Data, $ptr++, $Length--));}
        if ($Length) {$b = ord(substr ($Data, $ptr++, $Length--));}
        if ($Length) {$c = ord(substr ($Data, $ptr++, $Length--));}

        my $ac = ord("=") +    ($a >> 2);
        my $bc = ord("=") + ( (($a & 0x03) << 4) | (($b & 0xf0) >> 4) );
        my $cc = ord("=") + ( (($b & 0x0f) << 2) | (($c & 0xc0) >> 6) );
        my $dc = ord("=") +    ($c & 0x3f);
        $TxBuf = "$TxBuf" . pack ("C4", $ac, $bc, $cc, $dc);
        }
    return ($TxBuf);
    }

	
# Empfangenen Datensatz verarbeiten
sub ProcessRx()
    {
    my ($Header, $Data) = @_;

    my $Adr = substr ($Header, 1, 1);    # b=FC, c=NC, d=MK3MAG
    my $Id  = substr ($Header, 2, 1);

    if ( $Id eq "O" )
        {
        #
        # OSD-Daten nach %MkOsd einlesen
        #
		
        # Struktur Datensatz:
        #     u8  Version                                     // version of the data structure
        #     GPS_Pos_t CurrentPosition;
        #     GPS_Pos_t TargetPosition;
        #     GPS_PosDev_t TargetPositionDeviation;
        #     GPS_Pos_t HomePosition;
        #     GPS_PosDev_t HomePositionDeviation;
        #     u8  WaypointIndex;                              // index of current waypoints running from 0 to WaypointNumber-1
        #     u8  WaypointNumber;                             // number of stored waypoints
        #     u8  SatsInUse;                                  // no of satellites used for position solution
        #     s16 Altimeter;                                  // hight according to air pressure
        #     s16 Variometer;                                 // climb(+) and sink(-) rate
        #     u16 FlyingTime;                                 // in seconds
        #     u8  UBat;                                       // Battery Voltage in 0.1 Volts
        #     u16 GroundSpeed;                                // speed over ground in cm/s (2D)
        #     s16 Heading;                                    // current flight direction in deg as angle to north
        #     s16 CompassHeading;                             // current compass value
        #     s8  AngleNick;                                  // current Nick angle in 1
        #     s8  AngleRoll;                                  // current Rick angle in 1
        #     u8  RC_Quality;                                 // RC_Quality
        #     u8  FCStatusFlags;                              // Flags from FC
        #     u8  NCFlags;                                    // Flags from NC
        #     u8  Errorcode;                                  // 0 --> okay
        #     u8  OperatingRadius                             // current operation radius around the Home Position in m
        #     s16 TopSpeed;                                   // velocity in vertical direction in cm/s
        #     u8  TargetHoldTime;                             // time in s to stay at the given target, counts down to 0 if target has been reached
        #     u8  FCStatusFlags2;                             // Flags from FC
        #     s16 SetpointAltitude;                           // setpoint for altitude
        #     u8  Gas;                                        // for future use
	#     u16 Current;                                    // actual current in 0.1A steps
	#     u16 UsedCapacity;                               // used capacity in mAh

        
        # GPS_Pos_t:
        #     s32 Longitude;   // in 1E-7 deg
        #     s32 Latitude;    // in 1E-7 deg
        #     s32 Altitude;    // in mm
        #     u8  Status;      // validity of data

        # GPS_PosDev_t:
        #     s16 Distance;                                   // distance to target in dm
        #     s16 Bearing;                                    // course to target in deg      

        # Status:
        #     INVALID         = 0
        #     NEWDATA         = 1
        #     PROCESSED       = 2
		
        # FCStatusFlags  0x01: MOTOR_RUN, 0x02: FLY, 0x04: CALIBRATE, 0x08: START, 0x10: EMERGENCY_LANDING
        #                0x20: LOWBAT, 0x40: VARIO_TRIM_UP, 0x80: VARIO_TRIM_DOWN
        # FCStatusFlags2 0x01: CAREFREE, 0x02: ALTITUDE_CONTROL, 0x04: RC_FAILSAVE_ACTIVE
        #                0x08: OUT1_ACTIVE, 0x10: OUT2_ACTIVE
        # NCFlags  0x01: FLAG_FREE, 0x02: FLAG_PH, 0x04: FLAG_CH, 0x08: FLAG_RANGE_LIMIT
        #          0x10: FLAG_NOSERIALLINK, 0x20: FLAG_TARGET_REACHED, FLAG_MANUAL_CONTROL: 0x40
        #          0x80: FLAG_GPS_OK
	 	    
        lock (%MkOsd);  # until end of Block

        (
        $MkOsd{'Version'},
        $MkOsd{'CurPos_Lon'},
        $MkOsd{'CurPos_Lat'},
        $MkOsd{'CurPos_Alt'},
        $MkOsd{'CurPos_Stat'},
        $MkOsd{'TargetPos_Lon'},
        $MkOsd{'TargetPos_Lat'},
        $MkOsd{'TargetPos_Alt'},
        $MkOsd{'TargetPos_Stat'},
        $MkOsd{'TargetPosDev_Dist'},
        $MkOsd{'TargetPosDev_Bearing'},
        $MkOsd{'HomePos_Lon'},
        $MkOsd{'HomePos_Lat'},
        $MkOsd{'HomePos_Alt'},
        $MkOsd{'HomePos_Stat'},
        $MkOsd{'HomePosDev_Dist'},
        $MkOsd{'HomePosDev_Bearing'},
        $MkOsd{'WaypointIndex'},
        $MkOsd{'WaypointNumber'},
        $MkOsd{'SatsInUse'},
        $MkOsd{'Altimeter'},
        $MkOsd{'Variometer'},
        $MkOsd{'FlyingTime'},
        $MkOsd{'UBat'},
        $MkOsd{'GroundSpeed'},
        $MkOsd{'Heading'},
        $MkOsd{'CompassHeading'},
        $MkOsd{'AngleNick'},
        $MkOsd{'AngleRoll'},
        $MkOsd{'RC_Quality'},
        $MkOsd{'MKFlags'},
        $MkOsd{'NCFlags'},
        $MkOsd{'Errorcode'},
        $MkOsd{'OperatingRadius'},
        $MkOsd{'TopSpeed'},
        $MkOsd{'TargetHoldTime'},
        $MkOsd{'MKFlags2'},
        $MkOsd{'SetPointAltitude'},
        $MkOsd{'Gas'},
        $MkOsd{'Current'},
        $MkOsd{'UsedCapacity'},
        ) = unpack ('ClllClllCsslllCssCCCssSCSssccCCCCCsCCsCSs', $Data); 
		
        $MkOsd{'CurPos_Lon'}     = sprintf("%.7f", $MkOsd{'CurPos_Lon'} / 10000000);
        $MkOsd{'CurPos_Lat'}     = sprintf("%.7f", $MkOsd{'CurPos_Lat'} / 10000000);
        $MkOsd{'CurPos_Alt'}     = sprintf("%.3f", $MkOsd{'CurPos_Alt'} / 1000);
        $MkOsd{'TargetPos_Lon'}  = sprintf("%.7f", $MkOsd{'TargetPos_Lon'} / 10000000);
        $MkOsd{'TargetPos_Lat'}  = sprintf("%.7f", $MkOsd{'TargetPos_Lat'} / 10000000);
        $MkOsd{'TargetPos_Alt'}  = sprintf("%.3f", $MkOsd{'TargetPos_Alt'} / 1000);
        $MkOsd{'HomePos_Lon'}    = sprintf("%.7f", $MkOsd{'HomePos_Lon'} / 10000000);
        $MkOsd{'HomePos_Lat'}    = sprintf("%.7f", $MkOsd{'HomePos_Lat'} / 10000000);
        $MkOsd{'HomePos_Alt'}    = sprintf("%.3f", $MkOsd{'HomePos_Alt'} / 1000);
        $MkOsd{'UBat'}           = sprintf("%.1f", $MkOsd{'UBat'} / 10);
        $MkOsd{'Current'}        = sprintf("%.1f", $MkOsd{'Current'} / 10);
		
        # Timestamp, wann der Datensatz geschtieben wurde
        $MkOsd{'_Timestamp'} = time;
        $MkOsd{'_FrameCount'} ++;
        }
		
    elsif ( $Id eq "s" )
        {
        #
        # NC Target position in %MkTarget
        #
        # Datenstruktur: (see NC waypoints.h)
        #    GPS_Pos_t Position;     // the gps position of the waypoint, see ubx.h for details
        #    s16 Heading;            // orientation, future implementation
        #    u8  ToleranceRadius;    // in meters, if the MK is within that range around the target, then the next target is 
        #    u8  HoldTime;           // in seconds, if the MK was once in the tolerance area around a WP,
        #                            // this time defines the delay before the next WP is triggered
        #    u8  Event_Flag;         // future emplementation
        #    u8  Index;              // to indentify different waypoints, workaround for bad communications PC <-> NC
        #    u8  Type;               // typeof Waypoint
        #    u8  WP_EventChannelValue;
        #    u8  AltitudeRate;       // rate to change the setpoint
        #    u8  Speed;              // rate to change the Position
        #    u8  CamAngle;           // Camera servo angle  (-1 = POI)
        #    u8  reserve[6];         // reserve
		
        lock (%MkTarget);  # until end of block
		
        (
        $MkTarget{'Pos_Lon'},
        $MkTarget{'Pos_Lat'},
        $MkTarget{'Pos_Alt'},
        $MkTarget{'Pos_Stat'},
        $MkTarget{'Heading'},
        $MkTarget{'ToleranceRadius'},
        $MkTarget{'HoldTime'},
        $MkTarget{'EventFlag'},

        $MkTarget{'Index'},
        $MkTarget{'Type'},
        $MkTarget{'WP_EventChannelValue'},
        $MkTarget{'AltitudeRate'},
        $MkTarget{'Speed'},
        $MkTarget{'CamAngle'},
        ) = unpack ('lllCsCCCCCCCCC', $Data); 

        $MkTarget{'Pos_Lon'} = sprintf("%.7f", $MkTarget{'Pos_Lon'} / 10000000);
        $MkTarget{'Pos_Lat'} = sprintf("%.7f", $MkTarget{'Pos_Lat'} / 10000000);
        $MkTarget{'Pos_Alt'} = sprintf("%.3f", $MkTarget{'Pos_Alt'} / 1000);
	
        # Timestamp, wann der Datensatz geschrieben wurde
        $MkTarget{'_Timestamp'} = time;
        $MkTarget{'_FrameCount'} ++;
        }
	
    elsif ( $Id eq "W" )
        {
        #
        # Request new waypoint
        #
        # Datenstruktur:
        #    u8 Number of waypoint
        
        ($WpNumber) = unpack ('C', $Data);
 
        # keine Ahnung wofuer das gut sein soll

        # print "Request new Waypoint Number: $WpNumber\n";
	    
        }
		
    elsif ( $Id eq "V" )
        {
        #
        # Version
        #
        # Datenstruktur:
        #    u8 SWMajor
        #    u8 SWMinor
        #    u8 ProtoMajor
        #    u8 ProtoMinor
        #    u8 SWPatch
        #    u8 HardwareError[5]

        (
        $Mk{'SWMajor'},
        $Mk{'SWMinor'},
        $Mk{'ProtoMajor'},
        $Mk{'ProtoMinor'},
        $Mk{'SWPatch'},
        $Mk{'HardwareError1'},
        $Mk{'HardwareError2'},
        $Mk{'HardwareError3'},
        $Mk{'HardwareError4'},
        $Mk{'HardwareError5'},
        ) = unpack ('C10', $Data); 

        $Mk{'_Timestamp'} = time;
        $Mk{'_FrameCount'} ++;
        }

    elsif ( $Id eq "E" )
        {
        #
        # Error Text
        #
        # Datenstruktur:
        #    s8 ErrorMsg[25]

        $Mk{'ErrorMsg'} = unpack ('Z25', $Data); 
        }
	
    elsif ( $Id eq "D" )
        {
        #
        # NC Debug %MkNcDebug
        #
        # Datenstruktur:
        #    u8 Digital[2];
        #    u16 Analog[32];

        lock (%MkNcDebug);    # until end of block

        (
        $MkNcDebug{'Digital_00'},
        $MkNcDebug{'Digital_01'},
        $MkNcDebug{'Analog_00'},
        $MkNcDebug{'Analog_01'},
        $MkNcDebug{'Analog_02'},
        $MkNcDebug{'Analog_03'},
        $MkNcDebug{'Analog_04'},
        $MkNcDebug{'Analog_05'},
        $MkNcDebug{'Analog_06'},
        $MkNcDebug{'Analog_07'},
        $MkNcDebug{'Analog_08'},
        $MkNcDebug{'Analog_09'},
        $MkNcDebug{'Analog_10'},
        $MkNcDebug{'Analog_11'},
        $MkNcDebug{'Analog_12'},
        $MkNcDebug{'Analog_13'},
        $MkNcDebug{'Analog_14'},
        $MkNcDebug{'Analog_15'},
        $MkNcDebug{'Analog_16'},
        $MkNcDebug{'Analog_17'},
        $MkNcDebug{'Analog_18'},
        $MkNcDebug{'Analog_19'},
        $MkNcDebug{'Analog_20'},
        $MkNcDebug{'Analog_21'},		
        $MkNcDebug{'Analog_22'},
        $MkNcDebug{'Analog_23'},
        $MkNcDebug{'Analog_24'},
        $MkNcDebug{'Analog_25'},
        $MkNcDebug{'Analog_26'},
        $MkNcDebug{'Analog_27'},
        $MkNcDebug{'Analog_28'},
        $MkNcDebug{'Analog_29'},
        $MkNcDebug{'Analog_30'},
        $MkNcDebug{'Analog_31'},
        ) = unpack ('C2s32', $Data); 

        # Timestamp, wann der Datensatz geschrieben wurde
        $MkNcDebug{'_Timestamp'} = time;
        $MkNcDebug{'_FrameCount'} ++;
        }

    elsif ( $Id eq "B" )
        {
        #
        # External Control
        #
        # Datenstruktur:
        #    u8 ConfirmFrame;

        my ($ConfirmFrame) = unpack ('C5', $Data); 

        }

    elsif ( $Id eq "L" )
        {
        #
        # LCD Screen
        #
        # Datenstruktur:
        #    u8 Menuitem
        #    u8 MaxMenuItem
        #    char[80] Display Text

        my ($MenuItem, $MaxMenuItem, $LcdLine) = unpack ('CCA80', $Data);
        if ( $LcdLine =~ /Po1:\s*(\d+)\s*Po2:\s*(\d+)\s*Po3:\s*(\d+)\s*Po4:\s*(\d+)/i )
            {
            $Stick{'RcPoti1'} = $1;
            $Stick{'RcPoti2'} = $2;
            $Stick{'RcPoti3'} = $3;
            $Stick{'RcPoti4'} = $4;
            }
        elsif ( $LcdLine =~ /Po5:\s*(\d+)\s*Po6:\s*(\d+)\s*Po7:\s*(\d+)\s*Po8:\s*(\d+)/i )
            {
            $Stick{'RcPoti5'} = $1;
            $Stick{'RcPoti6'} = $2;
            $Stick{'RcPoti7'} = $3;
            $Stick{'RcPoti8'} = $4;
            }
        elsif ( $LcdLine =~ /Ni:\s*(-*\d+)\s*Ro:\s*(-*\d+)\s*Gs:\s*(-*\d+)\s*Ya:\s*(-*\d+)/i )
            {
            $Stick{'RcStickNick'} = $1;
            $Stick{'RcStickRoll'} = $2;
            $Stick{'RcStickGas'} = $3;
            $Stick{'RcStickGier'} = $4;
            }
        elsif ( $LcdLine =~ /Sat:\s*(-*\d+)\s*(-*\S+)\[1\]/i )
            {
            $Mk{'GpsFix'} = $2;
            }

        $Stick{'_RcTimestamp'} = time;
        $Stick{'_RcFrameCount'} ++;
        }

    elsif ( $Id eq "K" )
        {
        #
        # Motor parameter
        #
        # Datenstruktur: (see NC spi_slave.h.h)
        #     u8 Current;
        #     u8 Temperature;
        #     u8 MaxPWM;
        #     u8 State;

        lock (%MkMotor);  # until end of block
		
        (
        $Num,
        $Current,
        $Temperature,
        $MaxPWM,
        $State,
        ) = unpack ('C5', $Data); 

        $Num = sprintf ("%02d", $Num + 1);
	
        $MkMotor{"Current_$Num"}     = $Current;
        $MkMotor{"Temperature_$Num"} = $Temperature;
        $MkMotor{"MaxPWM_$Num"}      = $MaxPWM;
        $MkMotor{"State_$Num"}       = $State;

        # Timestamp, wann der Datensatz geschrieben wurde
        $MkMotor{'_Timestamp'} = time;
        $MkMotor{'_FrameCount'} ++;
        }

    elsif ( $Id eq "Q" )
        #
        # FC Setting - see eeprom.h
        #
        {
        my $Revision = unpack ('C', substr $Data, 1, 1);
        print "mkcomm.pl: Settings structure has revision $Revision\n";

        if ( $Revision >= 97 )
            {
            
            (
	    $MkSetting{'SettingNumber'},          # setting number
            $MkSetting{'Revision'},               # version of setting structure
            $MkSetting{'Kanalbelegung_01'},
	    $MkSetting{'Kanalbelegung_02'},
	    $MkSetting{'Kanalbelegung_03'},
	    $MkSetting{'Kanalbelegung_04'},
	    $MkSetting{'Kanalbelegung_05'},
	    $MkSetting{'Kanalbelegung_06'},
	    $MkSetting{'Kanalbelegung_07'},
	    $MkSetting{'Kanalbelegung_08'},
	    $MkSetting{'Kanalbelegung_09'},
	    $MkSetting{'Kanalbelegung_10'},
	    $MkSetting{'Kanalbelegung_11'},
	    $MkSetting{'Kanalbelegung_12'},
	    $MkSetting{'GlobalConfig'},            # CFG_HOEHENREGELUNG       0x01
                                                   # CFG_HOEHEN_SCHALTER      0x02
                                                   # CFG_HEADING_HOLD         0x04
                                                   # CFG_KOMPASS_AKTIV        0x08
                                                   # CFG_KOMPASS_FIX          0x10
                                                   # CFG_GPS_AKTIV            0x20
                                                   # CFG_ACHSENKOPPLUNG_AKTIV 0x40
                                                   # CFG_DREHRATEN_BEGRENZER  0x80
	    $MkSetting{'Hoehe_MinGas'},            # Wert : 0-100
	    $MkSetting{'Luftdruck_D'},             # Wert : 0-250
	    $MkSetting{'HoeheCHannel'},            # Wert : 0-32
	    $MkSetting{'Hoehe_P'},                 # Wert : 0-32
	    $MkSetting{'Hoehe_Verstaerkung'},      # Wert : 0-50
	    $MkSetting{'Hoehe_ACC_Wirkung'},       # Wert : 0-250
	    $MkSetting{'Hoehe_HoverBand'},         # Wert : 0-250
	    $MkSetting{'Hoehe_GPS_Z'},             # Wert : 0-250
	    $MkSetting{'Hoehe_StickNeutralPoint'}, # Wert : 0-250
	    $MkSetting{'Stick_P'},                 # Wert : 1-6
	    $MkSetting{'Stick_D'},                 # Wert : 0-64
	    $MkSetting{'Gier_P'},                  # Wert : 1-20
	    $MkSetting{'Gas_Min'},                 # Wert : 0-32
	    $MkSetting{'Gas_Max'},                 # Wert : 33-250
	    $MkSetting{'GyroAccFaktor'},           # Wert : 1-64
	    $MkSetting{'KompassWirkung'},          # Wert : 0-32
	    $MkSetting{'Gyro_P'},                  # Wert : 10-250
	    $MkSetting{'Gyro_I'},                  # Wert : 0-250
	    $MkSetting{'Gyro_D'},                  # Wert : 0-250
	    $MkSetting{'Gyro_Gier_P'},             # Wert : 10-250
	    $MkSetting{'Gyro_Gier_I'},             # Wert : 0-250
	    $MkSetting{'Gyro_Stability'},          # Wert : 0-16
	    $MkSetting{'UnterspannungsWarnung'},   # Wert : 0-250
	    $MkSetting{'NotGas'},                  # Wert : 0-250     Gaswert bei Empngsverlust
	    $MkSetting{'NotGasZeit'},              # Wert : 0-250     Zeit bis auf NotGas geschaltet wird, wg. Rx-Problemen
	    $MkSetting{'Receiver'},                # RECEIVER_PPM              0
                                                   # RECEIVER_SPEKTRUM         1
                                                   # RECEIVER_SPEKTRUM_HI_RES  2
                                                   # RECEIVER_SPEKTRUM_LOW_RES 3
                                                   # RECEIVER_JETI             4
                                                   # RECEIVER_ACT_DSL          5
                                                   # RECEIVER_UNKNOWN          0xFF
	    $MkSetting{'I_Faktor'},                # Wert : 0-250
	    $MkSetting{'UserParam1'},              # Wert : 0-250
	    $MkSetting{'UserParam2'},              # Wert : 0-250
	    $MkSetting{'UserParam3'},              # Wert : 0-250
	    $MkSetting{'UserParam4'},              # Wert : 0-250
	    $MkSetting{'ServoNickControl'},        # Wert : 0-250     # Stellung des Servos
	    $MkSetting{'ServoNickComp'},           # Wert : 0-250     # Einfluss Gyro/Servo
	    $MkSetting{'ServoNickMin'},            # Wert : 0-250     # Anschlag
	    $MkSetting{'ServoNickMax'},            # Wert : 0-250     # Anschlag
	    $MkSetting{'ServoRollControl'},        # Wert : 0-250     # Stellung des Servos
	    $MkSetting{'ServoRollComp'},           # Wert : 0-250
	    $MkSetting{'ServoRollMin'},            # Wert : 0-250
	    $MkSetting{'ServoRollMax'},            # Wert : 0-250
	    $MkSetting{'ServoNickRefresh'},        # Speed of the Servo
            $MkSetting{'ServoManualControlSpeed'},
            $MkSetting{'CamOrientation'},
	    $MkSetting{'Servo3'},                  # Value or mapping of the Servo Output
            $MkSetting{'Servo4'},                  # Value or mapping of the Servo Output
	    $MkSetting{'Servo5'},                  # Value or mapping of the Servo Output
	    $MkSetting{'LoopGasLimit'},            # Wert: 0-250  max. Gas whrend Looping
	    $MkSetting{'LoopThreshold'},           # Wert: 0-250  Schwelle fr Stickausschlag
	    $MkSetting{'LoopHysterese'},           # Wert: 0-250  Hysterese fr Stickausschlag
	    $MkSetting{'AchsKopplung1'},           # Wert: 0-250  Faktor, mit dem Gier die Achsen Roll und Nick koppelt (NickRollMitkopplung)
	    $MkSetting{'AchsKopplung2'},           # Wert: 0-250  Faktor, mit dem Nick und Roll verkoppelt werden
	    $MkSetting{'CouplingYawCorrection'},   # Wert: 0-250  Faktor, mit dem Nick und Roll verkoppelt werden
	    $MkSetting{'WinkelUmschlagNick'},      # Wert: 0-250  180-Punkt
	    $MkSetting{'WinkelUmschlagRoll'},      # Wert: 0-250  180-Punkt
	    $MkSetting{'GyroAccAbgleich'},         # 1/k  (Koppel_ACC_Wirkung)
	    $MkSetting{'Driftkomp'},
	    $MkSetting{'DynamicStability'},
	    $MkSetting{'UserParam5'},              # Wert : 0-250
	    $MkSetting{'UserParam6'},              # Wert : 0-250
	    $MkSetting{'UserParam7'},              # Wert : 0-250
	    $MkSetting{'UserParam8'},              # Wert : 0-250
	    $MkSetting{'J16Bitmask'},              # for the J16 Output
	    $MkSetting{'J16Timing'},               # for the J16 Output
	    $MkSetting{'J17Bitmask'},              # for the J17 Output
	    $MkSetting{'J17Timing'},               # for the J17 Output
	    $MkSetting{'WARN_J16_Bitmask'},        # for the J16 Output
	    $MkSetting{'WARN_J17_Bitmask'},        # for the J17 Output
	    $MkSetting{'NaviOut1Parameter'},       # for the J16 Output
	    $MkSetting{'NaviGpsModeControl'},      # Parameters for the Naviboard
	    $MkSetting{'NaviGpsGain'},
	    $MkSetting{'NaviGpsP'},
	    $MkSetting{'NaviGpsI'},
	    $MkSetting{'NaviGpsD'},
	    $MkSetting{'NaviGpsPLimit'},
	    $MkSetting{'NaviGpsILimit'},
	    $MkSetting{'NaviGpsDLimit'},
	    $MkSetting{'NaviGpsA'},
	    $MkSetting{'NaviGpsMinSat'},
	    $MkSetting{'NaviStickThreshold'},
	    $MkSetting{'NaviWindCorrection'},
	    $MkSetting{'NaviAccCompensation'},
	    $MkSetting{'NaviOperatingRadius'},
	    $MkSetting{'NaviAngleLimitation'},
	    $MkSetting{'NaviPH_LoginTime'},
	    $MkSetting{'ExternalControl'},         # for serial Control
	    $MkSetting{'OrientationAngle'},        # Where is the front-direction?
	    $MkSetting{'CareFreeChannel'},         # switch for CareFree
            $MkSetting{'MotorSafetySwitch'},
            $MkSetting{'MotorSmooth'},
            $MkSetting{'ComingHomeALtitude'},
            $MkSetting{'FailSafeTime'},
            $MkSetting{'MaxAltitude'},
            $MkSetting{'FailsafeChannel'},
            $MkSetting{'ServoFilterNick'},
            $MkSetting{'ServoFilterRoll'},
            $MkSetting{'StartLandChannel'},
            $MkSetting{'LandingSpeed'},
	    $MkSetting{'CompassOffset'},
	    $MkSetting{'AutoLandingVoltage'},      # in 0,1V  0 -> disabled
	    $MkSetting{'BitConfig'},               # CFG_LOOP_OBEN            0x01
                                                   # CFG_LOOP_UNTEN           0x02
                                                   # CFG_LOOP_LINKS           0x04
                                                   # CFG_LOOP_RECHTS          0x08
                                                   # CFG_MOTOR_BLINK	      0x10
                                                   # CFG_MOTOR_OFF_LED1       0x20
                                                   # CFG_MOTOR_OFF_LED2       0x40
                                                   # CFG_RES4		      0x80
	    $MkSetting{'ServoCompInvert'},         # 0x01 = Nick, 0x02 = Roll
	    $MkSetting{'ExtraConfig'},             # CFG2_HEIGHT_LIMIT        0x01
                                                   # CFG2_VARIO_BEEP          0x02
                                                   # CFG_SENSITIVE_RC         0x04
                                                   # CFG_3_3V_REFERENCE       0x08
            $MkSetting{'GlobalConfig3'},
	    $MkSetting{'Name'},
            $MkSetting{'crc'},
            ) = unpack ('C117A12C1', $Data); 
            }

        elsif ( $Revision >= 95 )
            {
            
            (
	    $MkSetting{'SettingNumber'},          # setting number
            $MkSetting{'Revision'},               # version of setting structure
            $MkSetting{'Kanalbelegung_01'},
	    $MkSetting{'Kanalbelegung_02'},
	    $MkSetting{'Kanalbelegung_03'},
	    $MkSetting{'Kanalbelegung_04'},
	    $MkSetting{'Kanalbelegung_05'},
	    $MkSetting{'Kanalbelegung_06'},
	    $MkSetting{'Kanalbelegung_07'},
	    $MkSetting{'Kanalbelegung_08'},
	    $MkSetting{'Kanalbelegung_09'},
	    $MkSetting{'Kanalbelegung_10'},
	    $MkSetting{'Kanalbelegung_11'},
	    $MkSetting{'Kanalbelegung_12'},
	    $MkSetting{'GlobalConfig'},            # CFG_HOEHENREGELUNG       0x01
                                                   # CFG_HOEHEN_SCHALTER      0x02
                                                   # CFG_HEADING_HOLD         0x04
                                                   # CFG_KOMPASS_AKTIV        0x08
                                                   # CFG_KOMPASS_FIX          0x10
                                                   # CFG_GPS_AKTIV            0x20
                                                   # CFG_ACHSENKOPPLUNG_AKTIV 0x40
                                                   # CFG_DREHRATEN_BEGRENZER  0x80
	    $MkSetting{'Hoehe_MinGas'},            # Wert : 0-100
	    $MkSetting{'Luftdruck_D'},             # Wert : 0-250
	    $MkSetting{'HoeheCHannel'},            # Wert : 0-32
	    $MkSetting{'Hoehe_P'},                 # Wert : 0-32
	    $MkSetting{'Hoehe_Verstaerkung'},      # Wert : 0-50
	    $MkSetting{'Hoehe_ACC_Wirkung'},       # Wert : 0-250
	    $MkSetting{'Hoehe_HoverBand'},         # Wert : 0-250
	    $MkSetting{'Hoehe_GPS_Z'},             # Wert : 0-250
	    $MkSetting{'Hoehe_StickNeutralPoint'}, # Wert : 0-250
	    $MkSetting{'Stick_P'},                 # Wert : 1-6
	    $MkSetting{'Stick_D'},                 # Wert : 0-64
	    $MkSetting{'Gier_P'},                  # Wert : 1-20
	    $MkSetting{'Gas_Min'},                 # Wert : 0-32
	    $MkSetting{'Gas_Max'},                 # Wert : 33-250
	    $MkSetting{'GyroAccFaktor'},           # Wert : 1-64
	    $MkSetting{'KompassWirkung'},          # Wert : 0-32
	    $MkSetting{'Gyro_P'},                  # Wert : 10-250
	    $MkSetting{'Gyro_I'},                  # Wert : 0-250
	    $MkSetting{'Gyro_D'},                  # Wert : 0-250
	    $MkSetting{'Gyro_Gier_P'},             # Wert : 10-250
	    $MkSetting{'Gyro_Gier_I'},             # Wert : 0-250
	    $MkSetting{'Gyro_Stability'},          # Wert : 0-16
	    $MkSetting{'UnterspannungsWarnung'},   # Wert : 0-250
	    $MkSetting{'NotGas'},                  # Wert : 0-250     Gaswert bei Empngsverlust
	    $MkSetting{'NotGasZeit'},              # Wert : 0-250     Zeit bis auf NotGas geschaltet wird, wg. Rx-Problemen
	    $MkSetting{'Receiver'},                # RECEIVER_PPM              0
                                                   # RECEIVER_SPEKTRUM         1
                                                   # RECEIVER_SPEKTRUM_HI_RES  2
                                                   # RECEIVER_SPEKTRUM_LOW_RES 3
                                                   # RECEIVER_JETI             4
                                                   # RECEIVER_ACT_DSL          5
                                                   # RECEIVER_UNKNOWN          0xFF
	    $MkSetting{'I_Faktor'},                # Wert : 0-250
	    $MkSetting{'UserParam1'},              # Wert : 0-250
	    $MkSetting{'UserParam2'},              # Wert : 0-250
	    $MkSetting{'UserParam3'},              # Wert : 0-250
	    $MkSetting{'UserParam4'},              # Wert : 0-250
	    $MkSetting{'ServoNickControl'},        # Wert : 0-250     # Stellung des Servos
	    $MkSetting{'ServoNickComp'},           # Wert : 0-250     # Einfluss Gyro/Servo
	    $MkSetting{'ServoNickMin'},            # Wert : 0-250     # Anschlag
	    $MkSetting{'ServoNickMax'},            # Wert : 0-250     # Anschlag
	    $MkSetting{'ServoRollControl'},        # Wert : 0-250     # Stellung des Servos
	    $MkSetting{'ServoRollComp'},           # Wert : 0-250
	    $MkSetting{'ServoRollMin'},            # Wert : 0-250
	    $MkSetting{'ServoRollMax'},            # Wert : 0-250
	    $MkSetting{'ServoNickRefresh'},        # Speed of the Servo
            $MkSetting{'ServoManualControlSpeed'},
            $MkSetting{'CamOrientation'},
	    $MkSetting{'Servo3'},                  # Value or mapping of the Servo Output
            $MkSetting{'Servo4'},                  # Value or mapping of the Servo Output
	    $MkSetting{'Servo5'},                  # Value or mapping of the Servo Output
	    $MkSetting{'LoopGasLimit'},            # Wert: 0-250  max. Gas whrend Looping
	    $MkSetting{'LoopThreshold'},           # Wert: 0-250  Schwelle fr Stickausschlag
	    $MkSetting{'LoopHysterese'},           # Wert: 0-250  Hysterese fr Stickausschlag
	    $MkSetting{'AchsKopplung1'},           # Wert: 0-250  Faktor, mit dem Gier die Achsen Roll und Nick koppelt (NickRollMitkopplung)
	    $MkSetting{'AchsKopplung2'},           # Wert: 0-250  Faktor, mit dem Nick und Roll verkoppelt werden
	    $MkSetting{'CouplingYawCorrection'},   # Wert: 0-250  Faktor, mit dem Nick und Roll verkoppelt werden
	    $MkSetting{'WinkelUmschlagNick'},      # Wert: 0-250  180-Punkt
	    $MkSetting{'WinkelUmschlagRoll'},      # Wert: 0-250  180-Punkt
	    $MkSetting{'GyroAccAbgleich'},         # 1/k  (Koppel_ACC_Wirkung)
	    $MkSetting{'Driftkomp'},
	    $MkSetting{'DynamicStability'},
	    $MkSetting{'UserParam5'},              # Wert : 0-250
	    $MkSetting{'UserParam6'},              # Wert : 0-250
	    $MkSetting{'UserParam7'},              # Wert : 0-250
	    $MkSetting{'UserParam8'},              # Wert : 0-250
	    $MkSetting{'J16Bitmask'},              # for the J16 Output
	    $MkSetting{'J16Timing'},               # for the J16 Output
	    $MkSetting{'J17Bitmask'},              # for the J17 Output
	    $MkSetting{'J17Timing'},               # for the J17 Output
	    $MkSetting{'WARN_J16_Bitmask'},        # for the J16 Output
	    $MkSetting{'WARN_J17_Bitmask'},        # for the J17 Output
	    $MkSetting{'NaviOut1Parameter'},       # for the J16 Output
	    $MkSetting{'NaviGpsModeControl'},      # Parameters for the Naviboard
	    $MkSetting{'NaviGpsGain'},
	    $MkSetting{'NaviGpsP'},
	    $MkSetting{'NaviGpsI'},
	    $MkSetting{'NaviGpsD'},
	    $MkSetting{'NaviGpsPLimit'},
	    $MkSetting{'NaviGpsILimit'},
	    $MkSetting{'NaviGpsDLimit'},
	    $MkSetting{'NaviGpsA'},
	    $MkSetting{'NaviGpsMinSat'},
	    $MkSetting{'NaviStickThreshold'},
	    $MkSetting{'NaviWindCorrection'},
	    $MkSetting{'NaviAccCompensation'},
	    $MkSetting{'NaviOperatingRadius'},
	    $MkSetting{'NaviAngleLimitation'},
	    $MkSetting{'NaviPH_LoginTime'},
	    $MkSetting{'ExternalControl'},         # for serial Control
	    $MkSetting{'OrientationAngle'},        # Where is the front-direction?
	    $MkSetting{'CareFreeChannel'},         # switch for CareFree
            $MkSetting{'MotorSafetySwitch'},
            $MkSetting{'MotorSmooth'},
            $MkSetting{'ComingHomeALtitude'},
            $MkSetting{'FailSafeTime'},
            $MkSetting{'MaxAltitude'},
            $MkSetting{'FailsafeChannel'},
            $MkSetting{'ServoFilterNick'},
            $MkSetting{'ServoFilterRoll'},
            $MkSetting{'StartLandChannel'},
            $MkSetting{'LandingSpeed'},
	    $MkSetting{'BitConfig'},               # CFG_LOOP_OBEN            0x01
                                                   # CFG_LOOP_UNTEN           0x02
                                                   # CFG_LOOP_LINKS           0x04
                                                   # CFG_LOOP_RECHTS          0x08
                                                   # CFG_MOTOR_BLINK	      0x10
                                                   # CFG_MOTOR_OFF_LED1       0x20
                                                   # CFG_MOTOR_OFF_LED2       0x40
                                                   # CFG_RES4		      0x80
	    $MkSetting{'ServoCompInvert'},         # 0x01 = Nick, 0x02 = Roll
	    $MkSetting{'ExtraConfig'},             # CFG2_HEIGHT_LIMIT        0x01
                                                   # CFG2_VARIO_BEEP          0x02
                                                   # CFG_SENSITIVE_RC         0x04
                                                   # CFG_3_3V_REFERENCE       0x08
            $MkSetting{'GlobalConfig3'},
	    $MkSetting{'Name'},
            $MkSetting{'crc'},
            ) = unpack ('C115A12C1', $Data); 
            }

        elsif ( $Revision >= 90 )   # FC 0.88 NC 0.28
            {
            
            (
	    $MkSetting{'SettingNumber'},          # setting number
            $MkSetting{'Revision'},               # version of setting structure
            $MkSetting{'Kanalbelegung_01'},
	    $MkSetting{'Kanalbelegung_02'},
	    $MkSetting{'Kanalbelegung_03'},
	    $MkSetting{'Kanalbelegung_04'},
	    $MkSetting{'Kanalbelegung_05'},
	    $MkSetting{'Kanalbelegung_06'},
	    $MkSetting{'Kanalbelegung_07'},
	    $MkSetting{'Kanalbelegung_08'},
	    $MkSetting{'Kanalbelegung_09'},
	    $MkSetting{'Kanalbelegung_10'},
	    $MkSetting{'Kanalbelegung_11'},
	    $MkSetting{'Kanalbelegung_12'},
	    $MkSetting{'GlobalConfig'},            # CFG_HOEHENREGELUNG       0x01
                                                   # CFG_HOEHEN_SCHALTER      0x02
                                                   # CFG_HEADING_HOLD         0x04
                                                   # CFG_KOMPASS_AKTIV        0x08
                                                   # CFG_KOMPASS_FIX          0x10
                                                   # CFG_GPS_AKTIV            0x20
                                                   # CFG_ACHSENKOPPLUNG_AKTIV 0x40
                                                   # CFG_DREHRATEN_BEGRENZER  0x80
	    $MkSetting{'Hoehe_MinGas'},            # Wert : 0-100
	    $MkSetting{'Luftdruck_D'},             # Wert : 0-250
	    $MkSetting{'MaxHoehe'},                # Wert : 0-32
	    $MkSetting{'Hoehe_P'},                 # Wert : 0-32
	    $MkSetting{'Hoehe_Verstaerkung'},      # Wert : 0-50
	    $MkSetting{'Hoehe_ACC_Wirkung'},       # Wert : 0-250
	    $MkSetting{'Hoehe_HoverBand'},         # Wert : 0-250
	    $MkSetting{'Hoehe_GPS_Z'},             # Wert : 0-250
	    $MkSetting{'Hoehe_StickNeutralPoint'}, # Wert : 0-250
	    $MkSetting{'Stick_P'},                 # Wert : 1-6
	    $MkSetting{'Stick_D'},                 # Wert : 0-64
	    $MkSetting{'StickGier_P'},             # Wert : 1-20
	    $MkSetting{'Gas_Min'},                 # Wert : 0-32
	    $MkSetting{'Gas_Max'},                 # Wert : 33-250
	    $MkSetting{'GyroAccFaktor'},           # Wert : 1-64
	    $MkSetting{'KompassWirkung'},          # Wert : 0-32
	    $MkSetting{'Gyro_P'},                  # Wert : 10-250
	    $MkSetting{'Gyro_I'},                  # Wert : 0-250
	    $MkSetting{'Gyro_D'},                  # Wert : 0-250
	    $MkSetting{'Gyro_Gier_P'},             # Wert : 10-250
	    $MkSetting{'Gyro_Gier_I'},             # Wert : 0-250
	    $MkSetting{'Gyro_Stability'},          # Wert : 0-16
	    $MkSetting{'UnterspannungsWarnung'},   # Wert : 0-250
	    $MkSetting{'NotGas'},                  # Wert : 0-250     Gaswert bei Empngsverlust
	    $MkSetting{'NotGasZeit'},              # Wert : 0-250     Zeit bis auf NotGas geschaltet wird, wg. Rx-Problemen
	    $MkSetting{'Receiver'},                # RECEIVER_PPM              0
                                                   # RECEIVER_SPEKTRUM         1
                                                   # RECEIVER_SPEKTRUM_HI_RES  2
                                                   # RECEIVER_SPEKTRUM_LOW_RES 3
                                                   # RECEIVER_JETI             4
                                                   # RECEIVER_ACT_DSL          5
                                                   # RECEIVER_UNKNOWN          0xFF
	    $MkSetting{'I_Faktor'},                # Wert : 0-250
	    $MkSetting{'UserParam1'},              # Wert : 0-250
	    $MkSetting{'UserParam2'},              # Wert : 0-250
	    $MkSetting{'UserParam3'},              # Wert : 0-250
	    $MkSetting{'UserParam4'},              # Wert : 0-250
	    $MkSetting{'ServoNickControl'},        # Wert : 0-250     # Stellung des Servos
	    $MkSetting{'ServoNickComp'},           # Wert : 0-250     # Einfluss Gyro/Servo
	    $MkSetting{'ServoNickMin'},            # Wert : 0-250     # Anschlag
	    $MkSetting{'ServoNickMax'},            # Wert : 0-250     # Anschlag
	    $MkSetting{'ServoRollControl'},        # Wert : 0-250     # Stellung des Servos
	    $MkSetting{'ServoRollComp'},           # Wert : 0-250
	    $MkSetting{'ServoRollMin'},            # Wert : 0-250
	    $MkSetting{'ServoRollMax'},            # Wert : 0-250
	    $MkSetting{'ServoNickRefresh'},        # Speed of the Servo
            $MkSetting{'ServoManualControlSpeed'},
            $MkSetting{'CamOrientation'},
	    $MkSetting{'Servo3'},                  # Value or mapping of the Servo Output
            $MkSetting{'Servo4'},                  # Value or mapping of the Servo Output
	    $MkSetting{'Servo5'},                  # Value or mapping of the Servo Output
	    $MkSetting{'LoopGasLimit'},            # Wert: 0-250  max. Gas whrend Looping
	    $MkSetting{'LoopThreshold'},           # Wert: 0-250  Schwelle fr Stickausschlag
	    $MkSetting{'LoopHysterese'},           # Wert: 0-250  Hysterese fr Stickausschlag
	    $MkSetting{'AchsKopplung1'},           # Wert: 0-250  Faktor, mit dem Gier die Achsen Roll und Nick koppelt (NickRollMitkopplung)
	    $MkSetting{'AchsKopplung2'},           # Wert: 0-250  Faktor, mit dem Nick und Roll verkoppelt werden
	    $MkSetting{'CouplingYawCorrection'},   # Wert: 0-250  Faktor, mit dem Nick und Roll verkoppelt werden
	    $MkSetting{'WinkelUmschlagNick'},      # Wert: 0-250  180-Punkt
	    $MkSetting{'WinkelUmschlagRoll'},      # Wert: 0-250  180-Punkt
	    $MkSetting{'GyroAccAbgleich'},         # 1/k  (Koppel_ACC_Wirkung)
	    $MkSetting{'Driftkomp'},
	    $MkSetting{'DynamicStability'},
	    $MkSetting{'UserParam5'},              # Wert : 0-250
	    $MkSetting{'UserParam6'},              # Wert : 0-250
	    $MkSetting{'UserParam7'},              # Wert : 0-250
	    $MkSetting{'UserParam8'},              # Wert : 0-250
	    $MkSetting{'J16Bitmask'},              # for the J16 Output
	    $MkSetting{'J16Timing'},               # for the J16 Output
	    $MkSetting{'J17Bitmask'},              # for the J17 Output
	    $MkSetting{'J17Timing'},               # for the J17 Output
	    $MkSetting{'WARN_J16_Bitmask'},        # for the J16 Output
	    $MkSetting{'WARN_J17_Bitmask'},        # for the J17 Output
	    $MkSetting{'NaviGpsModeControl'},      # Parameters for the Naviboard
	    $MkSetting{'NaviGpsGain'},
	    $MkSetting{'NaviGpsP'},
	    $MkSetting{'NaviGpsI'},
	    $MkSetting{'NaviGpsD'},
	    $MkSetting{'NaviGpsPLimit'},
	    $MkSetting{'NaviGpsILimit'},
	    $MkSetting{'NaviGpsDLimit'},
	    $MkSetting{'NaviGpsACC'},
	    $MkSetting{'NaviGpsMinSat'},
	    $MkSetting{'NaviStickThreshold'},
	    $MkSetting{'NaviWindCorrection'},
	    $MkSetting{'NaviAccCompensation'},
	    $MkSetting{'NaviOperatingRadius'},
	    $MkSetting{'NaviAngleLimitation'},
	    $MkSetting{'NaviPH_LoginTime'},
	    $MkSetting{'ExternalControl'},         # for serial Control
	    $MkSetting{'OrientationAngle'},        # Where is the front-direction?
	    $MkSetting{'CareFreeModeControl'},     # switch for CareFree
            $MkSetting{'MotorSafetySwitch'},
            $MkSetting{'MotorSmooth'},
            $MkSetting{'ComingHomeAltitude'},
            $MkSetting{'FailsaveTime'},
            $MkSetting{'MaxAltitude'},
	    $MkSetting{'FailsafeChannel'},         #if the value of this channel is > 100, the MK reports "RC-Lost"
            $MkSetting{'ServoFilterNick'},
            $MkSetting{'ServoFilterRoll'},
	    $MkSetting{'BitConfig'},               # CFG_LOOP_OBEN            0x01
                                                   # CFG_LOOP_UNTEN           0x02
                                                   # CFG_LOOP_LINKS           0x04
                                                   # CFG_LOOP_RECHTS          0x08
                                                   # CFG_MOTOR_BLINK	      0x10
                                                   # CFG_MOTOR_OFF_LED1       0x20
                                                   # CFG_MOTOR_OFF_LED2       0x40
                                                   # CFG_RES4		      0x80
	    $MkSetting{'ServoCompInvert'},         # 0x01 = Nick, 0x02 = Roll
	    $MkSetting{'ExtraConfig'},             # CFG2_HEIGHT_LIMIT        0x01
                                                   # CFG2_VARIO_BEEP          0x02
                                                   # CFG_SENSITIVE_RC         0x04
                                                   # CFG_3_3V_REFERENCE       0x08
                                                   # CFG_NO_RCOFF_BEEPING     0x10
                                                   # CFG_GPS_AID              0x20
                                                   # CFG_LEARNABLE_CAREFREE   0x40
                                                   # CFG_IGNORE_MAG_ERR_AT_STARTUP 0x80
            $MkSetting{'GlobalConfig3'},           # CFG3_NO_SDCARD_NO_START  0x01
                                                   # CFG3_DPH_MAX_RADIUS      0x02
                                                   # CFG3_VARIO_FAILSAFE      0x04
	    $MkSetting{'Name'},
            ) = unpack ('C112A12', $Data); 
            }

        elsif ( $Revision >= 88 )
            {
            
            (
	    $MkSetting{'SettingNumber'},          # setting number
            $MkSetting{'Revision'},               # version of setting structure
            $MkSetting{'Kanalbelegung_01'},
	    $MkSetting{'Kanalbelegung_02'},
	    $MkSetting{'Kanalbelegung_03'},
	    $MkSetting{'Kanalbelegung_04'},
	    $MkSetting{'Kanalbelegung_05'},
	    $MkSetting{'Kanalbelegung_06'},
	    $MkSetting{'Kanalbelegung_07'},
	    $MkSetting{'Kanalbelegung_08'},
	    $MkSetting{'Kanalbelegung_09'},
	    $MkSetting{'Kanalbelegung_10'},
	    $MkSetting{'Kanalbelegung_11'},
	    $MkSetting{'Kanalbelegung_12'},
	    $MkSetting{'GlobalConfig'},            # CFG_HOEHENREGELUNG       0x01
                                                   # CFG_HOEHEN_SCHALTER      0x02
                                                   # CFG_HEADING_HOLD         0x04
                                                   # CFG_KOMPASS_AKTIV        0x08
                                                   # CFG_KOMPASS_FIX          0x10
                                                   # CFG_GPS_AKTIV            0x20
                                                   # CFG_ACHSENKOPPLUNG_AKTIV 0x40
                                                   # CFG_DREHRATEN_BEGRENZER  0x80
	    $MkSetting{'Hoehe_MinGas'},            # Wert : 0-100
	    $MkSetting{'Luftdruck_D'},             # Wert : 0-250
	    $MkSetting{'MaxHoehe'},                # Wert : 0-32
	    $MkSetting{'Hoehe_P'},                 # Wert : 0-32
	    $MkSetting{'Hoehe_Verstaerkung'},      # Wert : 0-50
	    $MkSetting{'Hoehe_ACC_Wirkung'},       # Wert : 0-250
	    $MkSetting{'Hoehe_HoverBand'},         # Wert : 0-250
	    $MkSetting{'Hoehe_GPS_Z'},             # Wert : 0-250
	    $MkSetting{'Hoehe_StickNeutralPoint'}, # Wert : 0-250
	    $MkSetting{'Stick_P'},                 # Wert : 1-6
	    $MkSetting{'Stick_D'},                 # Wert : 0-64
	    $MkSetting{'Gier_P'},                  # Wert : 1-20
	    $MkSetting{'Gas_Min'},                 # Wert : 0-32
	    $MkSetting{'Gas_Max'},                 # Wert : 33-250
	    $MkSetting{'GyroAccFaktor'},           # Wert : 1-64
	    $MkSetting{'KompassWirkung'},          # Wert : 0-32
	    $MkSetting{'Gyro_P'},                  # Wert : 10-250
	    $MkSetting{'Gyro_I'},                  # Wert : 0-250
	    $MkSetting{'Gyro_D'},                  # Wert : 0-250
	    $MkSetting{'Gyro_Gier_P'},             # Wert : 10-250
	    $MkSetting{'Gyro_Gier_I'},             # Wert : 0-250
	    $MkSetting{'Gyro_Stability'},          # Wert : 0-16
	    $MkSetting{'UnterspannungsWarnung'},   # Wert : 0-250
	    $MkSetting{'NotGas'},                  # Wert : 0-250     Gaswert bei Empngsverlust
	    $MkSetting{'NotGasZeit'},              # Wert : 0-250     Zeit bis auf NotGas geschaltet wird, wg. Rx-Problemen
	    $MkSetting{'Receiver'},                # RECEIVER_PPM              0
                                                   # RECEIVER_SPEKTRUM         1
                                                   # RECEIVER_SPEKTRUM_HI_RES  2
                                                   # RECEIVER_SPEKTRUM_LOW_RES 3
                                                   # RECEIVER_JETI             4
                                                   # RECEIVER_ACT_DSL          5
                                                   # RECEIVER_HOTT             6
                                                   # RECEIVER_SBUS             7
                                                   # RECEIVER_UNKNOWN          0xFF
	    $MkSetting{'I_Faktor'},                # Wert : 0-250
	    $MkSetting{'UserParam1'},              # Wert : 0-250
	    $MkSetting{'UserParam2'},              # Wert : 0-250
	    $MkSetting{'UserParam3'},              # Wert : 0-250
	    $MkSetting{'UserParam4'},              # Wert : 0-250
	    $MkSetting{'ServoNickControl'},        # Wert : 0-250     # Stellung des Servos
	    $MkSetting{'ServoNickComp'},           # Wert : 0-250     # Einfluss Gyro/Servo
	    $MkSetting{'ServoNickMin'},            # Wert : 0-250     # Anschlag
	    $MkSetting{'ServoNickMax'},            # Wert : 0-250     # Anschlag
	    $MkSetting{'ServoRollControl'},        # Wert : 0-250     # Stellung des Servos
	    $MkSetting{'ServoRollComp'},           # Wert : 0-250
	    $MkSetting{'ServoRollMin'},            # Wert : 0-250
	    $MkSetting{'ServoRollMax'},            # Wert : 0-250
	    $MkSetting{'ServoNickRefresh'},        # Speed of the Servo
            $MkSetting{'ServoManualControlSpeed'},
            $MkSetting{'CamOrientation'},
	    $MkSetting{'Servo3'},                  # Value or mapping of the Servo Output
            $MkSetting{'Servo4'},                  # Value or mapping of the Servo Output
	    $MkSetting{'Servo5'},                  # Value or mapping of the Servo Output
	    $MkSetting{'LoopGasLimit'},            # Wert: 0-250  max. Gas whrend Looping
	    $MkSetting{'LoopThreshold'},           # Wert: 0-250  Schwelle fr Stickausschlag
	    $MkSetting{'LoopHysterese'},           # Wert: 0-250  Hysterese fr Stickausschlag
	    $MkSetting{'AchsKopplung1'},           # Wert: 0-250  Faktor, mit dem Gier die Achsen Roll und Nick koppelt (NickRollMitkopplung)
	    $MkSetting{'AchsKopplung2'},           # Wert: 0-250  Faktor, mit dem Nick und Roll verkoppelt werden
	    $MkSetting{'CouplingYawCorrection'},   # Wert: 0-250  Faktor, mit dem Nick und Roll verkoppelt werden
	    $MkSetting{'WinkelUmschlagNick'},      # Wert: 0-250  180-Punkt
	    $MkSetting{'WinkelUmschlagRoll'},      # Wert: 0-250  180-Punkt
	    $MkSetting{'GyroAccAbgleich'},         # 1/k  (Koppel_ACC_Wirkung)
	    $MkSetting{'Driftkomp'},
	    $MkSetting{'DynamicStability'},
	    $MkSetting{'UserParam5'},              # Wert : 0-250
	    $MkSetting{'UserParam6'},              # Wert : 0-250
	    $MkSetting{'UserParam7'},              # Wert : 0-250
	    $MkSetting{'UserParam8'},              # Wert : 0-250
	    $MkSetting{'J16Bitmask'},              # for the J16 Output
	    $MkSetting{'J16Timing'},               # for the J16 Output
	    $MkSetting{'J17Bitmask'},              # for the J17 Output
	    $MkSetting{'J17Timing'},               # for the J17 Output
	    $MkSetting{'WARN_J16_Bitmask'},        # for the J16 Output
	    $MkSetting{'WARN_J17_Bitmask'},        # for the J17 Output
	    $MkSetting{'NaviGpsModeControl'},      # Parameters for the Naviboard
	    $MkSetting{'NaviGpsGain'},
	    $MkSetting{'NaviGpsP'},
	    $MkSetting{'NaviGpsI'},
	    $MkSetting{'NaviGpsD'},
	    $MkSetting{'NaviGpsPLimit'},
	    $MkSetting{'NaviGpsILimit'},
	    $MkSetting{'NaviGpsDLimit'},
	    $MkSetting{'NaviGpsACC'},
	    $MkSetting{'NaviGpsMinSat'},
	    $MkSetting{'NaviStickThreshold'},
	    $MkSetting{'NaviAccCompensation'},
	    $MkSetting{'NaviSpeedCompensation'},
	    $MkSetting{'NaviOperatingRadius'},
	    $MkSetting{'NaviAngleLimitation'},
	    $MkSetting{'NaviPH_LoginTime'},
	    $MkSetting{'ExternalControl'},         # for serial Control
	    $MkSetting{'OrientationAngle'},        # Where is the front-direction?
	    $MkSetting{'CareFreeModeControl'},     # switch for CareFree
            $MkSetting{'MotorSafetySwitch'},
            $MkSetting{'MotorSmooth'},
            $MkSetting{'ComingHomeAltitude'},
            $MkSetting{'FailsaveTime'},
            $MkSetting{'MaxAltitude'},

	    $MkSetting{'BitConfig'},               # CFG_LOOP_OBEN            0x01
                                                   # CFG_LOOP_UNTEN           0x02
                                                   # CFG_LOOP_LINKS           0x04
                                                   # CFG_LOOP_RECHTS          0x08
                                                   # CFG_MOTOR_BLINK	      0x10
                                                   # CFG_MOTOR_OFF_LED1       0x20
                                                   # CFG_MOTOR_OFF_LED2       0x40
                                                   # CFG_MOTOR_BLINK2         0x80
	    $MkSetting{'ServoCompInvert'},         # 0x01 = Nick, 0x02 = Roll
	    $MkSetting{'ExtraConfig'},             # CFG2_HEIGHT_LIMIT        0x01
                                                   # CFG2_VARIO_BEEP          0x02
                                                   # CFG_SENSITIVE_RC         0x04
                                                   # CFG_3_3V_REFERENCE       0x08
	    $MkSetting{'Name'},
            ) = unpack ('C108A12', $Data); 
            }


        elsif ( $Revision >= 85 )
            {
            
            (
	    $MkSetting{'SettingNumber'},          # setting number
            $MkSetting{'Revision'},               # version of setting structure
            $MkSetting{'Kanalbelegung_01'},
	    $MkSetting{'Kanalbelegung_02'},
	    $MkSetting{'Kanalbelegung_03'},
	    $MkSetting{'Kanalbelegung_04'},
	    $MkSetting{'Kanalbelegung_05'},
	    $MkSetting{'Kanalbelegung_06'},
	    $MkSetting{'Kanalbelegung_07'},
	    $MkSetting{'Kanalbelegung_08'},
	    $MkSetting{'Kanalbelegung_09'},
	    $MkSetting{'Kanalbelegung_10'},
	    $MkSetting{'Kanalbelegung_11'},
	    $MkSetting{'Kanalbelegung_12'},
	    $MkSetting{'GlobalConfig'},            # CFG_HOEHENREGELUNG       0x01
                                                   # CFG_HOEHEN_SCHALTER      0x02
                                                   # CFG_HEADING_HOLD         0x04
                                                   # CFG_KOMPASS_AKTIV        0x08
                                                   # CFG_KOMPASS_FIX          0x10
                                                   # CFG_GPS_AKTIV            0x20
                                                   # CFG_ACHSENKOPPLUNG_AKTIV 0x40
                                                   # CFG_DREHRATEN_BEGRENZER  0x80
	    $MkSetting{'Hoehe_MinGas'},            # Wert : 0-100
	    $MkSetting{'Luftdruck_D'},             # Wert : 0-250
	    $MkSetting{'MaxHoehe'},                # Wert : 0-32
	    $MkSetting{'Hoehe_P'},                 # Wert : 0-32
	    $MkSetting{'Hoehe_Verstaerkung'},      # Wert : 0-50
	    $MkSetting{'Hoehe_ACC_Wirkung'},       # Wert : 0-250
	    $MkSetting{'Hoehe_HoverBand'},         # Wert : 0-250
	    $MkSetting{'Hoehe_GPS_Z'},             # Wert : 0-250
	    $MkSetting{'Hoehe_StickNeutralPoint'}, # Wert : 0-250
	    $MkSetting{'Stick_P'},                 # Wert : 1-6
	    $MkSetting{'Stick_D'},                 # Wert : 0-64
	    $MkSetting{'Gier_P'},                  # Wert : 1-20
	    $MkSetting{'Gas_Min'},                 # Wert : 0-32
	    $MkSetting{'Gas_Max'},                 # Wert : 33-250
	    $MkSetting{'GyroAccFaktor'},           # Wert : 1-64
	    $MkSetting{'KompassWirkung'},          # Wert : 0-32
	    $MkSetting{'Gyro_P'},                  # Wert : 10-250
	    $MkSetting{'Gyro_I'},                  # Wert : 0-250
	    $MkSetting{'Gyro_D'},                  # Wert : 0-250
	    $MkSetting{'Gyro_Gier_P'},             # Wert : 10-250
	    $MkSetting{'Gyro_Gier_I'},             # Wert : 0-250
	    $MkSetting{'Gyro_Stability'},          # Wert : 0-16
	    $MkSetting{'UnterspannungsWarnung'},   # Wert : 0-250
	    $MkSetting{'NotGas'},                  # Wert : 0-250     Gaswert bei Empngsverlust
	    $MkSetting{'NotGasZeit'},              # Wert : 0-250     Zeit bis auf NotGas geschaltet wird, wg. Rx-Problemen
	    $MkSetting{'Receiver'},                # RECEIVER_PPM              0
                                                   # RECEIVER_SPEKTRUM         1
                                                   # RECEIVER_SPEKTRUM_HI_RES  2
                                                   # RECEIVER_SPEKTRUM_LOW_RES 3
                                                   # RECEIVER_JETI             4
                                                   # RECEIVER_ACT_DSL          5
                                                   # RECEIVER_UNKNOWN          0xFF
	    $MkSetting{'I_Faktor'},                # Wert : 0-250
	    $MkSetting{'UserParam1'},              # Wert : 0-250
	    $MkSetting{'UserParam2'},              # Wert : 0-250
	    $MkSetting{'UserParam3'},              # Wert : 0-250
	    $MkSetting{'UserParam4'},              # Wert : 0-250
	    $MkSetting{'ServoNickControl'},        # Wert : 0-250     # Stellung des Servos
	    $MkSetting{'ServoNickComp'},           # Wert : 0-250     # Einfluss Gyro/Servo
	    $MkSetting{'ServoNickMin'},            # Wert : 0-250     # Anschlag
	    $MkSetting{'ServoNickMax'},            # Wert : 0-250     # Anschlag
	    $MkSetting{'ServoRollControl'},        # Wert : 0-250     # Stellung des Servos
	    $MkSetting{'ServoRollComp'},           # Wert : 0-250
	    $MkSetting{'ServoRollMin'},            # Wert : 0-250
	    $MkSetting{'ServoRollMax'},            # Wert : 0-250
	    $MkSetting{'ServoNickRefresh'},        # Speed of the Servo
            $MkSetting{'ServoManualControlSpeed'},
            $MkSetting{'CamOrientation'},
	    $MkSetting{'Servo3'},                  # Value or mapping of the Servo Output
            $MkSetting{'Servo4'},                  # Value or mapping of the Servo Output
	    $MkSetting{'Servo5'},                  # Value or mapping of the Servo Output
	    $MkSetting{'LoopGasLimit'},            # Wert: 0-250  max. Gas whrend Looping
	    $MkSetting{'LoopThreshold'},           # Wert: 0-250  Schwelle fr Stickausschlag
	    $MkSetting{'LoopHysterese'},           # Wert: 0-250  Hysterese fr Stickausschlag
	    $MkSetting{'AchsKopplung1'},           # Wert: 0-250  Faktor, mit dem Gier die Achsen Roll und Nick koppelt (NickRollMitkopplung)
	    $MkSetting{'AchsKopplung2'},           # Wert: 0-250  Faktor, mit dem Nick und Roll verkoppelt werden
	    $MkSetting{'CouplingYawCorrection'},   # Wert: 0-250  Faktor, mit dem Nick und Roll verkoppelt werden
	    $MkSetting{'WinkelUmschlagNick'},      # Wert: 0-250  180-Punkt
	    $MkSetting{'WinkelUmschlagRoll'},      # Wert: 0-250  180-Punkt
	    $MkSetting{'GyroAccAbgleich'},         # 1/k  (Koppel_ACC_Wirkung)
	    $MkSetting{'Driftkomp'},
	    $MkSetting{'DynamicStability'},
	    $MkSetting{'UserParam5'},              # Wert : 0-250
	    $MkSetting{'UserParam6'},              # Wert : 0-250
	    $MkSetting{'UserParam7'},              # Wert : 0-250
	    $MkSetting{'UserParam8'},              # Wert : 0-250
	    $MkSetting{'J16Bitmask'},              # for the J16 Output
	    $MkSetting{'J16Timing'},               # for the J16 Output
	    $MkSetting{'J17Bitmask'},              # for the J17 Output
	    $MkSetting{'J17Timing'},               # for the J17 Output
	    $MkSetting{'WARN_J16_Bitmask'},        # for the J16 Output
	    $MkSetting{'WARN_J17_Bitmask'},        # for the J17 Output
	    $MkSetting{'NaviGpsModeControl'},      # Parameters for the Naviboard
	    $MkSetting{'NaviGpsGain'},
	    $MkSetting{'NaviGpsP'},
	    $MkSetting{'NaviGpsI'},
	    $MkSetting{'NaviGpsD'},
	    $MkSetting{'NaviGpsPLimit'},
	    $MkSetting{'NaviGpsILimit'},
	    $MkSetting{'NaviGpsDLimit'},
	    $MkSetting{'NaviGpsACC'},
	    $MkSetting{'NaviGpsMinSat'},
	    $MkSetting{'NaviStickThreshold'},
	    $MkSetting{'NaviWindCorrection'},
	    $MkSetting{'NaviAccCompensation'},
	    $MkSetting{'NaviOperatingRadius'},
	    $MkSetting{'NaviAngleLimitation'},
	    $MkSetting{'NaviPH_LoginTime'},
	    $MkSetting{'ExternalControl'},         # for serial Control
	    $MkSetting{'OrientationAngle'},        # Where is the front-direction?
	    $MkSetting{'OrientationModeControl'},  # switch for CareFree
            $MkSetting{'MotorSafetySwitch'},
	    $MkSetting{'BitConfig'},               # CFG_LOOP_OBEN            0x01
                                                   # CFG_LOOP_UNTEN           0x02
                                                   # CFG_LOOP_LINKS           0x04
                                                   # CFG_LOOP_RECHTS          0x08
                                                   # CFG_MOTOR_BLINK	      0x10
                                                   # CFG_MOTOR_OFF_LED1       0x20
                                                   # CFG_MOTOR_OFF_LED2       0x40
                                                   # CFG_RES4		      0x80
	    $MkSetting{'ServoCompInvert'},         # 0x01 = Nick, 0x02 = Roll
	    $MkSetting{'ExtraConfig'},             # CFG2_HEIGHT_LIMIT        0x01
                                                   # CFG2_VARIO_BEEP          0x02
                                                   # CFG_SENSITIVE_RC         0x04
                                                   # CFG_3_3V_REFERENCE       0x08
	    $MkSetting{'Name'},
            ) = unpack ('C104A12', $Data); 
            }

        else
            {
            print "mkcomm.pl: Settings structure revision $Revision not supported\n";
            }
        }

    elsif ( $Id eq "J" )
        #
        # NC Parameter
        #
        {
        my ($Id, $Value) = unpack ('Cs', $Data);
        }

    elsif ( $Id eq "k" )
        #
        # FC compass heading
        #
        {
        # ignore
        }

    elsif ( $Id eq "Z" )
        #
        # NC Connection Checker
        #
        {
        $Mk{'ConnectionChecker'} = unpack ('S', $Data);
        }

    else
        {
        print "Unknown Command: $Header $Data\n";
        }
    }	


# send Target or Waypoint to MK
sub MkFlyTo()
    {
    my %Param = @_;

    my $x                 = $Param{'-x'};
    my $y                 = $Param{'-y'};
    my $Lat               = $Param{'-lat'};
    my $Lon               = $Param{'-lon'};
    my $Alt               = $Param{'-alt'};
    my $AltRate           = $Param{'-altrate'};
    my $Heading           = $Param{'-heading'};          # <0:POI, 0:no change, 1..360, >360:invalid
    my $ToleranceRadius   = $Param{'-toleranceradius'};
    my $Holdtime          = $Param{'-holdtime'};
    my $EventFlag         = $Param{'-eventflag'};
    my $EventChannelValue = $Param{'-eventchannelvalue'};
    my $Mode              = $Param{'-mode'};
    my $Index             = $Param{'-index'};            # 1..n (dummy ... will be overwritten in NC:uart1.c)
    my $Type              = $Param{'-type'};             # 0=WP, 1=POI
    my $Speed             = $Param{'-speed'};            # km/h
    my $CamAngle          = $Param{'-camangle'};         # -1 = POI
    my $Name              = $Param{'-name'};             # ASCII

    if ( $x eq ""    and  $y   eq "" ) { ($x, $y)           = &MapGps2XY($Lat, $Lon); }
    if ( $Lat eq ""  and  $Lon eq "" ) { ($Lat, $Lon)       = &MapXY2Gps($x, $y); }
    if ( $Alt eq "" )                  { $Alt               = &Altitude(); }
    if ( $AltRate eq "" )              { $AltRate           = $Cfg->{'waypoint'}->{'DefaultAltRate'}; }
    if ( $Heading eq "" )              { $Heading           = $Cfg->{'waypoint'}->{'DefaultHeading'}; }
    if ( $ToleranceRadius eq "" )      { $ToleranceRadius   = $Cfg->{'waypoint'}->{'DefaultToleranceRadius'}; }
    if ( $Holdtime eq "" )             { $Holdtime          = $Cfg->{'waypoint'}->{'DefaultHoldtime'}; }
    if ( $EventFlag eq "" )            { $EventFlag         = $Cfg->{'waypoint'}->{'DefaultEventFlag'}; }
    if ( $EventChannelValue eq "" )    { $EventChannelValue = $Cfg->{'waypoint'}->{'DefaultEventChannelValue'}; }
    if ( $Speed eq "" )                { $Speed             = $Cfg->{'waypoint'}->{'DefaultSpeed'}; }
    if ( $CamAngle eq "" )             { $CamAngle          = $Cfg->{'waypoint'}->{'DefaultCamAngle'}; }
    if ( $Index eq "" )                { $Index = 0; }
    if ( $Type eq "" )                 { $Type = 0; }

    my $Status = 1;     # valid
    if ( $Mode =~ /delete/i )
        {
        $Status = 0;    # invalid -> delete NC WP-List
        $Index = -1;    # required from NC0.19 onward
        }

    $AltRate = abs ($AltRate);
    if ( $Mode =~ /target/i  and ! &MkIsCarefree() )
        {
        # Disable Altitude control, if not in CF mode
        $AltRate = 0;
        }

    # Altitude bypass, may be used to set Altitude in Event
    if ( $System{'AltBypass'} ne "" )
        {
        $AltRate = 0;    # no change
        }

    # Gier bypass, may be used to set Heading in Event
    if ( $System{'GierBypass'} ne "" )
        {
        $Heading = 0;    # no change
        }

    if ( $Mode =~ /waypoint/i  and  $Name eq "" )
        {
        $Name = sprintf ("P%03d", $Index + 1);
        }
    if ( $Mode =~ /target/i  and  $Name eq "" )
        {
        $Name = sprintf ("T%03d", $FlyToTargetNum % 1000);
        $FlyToTargetNum ++;
        }
	
    my $Lat_i = sprintf "%d", $Lat * 10000000;
    my $Lon_i = sprintf "%d", $Lon * 10000000;
    my $Alt_i = sprintf "%d", $Alt * 10 * $Cfg->{'map'}->{'AltFactor'} / 20;  # NC: 20 Steps/m

    $Speed = int($Speed / 3.6 * 10);  # km/h --> 0.1 m/s   

    # Datenstruktur:  (see NC waypoints.h)
    #    GPS_Pos_t Position;      // the gps position of the waypoint, see ubx.h for details
    #    s16 Heading;             // orientation, future implementation
    #    u8  ToleranceRadius;     // in meters, if the MK is within that range around the target, then the next target is 
    #    u8  HoldTime;            // in seconds, if the MK was once in the tolerance area around a WP,
    #                             // this time defines the delay before the next WP is triggered
    #    u8  Event_Flag;          // future emplementation
    #    u8  Index;               // to indentify different waypoints, workaround for bad communications PC <-> NC
    #    u8  Type;                // typeof Waypoint (0=WP, 1=POI)
    #    u8  WP_EventChannelValue // 
    #    u8  AltitudeRate;        // rate to change the setpoint (0.1 m/s)
    #    u8  Speed;               // 0.1 m/s
    #    u8  CamAngle;            // Camera servo angle
    #    u8  Name[4];             // Name of that Point (ASCII)
    #    u8  reserve[2];          // reserved

    my $Wp = pack ('lllCsC9A4C2', 
                   $Lon_i,
                   $Lat_i,
                   $Alt_i,
                   $Status,
                   $Heading,
                   $ToleranceRadius,
                   $Holdtime,
                   $EventFlag,
                   $Index + 1,
                   $Type,
                   $EventChannelValue,
                   $AltRate,
                   $Speed,
                   $CamAngle,
                   $Name,
                   0,0,
                  );

    if ( $Mode =~ /waypoint/i )
        {
        &Log ("Send WP $Name Index: $Index Lat: $Lat Lon: $Lon Alt: $Alt AltRate: $AltRate Heading: $Heading Speed: $Speed", "S");

        $MkSendQueue->enqueue( "w", "$AddrNC", $Wp );
        # &MkSend( "w", "$AddrNC", $Wp );
        }
    elsif ( $Mode =~ /target/i )
        {
        &Log ("Send Target $Name Index: $Index Lat: $Lat Lon: $Lon Alt: $Alt AltRate: $AltRate Heading: $Heading Speed: $Speed", "S");

        $MkSendQueue->enqueue( "s", "$AddrNC", $Wp );
        # &MkSend( "s", "$AddrNC", $Wp );

        # set System information
        $System{'TargetSetpoint_x'}        = $x;
        $System{'TargetSetpoint_y'}        = $y;
        $System{'TargetSetpoint_Lat'}      = $Lat;
        $System{'TargetSetpoint_Lon'}      = $Lon;
        $System{'TargetSetpoint_Alt'}      = $Alt;
        $System{'TargetSetpoint_AltRate'}  = $AltRate;
        $System{'TargetSetpoint_Speed'}    = $Speed;

        # set Target information for Simulator
        $MkSim{'Target_Lat'}               = $Lat;
        $MkSim{'Target_Lon'}               = $Lon;
        $MkSim{'Target_Alt'}               = $Alt;
        $MkSim{'Target_AltRate'}           = $AltRate;
        $MkSim{'Target_Status'}            = $Status;
        $MkSim{'Target_Heading'}           = $Heading;
        $MkSim{'Target_ToleranceRadius'}   = $ToleranceRadius;
        $MkSim{'Target_Holdtime'}          = $Holdtime;
        $MkSim{'Target_EventFlag'}         = $EventFlag;
        $MkSim{'Target_EventChannelValue'} = $EventChannelValue;
        $MkSim{'Target_Speed'} = $Speed;
        $MkSim{'Target_CamAngle'} = $CamAngle;

        # Timestamp, wann der Datensatz geschtieben wurde
        $MkSim{'_Timestamp'} = time;
        }
    else
        {
        # unknown mode, ignore
        }

    ## Set Heading
    ## Heading Parameter in WP/Target is not working (NC 0.23)
    ## if ( $Heading != 0 )
    ##     {
    ##     &NcSetParam(1, $Heading);
    ##     }

    return 0;
    }


# set NC Parameter
sub NcSetParam()
    {
    my ($Id, $Value) = @_;

    &Log ("Send NC Param Id: $Id Value: $Value", "S");

    # Id 0: Target Speed       0.1 m/s
    # Id 1: Gier Setpoint      degree
    # Id 2: Camera Elevation   0.1 degree
    # Id 3: Altitude Rate      0.1 m/s
    # Id 4: Altitude setpoint  0.1 m

    my $Param = pack ('CCs', 
                       1,     # Set Parameter
                       $Id,
                       $Value,
                     );

    $MkSendQueue->enqueue( "j", "$AddrNC", $Param );
    # &MkSend( "j", "$AddrNC", $Param );

    # Heading for simulator
    if ( $Id == 1 )
        {
        $MkSim{'Target_Heading'} = $Value;
        }

    return 0;
    }


# send External control to MK
sub SendExternalControl()
    {
    my %Param = @_;

    my $RemoteButtons = $Param{'-remotebuttons'};
    my $Nick   = $Param{'-nick'};
    my $Roll   = $Param{'-roll'};
    my $Gier   = $Param{'-gier'};
    my $Gas    = $Param{'-gas'};
    my $Hight  = $Param{'-hight'};
    my $Free   = $Param{'-free'};
    my $Frame  = $Param{'-frame'};
    my $Config = $Param{'-config'};

    &Log ("Send External Control Nick: $Nick Roll: $Roll Gier: $Gier Gas: $Gas", "S");

    # Datenstruktur:
    #    u8      Digital[2];
    #    u8      RemoteButtons;
    #    s8      Nick;
    #    s8      Roll;
    #    s8      Yaw;
    #    u8      Gas;
    #    s8      Height;
    #    u8      free;
    #    u8      Frame;
    #    u8      Config;

    # Config/Bit 0 and FC-Parameter ExternControl > 128:
    #    Nich/Roll/Yaw added to RC-Channel
    #    Gas wird auf max. RC-Gas begrenzt

    my $Ec = pack ('CCCcccCcCCC',
                   0, 0, 
                   $RemoteButtons,
                   $Nick,
                   $Roll,
                   $Gier,
                   $Gas,
                   $Hight,
                   $Free,
                   $Frame,       # Frame/Command counter, ungleich 0
                   $Config,
                  );

    $MkSendQueue->enqueue( "b", "$AddrFC", $Ec );
    # &MkSend( "b", "$AddrFC", $Ec );

    return 0;
    }


# send serial Channel values from %MkSerialChannel to MK/FC
sub SendSerialChannel()
    {

    # Datenstruktur:
    #    s8      Channel[12];

    lock (%MkSerialChannel);    # until end of block

    &Log ("Send Serial Channel " .
                   "$MkSerialChannel{'SerialChannel01'}," .
                   "$MkSerialChannel{'SerialChannel02'}," .
                   "$MkSerialChannel{'SerialChannel03'}," .
                   "$MkSerialChannel{'SerialChannel04'}," .
                   "$MkSerialChannel{'SerialChannel05'}," .
                   "$MkSerialChannel{'SerialChannel06'}," .
                   "$MkSerialChannel{'SerialChannel07'}," .
                   "$MkSerialChannel{'SerialChannel08'}," .
                   "$MkSerialChannel{'SerialChannel09'}," .
                   "$MkSerialChannel{'SerialChannel10'}," .
                   "$MkSerialChannel{'SerialChannel11'}," .
                   "$MkSerialChannel{'SerialChannel12'}", "S");
    
    my $SP = pack ('c12', 
                   $MkSerialChannel{'SerialChannel01'},
                   $MkSerialChannel{'SerialChannel02'},
                   $MkSerialChannel{'SerialChannel03'},
                   $MkSerialChannel{'SerialChannel04'},
                   $MkSerialChannel{'SerialChannel05'},
                   $MkSerialChannel{'SerialChannel06'},
                   $MkSerialChannel{'SerialChannel07'},
                   $MkSerialChannel{'SerialChannel08'},
                   $MkSerialChannel{'SerialChannel09'},
                   $MkSerialChannel{'SerialChannel10'},
                   $MkSerialChannel{'SerialChannel11'},
                   $MkSerialChannel{'SerialChannel12'},
                  );

    $MkSendQueue->enqueue( "y", "$AddrFC", $SP );
    # &MkSend( "y", "$AddrFC", $SP );

    return 0;
    }


# Switch NC Debug UART to NC
sub MkSwitchToNC()
    {
    # &Log ("Send switch UART to NC", "S");

    $MkSendQueue->enqueue( "_DIRECT", "$AddrNC", pack ("C5", 0x1B, 0x1B, 0x55, 0xAA, 0x00) );
    }

# Switch NC Debug UART to FC
sub MkSwitchToFC()
    {
    # &Log ("Send switch UART to FC", "S");

    $MkSendQueue->enqueue( "u", "$AddrNC", pack ("C", 0) );
    }

# Switch NC Debug UART to MK3MAG
sub MkSwitchToMK3MAG()
    {
    # &Log ("Send switch UART to MK3MAG", "S");

    $MkSendQueue->enqueue( "u", "$AddrNC", pack ("C", 1) );
    }

# Switch NC Debug UART to MKGPS
sub MkSwitchToMKGPS()
    {
    # &Log ("Send switch UART to GPS", "S");

    $MkSendQueue->enqueue( "u", "$AddrNC", pack ("C", 2) );
    }


# Request current setting from FC
sub MkRequestSetting()
    {
    for ( my $i=0; $i < 5; $i++)
        {
        &MkSwitchToFC();
        usleep (100000);

        $MkSendQueue->enqueue( "q", "$AddrFC", pack ("C", 255) );
        usleep (250000);

        if ( $MkSetting{'SettingNumber'} ne "" )
            {
            last;
            }
        usleep (100000);
        }
 
    &MkSwitchToNC();
    }


# when called as thread
sub MkCommLoop()
    {
    while (1)
       {
       &MkIO();
       }
    }


#
# Hauptprgramm
#	

if ( $0 =~ /mkcomm.pl$/i )
    {
    # Program wurde direkt aufgerufen
    &MkCommLoop();

    # should never exit
    }
	
1;

__END__

